// Vincia.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains global header information for Vincia.

#ifndef VINCIA_H
#define VINCIA_H

// Include Pythia 8 headers and name space
#include "Pythia.h"

// Now include headers that need this information
#include "VinciaAntenna.h"
#include "VinciaCommon.h"
#include "VinciaShower.h"
#include "VinciaMatching.h"
#include "VinciaMGInterface.h"

using namespace Pythia8;
using namespace Vincia;

// Define namespace inside which Vincia lives
namespace Vincia { 
  
  //*********
  
  // The VinciaPlugin class
  
  class VinciaPlugin {
    
  public:
    
    VinciaPlugin() {
      cerr << "\n (VinciaPlugin():) cannot construct VinciaPlugin :(\n\n"
	   <<"  Sorry, the Vincia plugin must be constructed with a pointer"
	   <<" to an\n  explicit Pythia object (since it's a plugin)." 
	   <<" Please use one of \n\n"
	   <<"     VinciaPlugin vincia(Pythia* pythia);\n"
	   <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile);\n"
	   <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile, string xmlDir);\n"
	   <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile, string xmlDir, string xmlIndexFile);\n"
	   <<"\n";
    }
    
    // Constructor
    VinciaPlugin(Pythia* pythia, string commandFile="none", 
		 string xmlDir="xmldoc", 
		 string xmlIndexFile="VinciaIndex.xml") {
      
      // First set pointers to Pythia classes 
      coupSMPtr         = &pythia->coupSM;
      infoPtr           = &pythia->info;
      particleDataPtr   = &pythia->particleData;
      partonSystemsPtr  = &pythia->partonSystems;
      rndmPtr           = &pythia->rndm;
      settingsPtr       = &pythia->settings;
      // Initial value for pointer to user hooks.
      userHooksPtr      = 0;
     
      // Add the Vincia list of parameters to the Pythia 8 settings database
      settingsPtr->init(xmlDir+"/"+xmlIndexFile, true); 
      addAntennaSettings();    
      
      // Init default shower pointers and default settings
      shower.RPtr        = &R;
      shower.matchingPtr = &matching;
      shower.vinCluPtr   = &vinClu;
      shower.initPtr(infoPtr, settingsPtr, particleDataPtr, rndmPtr,
		     coupSMPtr, partonSystemsPtr, userHooksPtr) ;

      // Read command file 1st time
      if (commandFile != "none") pythia->readFile(commandFile); 
      verbose = 0;
      int verboseGlo = settingsPtr->mode("Vincia:verbose");
      
      // Vincia on/off
      vinciaOn = settingsPtr->flag("Vincia");

      // Initialize standard Antennae       
      if (vinciaOn) {
	if (verboseGlo >= 3)
	  cout<<" VinciaPlugin(): initializing MAX antennae"<<endl;
	pythia->readFile(xmlDir+"/antennae-Max.cmnd"); 
	initAntennaSet(antennaeMax);
	if (verboseGlo >= 3) 
	  cout<<" VinciaPlugin(): initializing MIN antennae"<<endl;
	pythia->readFile(xmlDir+"/antennae-Min.cmnd");
	initAntennaSet(antennaeMin);
	if (verboseGlo >= 3)
	  cout<<" VinciaPlugin(): initializing DEF antennae"<<endl;
	pythia->readFile(xmlDir+"/antennae-Def.cmnd"); 
	initAntennaSet(antennaeDef);

	// If filename given for non-default antenna presets, read file.
	string antennaFile=settingsPtr->word("Vincia:antennaFile");
	if (antennaFile != "none" && antennaFile != "") { 
	  if (verboseGlo >= 2) 
	    cout << " VinciaPlugin(): reading USR antenna presets from "
		 <<antennaFile<<endl;
	  pythia->readFile(antennaFile);
	}
	
	// Read command file 2nd time
	if (commandFile != "none") pythia->readFile(commandFile); 
	
	// Initialize physical antennae
	if (verboseGlo >= 3)
	  cout<<" VinciaPlugin(): initializing USR antennae"<<endl;
	initAntennaSet(antennae);      
	
	// Now set verbose
	verbose = verboseGlo;
	
	// Hand antenna set pointers to shower and matching objects
	shower.initAntPtr(&antennae, &antennaeDef, &antennaeMax, &antennaeMin);
	matching.initAntPtr(&antennae, &antennaeDef, &antennaeMax, &antennaeMin);

	// Set Pythia 8 shower pointer to Vincia 
	if (settingsPtr->flag("Vincia:nonRes")) {
	  pythia->setShowerPtr(&shower,&shower,0);    
	} else {
	  pythia->setShowerPtr(&shower,0,0);    
	}

      } 

      // If VINCIA off, let us know about it
      else {
	if (settingsPtr->flag("Vincia:nonRes")) {
	  if (verbose >= 2) printOut("VinciaPlugin","Vincia is switched off");
	  settingsPtr->flag("Vincia:nonRes",false);
	}
      }

      // Pass pointers on to objects that require them
      R.initPtr(settingsPtr);
      matching.initPtr(&R,&mgInterface,settingsPtr,&vinClu);    
      mgInterface.initPtr(coupSMPtr,particleDataPtr,settingsPtr);
      rambo.initPtr(rndmPtr);
      vinClu.initPtr(rndmPtr);
      
    } 

    //*********
    
    // Add placeholders for antenna functions in Settings database
    // (only ones with non-zero coefficients will be saved later,
    // but the definitions are needed in case user wants to update.)

    void addAntennaSettings() {

      for (int iAnt=0; iAnt<=10; iAnt++) {
	string name;
	if (iAnt == iqqEmit) name="Vincia:qQ:emit";
	else if (iAnt == iqgEmit) name="Vincia:qg:emit";
	else if (iAnt == iqgSplit) name="Vincia:qg:split";
	else if (iAnt == iggEmit) name="Vincia:gg:emit";
	else if (iAnt == iggSplit) name="Vincia:gg:split";
	else continue;

	// Check if global switch exists
	if ( ! settingsPtr->isFlag(name) ) 
	  settingsPtr->addFlag(name,false);
	
	//Check if chargeFactor exists
	string parm(name+":chargeFactor");
	if ( ! settingsPtr->isParm(parm) )
	  settingsPtr->addParm(parm,0.0,false,false,0.0,0.0); 
	
	// Add the C, D, E, K Laurent coefficients
	for (int i=-1;i<=LaurentIndexMax;i++) {
	  string iStr(num2str(i,0));
	  for (int j=-1;j<=LaurentIndexMax;j++) {
	    string jStr(num2str(j,0));
	    parm=name+":LaurentC("+iStr+","+jStr+")";
	    if ( ! settingsPtr->isParm( parm ) )  
	      settingsPtr->addParm(parm,0.0,false,false,0.0,0.0); 
	  }	
	  parm=name+":LaurentD("+iStr+")";
	  if ( !  settingsPtr->isParm(parm) )  
	    settingsPtr->addParm(parm,0.0,false,false,0.0,0.0);   
	  parm=name+":LaurentE("+iStr+")";
	  if ( !  settingsPtr->isParm(parm) )  
	    settingsPtr->addParm(parm,0.0,false,false,0.0,0.0);   
	}
	// Add the K coefficient
	parm=name+":LaurentK";
	if ( !  settingsPtr->isParm(parm) )  
	  settingsPtr->addParm(parm,0.0,false,false,0.0,0.0);   
	// ...
      }
    }  

    //*********
    
    // Initialize an antenna set (e.g., Default, Max, Min, etc)

    void initAntennaSet(AntennaSet& antennaeIn) {

      // Read in and initialize generic Sudakovs 
      for (int iAnt=0; iAnt<=10; iAnt++) {
	string name;
	if (iAnt == iqqEmit) name="Vincia:qQ:emit";
	else if (iAnt == iqgEmit) name="Vincia:qg:emit";
	else if (iAnt == iggEmit) name="Vincia:gg:emit";
	else if (iAnt == iqgSplit) name="Vincia:qg:split";
	else if (iAnt == iggSplit) name="Vincia:gg:split";
	else continue;
	
	// If this name exists and is switched on, go ahead:
	if ( settingsPtr->isFlag(name) ) {
	  
	  // Construct and initialize generic antenna
	  Antenna myAntenna(name);
	  myAntenna.initPtr(particleDataPtr, rndmPtr, settingsPtr);
	  bool pass=myAntenna.init();      	
	  
	  // If all OK, add to antenna list, else print error message
	  if (pass) {	
	    
	    // Save LL physical antenna
	    antennaeIn.setAnt(iAnt,myAntenna);
	    
	    // Verbose output
	    if (verbose >= 3)
	      printOut(myAntenna.getHumanName()+":init","ok: added to antenna list");
	  } else 
	    if (verbose >= 1) printErr(name+":init",myAntenna.getHumanName()+
				       " has a problem and will not be used in this run!");
	}
      }      
    }

    // Automatically set verbose level in all inheriting members
    void setVerbose(int verboseIn) {
      R.setVerbose(verboseIn);
      matching.setVerbose(verboseIn);
      mgInterface.setVerbose(verboseIn);
      shower.setVerbose(verboseIn);
    }

    // Main VINCIA on/off switch
    bool vinciaOn;

    // Instantiate Vincia Objects
    EvolutionVariable R;
    Matching          matching;
    MGInterface       mgInterface;
    Rambo             rambo;
    VinciaShower      shower;
    VinClu            vinClu;  
    
    // Vectors of physical Antennae with uncertainty bands
    AntennaSet antennae;
    AntennaSet antennaeMax;
    AntennaSet antennaeDef;
    AntennaSet antennaeMin;

    // Pointers to Pythia classes
    CoupSM*        coupSMPtr;
    Info*          infoPtr; 
    ParticleData*  particleDataPtr;
    PartonSystems* partonSystemsPtr;
    Rndm*          rndmPtr; 
    Settings*      settingsPtr; 
    UserHooks*     userHooksPtr;
    
    // Shortcut for post-Vincia event weight
    double weight() {return shower.weight();}
    bool negativeWeightOccurred() {return shower.negativeWeightOccurred;}
    bool nonunityWeightOccurred() {return shower.nonunityWeightOccurred;}
    
    // Utilities for printing Info and Internal histograms
    void printInfo( ostream& os = cout ) { shower.printInfo(os); }
    void printHistos( ostream& os = cout ) { shower.printHistos(os); }

  private :
    
    // Verbose level
    int verbose;

  }; // end Vincia plugin
  
} // end Vincia namespace
#endif
