// VinciaMatching.cc is a part of the VINCIA plug-in to the PYTHIA 8 
// generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the Matching class. 

#include "VinciaMatching.h"

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  // Initialize Matching interface
  //***********************************************************************

  bool Matching::init() {

    // Check if pointers initialized
    if (! isInitPtr) {
      printOut("Matching::init","Cannot initialize, pointers not set.");
      return false;
    }
    
    // Innocent until proven guilty
    isInit = true;

    // Settings
    verbose          = settingsPtr->mode("Vincia:verbose");
    colorDepth       = settingsPtr->mode("Vincia:matchingColorDepth");
    uncertaintyBands = settingsPtr->flag("Vincia:uncertaintyBands");
    sectorOrdering   = settingsPtr->flag("Vincia:sectorOrdering");

    // Check MG interface
    isMGOn = settingsPtr->flag("Vincia:MGInterface");
    if (isMGOn) {
      isMGOn = mgInterfacePtr->init();
      settingsPtr->flag("Vincia:MGInterface",isMGOn);
      if (isMGOn && verbose >= 2) {
	printOut("Matching::init","MGInterface switched on.");
      }
    }

    return isInit;
  }
	
  //***********************************************************************
  // Main functions for VINCIA matching
  //***********************************************************************

  //*******

  // isMatched():
  // Check if any matching implemented for this number of legs

  bool Matching::isMatched(int idRes, int nEmit, int nSplit, bool tryMG) {
    
    // Return value false by default
    bool gotit = false;

    // -------------------------------------------------------------------
    // Matching: with MG interface
    // -------------------------------------------------------------------
    if (tryMG) {

      // Spin-1 Charge-0 Color-Singlet (Z0) Decays
      if (idRes == 23) {
	if (nEmit+nSplit <= 3) gotit = true;
	else if (nEmit == 4) gotit = true;
      } 
      
      // Spin-1 Charge-1 Color-Singlet (W) Decays
      // (currently using same as Charge-0)
      else if (idRes == 24) {
	if (nEmit+nSplit <= 3) gotit = true;
	else if (nEmit == 4) gotit = true;
      }
      
    }

    return gotit;

  } 
  
  //*******

  // getPME():
  // Return reweighting factors in PME
  //   PME[0] : reweighting factor with USR antennae
  //   PME[1:3] : reweighting factor with MAX/MIN/DEF antennae
  // = |M_n|^2 / (Sum_i A^phys_i |M_n-1|^2)

  bool Matching::getPME(int nIn, vector<int> id, vector<Vec4> p, 
			vector<double>& PME) {

    // Resize PME.
    PME.resize(0);
    bool done = false;

    // Matching using MadGraph
    if (isMGOn) {

      //cout<<" PME get MEn"<<endl;

      // Get color-ordered M2 (= LC + ...) for this n-parton state
      double Mn2     = mgInterfacePtr->ME2(nIn, id, p);

      // mgInterface returns -1 if not existing. 
      if (abs(Mn2+1) < TINY) return false;

      // Check if previous ME exists
      if (id.size()-nIn <= 3) {
	//	cout << " getPME : prevME not existing"<<endl;
	//	hasOldME = false;
      } else {
	// hasOldME = true;
	//nOldME   = id.size()
      }      

      //cout<<"         MEn = "<<Mn2<<endl;

      // Check if return was -1 ?
      if (abs(Mn2+1.0) < TINY) return false; 

      // Initialize (n-1)->n vector of approximations
      vector<double> Mn2Appr(0);
      for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {
	if (iSet >= 1 && ! uncertaintyBands) continue;
	Mn2Appr.push_back(0.);
      }

      // Loop over possible emitter antennae
      for (int iClu=0; iClu<int(id.size())-3; iClu++) {
	
	// Clustering:
	//cout<<" PME Clustering "<<iClu<<"/"<<int(int(id.size())-4)<<endl;

	// TMP! Hardcode for pT unordering
	int uoType = 1;
	
	// Compute new ordering scale
	int ir  = 2+iClu;
	int ida = id[ir-1];
	int idr = id[ir];
	int idb = id[ir+1];
	double Q2New = RPtr->Q2E(p[ir-1],p[ir],p[ir+1],uoType);

	//cout <<"  arb = "<<ir-1<<" "<<ir<<" "<<ir+1<<" QNew = "<<sqrt(Q2New)<<endl;

	// Reconstruct previous state
	vector<int> idClu(id);
	vector<Vec4> pClu;
	int idaHat  = ida;
	int idbHat  = idb;
	int antIndx = -1;
	bool swapAB = false;
	if (idr == 21) {
	  // To recluster gluon, just collapse parent ids
	  idClu.erase(idClu.begin()+ir);
	  if (ida == 21 && idb == 21) antIndx = iggEmit;
	  else if (idb == 21) antIndx = iqgEmit;
	  else if (ida == 21) {antIndx = iqgEmit; swapAB=true;}
	  else antIndx = iqqEmit;
	} else if (idr < 0) {
	  // To recluster antiquark, collapse with downstream quark 
	  idClu[ir+1]=21;
	  idClu.erase(idClu.begin()+ir);
	  idbHat = 21;
	  if (idaHat == 21) antIndx = iggSplit;
	  else antIndx = iqgSplit;
	} else {
	  // To recluster quark, collapse with upstream antiquark
	  idClu[ir-1]=21;
	  idClu.erase(idClu.begin()+ir);
	  idaHat = 21;
	  if (idbHat == 21) {antIndx = iggSplit; swapAB=true;}
	  else {antIndx = iqgSplit; swapAB=true;}
	}
	if (verbose >= 9) {
	  cout << "PME Clustering "<<iClu<<" : ";
	  for (int i=0;i<int(idClu.size());i++) cout << " " << idClu[i];
	  if (antIndx == iqgEmit) cout << (swapAB ? " gqbarEmit" : " qgEmit");
	  cout<<" Ant = "<<antennaSets[0]->getAnt(antIndx)->getHumanName();
	  cout<<endl;
	}
	// Reconstruct clustered momenta
	int iMap = antennaSets[0]->getAnt(antIndx)->getKineMapType();
	vinCluPtr->vc3to2(iMap,p,ir-1,ir,ir+1,pClu);
	if (verbose >= 9) {
	  for (int i=0;i<int(idClu.size());i++) {
	    cout << "   p" << i<<" "<<pClu[i].px()<<" "<<pClu[i].py()
		 <<" "<<pClu[i].pz()<<" "<<pClu[i].e()<<endl;
	  }
	}

	// Invariant mass of parent antenna and daughter invariants
	double sP  = m2(pClu[ir-1],pClu[ir]);
	double yar = m2(p[ir-1],p[ir])/sP;
	double yrb = m2(p[ir],p[ir+1])/sP;
	if (swapAB) {
	  double ytmp = yar;
	  yar = yrb;
	  yrb = ytmp;
	}

	// Compute Pimp factor 
	double Pimp = 1.0;
	// Only apply if at least 3 partons in final state
	if ( int(pClu.size()-2) > nIn) { 
	  double Q2Hat = RPtr->Q2E(pClu[1],pClu[2],pClu[3],uoType);
	  for (int jClu=1; jClu<int(pClu.size())-3; jClu++) {
	    int jaHat=jClu+1;
	    int jrHat=jClu+2;
	    int jbHat=jClu+3;
	    Q2Hat = min(Q2Hat,
			RPtr->Q2E(pClu[jaHat],pClu[jrHat],pClu[jbHat],uoType));
	  }	  
	  Pimp = Q2Hat/(Q2Hat+Q2New);
	  if (verbose >= 8) 
	    printOut("Matching::getPME","Pimp = "+num2str(Pimp));
	}

	// Compute ARIADNE factor (for g->qqbar splittings)
	double Pari = 1.0;
	if (antIndx == iqgSplit || antIndx == iggSplit) {
	  double sN;
	  if (!swapAB) {
	    // Only apply if neighbor is in final state
	    if (ir+1 < int(pClu.size()) ) { 
	      sN = m2(pClu[ir],pClu[ir+1]);
	      // Only apply if neighbor is a quark
	      if (idClu[ir+1] != 21) 
		Pari = 2.0*sN/(sP + sN);
	    }
	  } else { 
	    // Only apply if neighbor is in final state
	    if (ir-2 >= nIn) {
	      sN = m2(pClu[ir-2],pClu[ir-1]);	
	      // Only apply if neighbor is a quark
	      if (idClu[ir-2] != 21)  
		Pari = 2.0*sN/(sP + sN);
	    }
	  }
	  
	  // Verbose output
	  if (verbose >= 8) 
	    printOut("Matching::getPME","Pari = "+num2str(Pari));
	  
	}
	
	// Total prefactor
	double comFac = Pimp * Pari;

	// Get color-ordered M2 (= LC + ...) for this (n-1)-parton state
	double Mp2 = mgInterfacePtr->ME2(nIn, idClu, pClu);

	// TMP!
	if (verbose >= 9) {
	  cout << " PME Mp2 = "<<Mp2<<endl;

	  // Test
	  double sC01  = m2(pClu[1],pClu[2]);
	  double sC12  = m2(pClu[2],pClu[3]);
	  double sC    = m2(pClu[1],pClu[2],pClu[3]);
	  double A3chk = CF*antennaSets[0]->getAnt(iqqEmit)->a2(sC01/sC,sC12/sC)/sC;
	  double A3chk2 = antennaSets[0]->getAnt(iqqEmit)->getChargeFactor()
	    *antennaSets[0]->getAnt(iqqEmit)->a2(sC01/sC,sC12/sC)/sC;
	  cout << " PME A30 = "<<scientific<<A3chk<<" s = "<<" A30chk = "
	       << A3chk2<<endl;
	}

	// Check if return was -1 ?
	if (abs(Mp2+1.0) < TINY) return false; 

	// Loop over different antenna sets (for uncertainty evaluation)
	for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {

	  // Only include uncertainty variations if doing uncertainty bands.
	  if (iSet >= 1 && ! uncertaintyBands) continue;
	  
	  // Compute physical antenna for this set for this (n-1)-parton state
	  double AntPhys = antennaSets[iSet]->getAnt(antIndx)->a2(yar,yrb)/sP;
	  AntPhys *= antennaSets[iSet]->getAnt(antIndx)->getChargeFactor();
	  if (verbose >= 9 && iSet == 0) {
	    double sar = yar*sP;
	    double srb = yrb*sP;
	    double D30chk  = D30(sar,srb,sP-sar-srb,sP);
	    D30chk *= CF;
	    cout << scientific<<" PME D30 = "<<AntPhys<<" D30chk = "
		 <<D30chk<<endl;
	  }
	  AntPhys *= comFac;
	  // Add to Approximated Mn2 for this set
	  Mn2Appr[iSet] += Mp2 * AntPhys;

	}

      }
      
      // Store ME reweight factors
      for (int iSet=0; iSet<int(antennaSets.size()); iSet++) {
	if (iSet >= 1 && ! uncertaintyBands) continue;
	PME.push_back(Mn2/Mn2Appr[iSet]);
      }
      
      // Store Old ME
      //oldME2 = Mn2;
      

      // Done
      done = true;

    } 

    // For the time being, no non-MadGraph matching implemented
    return done;

  }

  //***********************************************************************
  // Functions for specific matrix element weights
  //***********************************************************************

  //***********************************************************************
  // A30 (q_a,g_r,qbar_b) = Z->3 at tree level / |M_2|^2
  // From GGG, JHEP 0509:056,2005

  double Matching::A30(double sar,double srb,double sab,double s) {

    // A30
    double A=(sar/srb+srb/sar+2*sab*s/sar/srb)/s;

    // Return answer
    return A;
  }

  // Same thing, but with a vector<Vec4> as input
  double Matching::A30(vector<Vec4>& pn, int a, int r, int b) {
    if (max(a,max(r,b)) > int(pn.size()) || min(a,min(r,b)) < 0) return 0.0;
    double sar = m2(pn[a],pn[r]);
    double srb = m2(pn[r],pn[b]);
    double sab = m2(pn[a],pn[b]);
    double s   = m2(pn[a],pn[r],pn[b]);
    double A   = A30(sar,srb,sab,s);
    return A;
  }

  // Same thing, but multiplied by s and with rescaled invariants as input
  double Matching::sA30(double yar,double yrb) {
    
    // s*A30
    double sA=(yar/yrb+yrb/yar+2*(1.-yar-yrb)/yar/yrb);
    
    // Return answer
    return sA;
  }

  //***********************************************************************
  // d30 (qg->q_a,g_r,g_b)
  // From GGG, JHEP 0509:056,2005

  double Matching::D30(double sar,double srb,double sab,double s) {
    
    // d30
    double A=(2*s*sab/sar/srb+(sab*srb+pow2(srb))/sar/s
		    + sar*sab/srb/s+2.5+0.5*srb/s)/s;

    // Return answer
    return A;

  }

  //***********************************************************************
  // d30sng = d30 with zero finite terms

  double Matching::D30sng(double sar,double srb,double sab,double s) {
    
    // d30
    double A=D30(sar,srb,sab,s) - 2.5/s + sar/pow2(s) - 1.5*srb/pow2(s);

    // Return answer
    return A;

  }

  //***********************************************************************
  // E30 (qg->q_a,qbar_r'q_b') 
  // From GGG, JHEP 0509:056,2005
  // Note: actual antenna function in xml file is 1/2 this.
  // Counter-checked since need 1/2*E30 to retrieve E40 in limit mqq->0

  double Matching::E30(double sar,double srb,double sab,double s) {
    
    // E30 
    double A=(pow2(sab)+pow2(sar))/srb/s+sab/s+sar/s;

    // Return answer
    return A/s;

  }

  //***********************************************************************
  // E30sng = E30 with zero finite terms
  // Corrected relative to LH proceedings. GGG form given there is
  // not correct. The correct finite terms are
  //    C00 = -0.5, C10=1.0, C01=0.0
  double Matching::E30sng(double sar,double srb,double sab,double s) {
    
    // e30 - finite (still twice what is in the code)
    double A=E30(sar,srb,sab,s) + 1.0/s - 2.*sar/pow2(s) ;

    // Return answer
    return A;

  }

  //***********************************************************************
  // E30pzs = E30 with finite terms tuned to Z->4 and Z->5
  double Matching::E30pzs(double sar,double srb,double sab,double s) {
    
    // e30 - finite (still twice what is in the code)
    double A=E30(sar,srb,sab,s) + (-0.5 + 0.5*srb/s)/s ;

    // Return answer
    return A;

  }

  //***********************************************************************
  // A30ar (qqbar->q_a,g_r,qbar_b)
  // Ariadne function for qqbar->qgqbar
  double Matching::A30ar(double sar, double srb, double sab, double s) {
    
    //A30 Ariadne is identical to A30 GGG
    return A30(sar, srb, sab, s); 

  }

  //***********************************************************************
  // d30ar (qg->q_a,g_r,g_b)
  // Ariadne function for qg->qgg
  double Matching::D30ar(double sar, double srb, double sab, double s) {
    
    //d30 Ariadne
    double A=2.0*sab*s/sar/srb-s/srb+srb/sar+3.*sar/srb-pow2(sar)/srb/s;

    // Return answer
    return A/s;

  }

  //***********************************************************************
  // f30ar (qg->q_a,g_r,g_b)
  // Ariadne function for qg->qgg
  double Matching::F30ar(double sar, double srb, double sab, double s) {
    
    //f30 Ariadne
    double A=2.0*sab*s/sar/srb-s/srb-s/sar+3.*srb/sar+3.*sar/srb
      -pow2(srb)/sar/s-pow2(sar)/srb/s;

    // Return answer
    return A/s;

  }

  //***********************************************************************
  // E30ar (qg->q_a,qbar_r'q_b')
  // From Gustafson / Ariadne 
  // Note: actual antenna function in xml is 1/2 this.
  // Counter-checked since need 1/2*E30ar to retrieve E40 in limit mqq->0

  double Matching::E30ar(double sar,double srb,double sab,double s) {
    
    // E30 Ariadne 
    double A=2.0*((sab-0.5*s)/srb+pow2(sar)/srb/s+sar/s+0.5*srb/s);

    // Return answer
    return A/s;

  }

  //***********************************************************************
  // f30 (gg->g_a,g_r,g_b)
  // From GGG, JHEP 0509:056,2005

  double Matching::F30(double sar,double srb,double sab,double s) {
    
    // f30 
    double A = 
      2*sab/sar/srb 
      + sab*sar/srb/pow2(s) + sab*srb/sar/pow2(s) 
      + 8.0/3.0/s;

    // Return answer
    return A;

  }

  //***********************************************************************
  // F30sng = f30 with zero finite terms

  double Matching::F30sng(double sar,double srb,double sab,double s) {
    
    // f30 
    double A = F30(sar,srb,sab,s) - 8.0/3.0/s + sar/pow2(s) + srb/pow2(s);

    // Return answer
    return A;

  }

  //***********************************************************************
  // G30 (gg->g_a,qbar_r'q_b')
  // From GGG, JHEP 0509:056,2005

  double Matching::G30(double sar,double srb,double sab,double s) {
    
    // G30 
    double A=((pow2(sab)+pow2(sar))/srb/s)/s;

    // Return answer
    return A;

  }

  //***********************************************************************
  // G30sng = G30 with zero finite terms
 
  double Matching::G30sng(double sar,double srb,double sab,double s) {
    
    // G30 
    double A=G30(sar,srb,sab,s) + 1.0/s - sar/pow2(s) - 0.5*srb/pow2(s);

    // Return answer
    return A;

  }

  //***********************************************************************
  // Q30 
  // From GGGH, JHEP 0711:058,2007

  double Matching::Q30(double sar,double srb,double s) {
    
    // Q30 
    double A=( - pow2(sar)/srb/s ) / s;
    // Add positive finite terms
    A += (5.0/2.0 + 1.5*srb/s)/s;
    // Subtract negative finite terms
    A +=  - 1.0*sar/s/s;
    // Return answer
    return A;

  }

  //***********************************************************************
  // E30vc = E30 with smaller finite terms
  //
  double Matching::E30med(double sar,double srb,double sab,double s) {
    
    // e30 - finite (still twice what is in the code)
    double A=E30(sar,srb,sab,s) - (s - srb - sar*srb/s)/pow2(s);

    // Return answer
    return A;

  }

  //***********************************************************************
  // a40 (q_a,g_b,g_c,qbar_d) = Z->4 at tree level / |M_2|^2: LC and NLC
  // From GGG, JHEP 0509:056,2005 paper

  double Matching::A40(vector<Vec4>& pn, int a, int b, int c, int d) {
    double A = ( A40aux(pn,a,b,c,d) + A40aux(pn,d,c,b,a) );
    return A;
  }
  // Leading-color a40
  double Matching::A40aux(vector<Vec4>& pn, int a,int b,int c,int d) {
    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A = 
      1.0/(2*s13*s24*s34) * (2*s12*s14 + 2*s12*s23 + 2*pow(s12,2) 
				      + pow(s14,2) + pow(s23,2) )
      + 1.0/(2*s13*s24*s134*s234)* (3*s12*pow(s34,2) - 4*pow(s12,2)*s34
				    + 2*pow(s12,3) - pow(s34,3))
      + 1.0/(s13*s24*s134) * (3*s12*s23 - 3*s12*s34 + 4*pow(s12,2)
			      - s23*s34 + pow(s23,2) + pow(s34,2) )
      + 3.0/(2*s13*s24) * (2*s12 + s14 + s23)
      + 1.0/(s13*s34) * (4*s12 + 3*s23 + 2*s24)
      + 1.0/(s13*pow(s134,2)) * (s12*s34 + s23*s34 + s24*s34)
      + 1.0/(s13*s134*s234) * (3*s12*s24 + 6*s12*s34 - 4*pow(s12,2)
			       - 3*s24*s34 - pow(s24,2) - 3*pow(s34,2) )
      + 1.0/(s13*s134) * ( -6*s12 - 3*s23 - s24 + 2*s34 )
      + 1.0/(s24*s34*s134) * ( 2*s12*s14 + 2*s12*s23 + 2*pow(s12,2) 
			       + 2*s14*s23 + pow(s14,2) + pow(s23,2) )
      + 1.0/(s24*s134) * ( -4*s12 - s14 - s23 + s34 )
      + 1.0/pow(s34,2) * ( s12 + 2*s13 - 2*s14 - s34 )
      + 1.0/(pow(s34,2)*pow(s134,2))  * ( 2*s12*pow(s14,2)
					  + 2*pow(s14,2)*(s23 + s24) )
      - (2.0*s12*s14*s24)/(pow(s34,2)*s134*s234) 
      + 1.0/(pow(s34,2)*s134) * (-2*s12*s14 - 4*s14*s24 + 2*pow(s14,2))
      + 1.0/(s34*s134*s234) * (- 2*s12*s14 - 4*pow(s12,2)
			       + 2*s14*s24 - pow(s14,2) - pow(s24,2) )
      + 1.0/(s34*s134) * (- 8*s12 - 2*s23 - 2*s24 )
      + 1.0/pow(s134,2)* (s12 + s23 + s24)
      + 3.0/(2*s134*s234) * (2*s12 + s14 - s24 - s34)
      + 1.0/(2*s134);
    
    // Normalize and return
    A /= s1234;
    return A;
  }

  // A40NLC
  // Only summing over 2 ~a40, -> singularities in 1/sab and 1/scd
  // To get full NLC must sum over b<->c as well.
  // When summed over b<->c, checked gives identical answer to GGG form code.
  double Matching::A40NLC(vector<Vec4>& pn, int a, int b, int c, int d) {
    double A = ( A40auxNLC(pn,a,b,c,d) + A40auxNLC(pn,d,c,b,a) ) ;
    return A;
  }

  // Subleading-color ~a40 
  double Matching::A40auxNLC(vector<Vec4>& pn, int a, int b, int c, int d) {
    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double ANLC =
      1.0/(s13*s24*s134*s234) * ( 3.0/2.0 * s12 * pow(s34,2)
				  - 2*pow(s12,2) * s34 + pow(s12,3)
				  - 1.0/2.0 * pow(s34,3) )
      + 1.0/(s13*s24*s134) * ( 3 * s12 * s23 - 3 * s12 * s34
			       + 4 * pow(s12,2) - s23 * s34
			       + pow(s23,2) + pow(s34,2) )
      + pow(s12,3) / (s13*s24*(s13+s23)*(s14+s24))
      + 1.0/(s13*s24*(s13+s23)) * ( 1.0/2.0*s12*s14 + pow(s12,2) )
      + 1.0/(s13*s24*(s14+s24)) * ( 1.0/2.0*s12*s23 + pow(s12,2) )
      + 1.0/(s13*s24) * (3*s12 + 3.0/2.0*s14 + 3.0/2.0*s23 )
      + 1.0/(s13*pow(s134,2)) * (s12*s34 + s23*s34 + s24*s34 )
      + (2*pow(s12,3)) / (s13*s134*s234*(s13+s23)) 
      + 1.0/(s13*s134*s234) * (3*s12*s34 - s24*s34 - 2*pow(s34,2) )
      + 1.0/(s13*s134*(s13+s23)) * (s12*s24 + s12*s34 + 2*pow(s12,2) )
      + 1.0/(s13*s134) * ( - s23 - s24 + 2*s34 )
      + 1.0/(s13*s234*(s13+s23)) * ( s12*s14 + s12*s34 + 2*pow(s12,2) )
      + 1.0/(s13*s234) * ( - 2*s12 - 2*s14 + s24 + 2*s34 )
      + (2*pow(s12,3)) / (s13*(s13+s23)*(s14+s24)*(s13+s14))
      + 1.0/(s13*(s13+s23)*(s13+s14)) * ( s12*s24 + 2*pow(s12,2) )
      + 1.0/(s13*(s14+s24)*(s13+s14)) * ( s12*s23 + 2*pow(s12,2) )
      + (2*s12) / (s13*(s13+s14))
      - 2.0/s13
      + 1.0/pow(s134,2) * ( s12 + s23 + s24 )
      + 1.0/(s134 * s234) * ( s12 - s34 ) + 1.0/s134;    
    
    // Normalize and return
    ANLC /= s1234;
    return ANLC;

  }

  // A40 from the GGG form code
  // Contains sum over all four ~a40 contributions
  double Matching::A40frm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A = 
      2*s12*pow(s13,-1)*s14*pow(s24,-1)*pow(s34,-1) + 3*s12*pow(s13,-1)
      *s14*pow(s24,-1)*pow(s234,-1) + 2*s12*pow(s13,-1)*s14*pow(s34,-1)
      *pow(s234,-1) + 2*s12*pow(s13,-1)*s23*pow(s24,-1)*pow(s34,-1) + 3
      *s12*pow(s13,-1)*s23*pow(s24,-1)*pow(s134,-1) - 3*s12*pow(s13,-1)
      *pow(s24,-1)*s34*pow(s134,-1) - 3*s12*pow(s13,-1)*pow(s24,-1)*s34
      *pow(s234,-1) + 3*s12*pow(s13,-1)*pow(s24,-1)*pow(s34,2)*pow(
      s134,-1)*pow(s234,-1) + 6*s12*pow(s13,-1)*pow(s24,-1) - 2*s12*
      pow(s13,-1)*s24*pow(s34,-1)*pow(s234,-1) + 3*s12*pow(s13,-1)*s24*
      pow(s134,-1)*pow(s234,-1) + 6*s12*pow(s13,-1)*pow(s34,-1) + s12*
      pow(s13,-1)*s34*pow(s134,-2) + 6*s12*pow(s13,-1)*s34*pow(s134,-1)
      *pow(s234,-1) - 6*s12*pow(s13,-1)*pow(s134,-1) - 6*s12*pow(
      s13,-1)*pow(s234,-1) + 2*s12*s14*pow(s24,-1)*pow(s34,-1)*pow(
      s134,-1) - 3*s12*s14*pow(s24,-1)*pow(s134,-1)*pow(s234,-1) - 4*
      s12*s14*s24*pow(s34,-2)*pow(s134,-1)*pow(s234,-1) - 4*s12*s14*
      pow(s34,-1)*pow(s134,-1)*pow(s234,-1);
      A +=  + 2*s12*pow(s14,2)*pow(s34,-2)*pow(s134,-2) + 2*s12*s23*
      pow(s24,-1)*pow(s34,-1)*pow(s134,-1) + 4*s12*pow(s24,-1)*pow(
      s34,-1) + 3*s12*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1) + s12*
      pow(s24,-1)*s34*pow(s234,-2) - 4*s12*pow(s24,-1)*pow(s134,-1) - 3
      *s12*pow(s24,-1)*pow(s234,-1) + 4*s12*s24*pow(s34,-1)*pow(
      s234,-2) + 2*s12*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 8*s12*pow(
      s34,-1)*pow(s134,-1) - 8*s12*pow(s34,-1)*pow(s234,-1) + s12*pow(
      s134,-2) + 6*s12*pow(s134,-1)*pow(s234,-1) + 3*s12*pow(s234,-2)
       + 2*pow(s12,2)*pow(s13,-1)*pow(s24,-1)*pow(s34,-1) - 4*pow(
      s12,2)*pow(s13,-1)*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1) + 4*
      pow(s12,2)*pow(s13,-1)*pow(s24,-1)*pow(s134,-1) + 4*pow(s12,2)*
      pow(s13,-1)*pow(s24,-1)*pow(s234,-1) + 2*pow(s12,2)*pow(s13,-1)*
      pow(s34,-1)*pow(s234,-1) - 4*pow(s12,2)*pow(s13,-1)*pow(s134,-1)*
      pow(s234,-1) + 2*pow(s12,2)*pow(s24,-1)*pow(s34,-1)*pow(s134,-1)
       - 4*pow(s12,2)*pow(s24,-1)*pow(s134,-1)*pow(s234,-1) - 8*pow(
      s12,2)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1);
      A +=  + 2*pow(s12,3)*pow(s13,-1)*pow(s24,-1)*pow(s134,-1)*pow(
      s234,-1) - pow(s13,-1)*s14*pow(s24,-1)*s34*pow(s234,-1) + 3*pow(
      s13,-1)*s14*pow(s24,-1) - 2*pow(s13,-1)*s14*s24*pow(s34,-1)*pow(
      s234,-1) + 2*pow(s13,-1)*s14*pow(s34,-1) - 3*pow(s13,-1)*s14*pow(
      s234,-1) + pow(s13,-1)*pow(s14,2)*pow(s24,-1)*pow(s34,-1) + pow(
      s13,-1)*pow(s14,2)*pow(s24,-1)*pow(s234,-1) + pow(s13,-1)*pow(
      s14,2)*pow(s34,-1)*pow(s234,-1) - pow(s13,-1)*s23*pow(s24,-1)*s34
      *pow(s134,-1) + 3*pow(s13,-1)*s23*pow(s24,-1) + 4*pow(s13,-1)*s23
      *pow(s34,-1) + pow(s13,-1)*s23*s34*pow(s134,-2) - 3*pow(s13,-1)*
      s23*pow(s134,-1) + pow(s13,-1)*pow(s23,2)*pow(s24,-1)*pow(s34,-1)
       + pow(s13,-1)*pow(s23,2)*pow(s24,-1)*pow(s134,-1) + pow(s13,-1)*
      pow(s24,-1)*pow(s34,2)*pow(s134,-1) + pow(s13,-1)*pow(s24,-1)*
      pow(s34,2)*pow(s234,-1) - pow(s13,-1)*pow(s24,-1)*pow(s34,3)*pow(
      s134,-1)*pow(s234,-1) + pow(s13,-1)*s24*pow(s34,-1) + pow(s13,-1)
      *s24*s34*pow(s134,-2) - 3*pow(s13,-1)*s24*s34*pow(s134,-1)*pow(
      s234,-1);
      A +=  - pow(s13,-1)*s24*pow(s134,-1) + 3*pow(s13,-1)*s24*pow(
      s234,-1) + pow(s13,-1)*pow(s24,2)*pow(s34,-1)*pow(s234,-1) - pow(
      s13,-1)*pow(s24,2)*pow(s134,-1)*pow(s234,-1) + 2*pow(s13,-1)*s34*
      pow(s134,-1) + 3*pow(s13,-1)*s34*pow(s234,-1) - 3*pow(s13,-1)*
      pow(s34,2)*pow(s134,-1)*pow(s234,-1) - 2*pow(s13,-1) + 2*s13*pow(
      s24,-1)*pow(s34,-1) + s13*pow(s24,-1)*s34*pow(s234,-2) - 2*s13*
      pow(s24,-1)*pow(s234,-1) + 4*s13*s24*pow(s34,-1)*pow(s234,-2) + 2
      *s13*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 3*s13*pow(s34,-1)*pow(
      s234,-1) + 3*s13*pow(s234,-2) + 2*s14*s23*pow(s24,-1)*pow(s34,-1)
      *pow(s134,-1) + 3*s14*pow(s24,-1)*pow(s34,-1) + s14*pow(s24,-1)*
      s34*pow(s134,-1)*pow(s234,-1) + s14*pow(s24,-1)*s34*pow(s234,-2)
       - s14*pow(s24,-1)*pow(s134,-1) - 2*s14*pow(s24,-1)*pow(s234,-1)
       - 4*s14*s24*pow(s34,-2)*pow(s134,-1) - 4*s14*s24*pow(s34,-2)*
      pow(s234,-1) + 4*s14*s24*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 
      4*s14*s24*pow(s34,-1)*pow(s234,-2) + 2*s14*pow(s24,2)*pow(s34,-2)
      *pow(s234,-2);
      A +=  - 2*s14*pow(s34,-1)*pow(s134,-1) - 5*s14*pow(s34,-1)*pow(
      s234,-1) + 3*s14*pow(s134,-1)*pow(s234,-1) + 3*s14*pow(s234,-2)
       + 2*pow(s14,2)*s23*pow(s34,-2)*pow(s134,-2) + pow(s14,2)*pow(
      s24,-1)*pow(s34,-1)*pow(s134,-1) - pow(s14,2)*pow(s24,-1)*pow(
      s134,-1)*pow(s234,-1) + 2*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-2)
       + 2*pow(s14,2)*pow(s34,-2)*pow(s134,-1) - 2*pow(s14,2)*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) - s23*pow(s24,-1)*pow(s134,-1)
       - 3*s23*pow(s34,-1)*pow(s134,-1) + s23*pow(s134,-2) + pow(s23,2)
      *pow(s24,-1)*pow(s34,-1)*pow(s134,-1) + pow(s24,-1)*s34*pow(
      s134,-1) - pow(s24,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) - s24
      *pow(s34,-1)*pow(s134,-1) + 2*s24*pow(s34,-1)*pow(s234,-1) + s24*
      pow(s134,-2) - 3*s24*pow(s134,-1)*pow(s234,-1) + 2*pow(s24,2)*
      pow(s34,-2)*pow(s234,-1) - 2*pow(s24,2)*pow(s34,-1)*pow(s134,-1)*
      pow(s234,-1) - 2*pow(s34,-1) - 3*s34*pow(s134,-1)*pow(s234,-1) + 
      pow(s134,-1);

    A /= s1234;

    return A;

  }

  // ~A40 from the GGG form code
  double Matching::A40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A =
      s12*pow(s13,-1)*pow(s14,-1)*pow(s23,-1)*s24 + s12*pow(s13,-1)*
      pow(s14,-1)*s23*pow(s24,-1) + 2*s12*pow(s13,-1)*pow(s14,-1) + s12
      *pow(s13,-1)*s14*pow(s23,-1)*pow(s24,-1) + s12*pow(s13,-1)*s14*
      pow(s23,-1)*pow(s234,-1) + 3*s12*pow(s13,-1)*s14*pow(s24,-1)*pow(
      s234,-1) + s12*pow(s13,-1)*pow(s23,-1)*s24*pow(s134,-1) + s12*
      pow(s13,-1)*pow(s23,-1)*s34*pow(s134,-1) + s12*pow(s13,-1)*pow(
      s23,-1)*s34*pow(s234,-1) + 3*s12*pow(s13,-1)*s23*pow(s24,-1)*pow(
      s134,-1) - 3*s12*pow(s13,-1)*pow(s24,-1)*s34*pow(s134,-1) - 3*s12
      *pow(s13,-1)*pow(s24,-1)*s34*pow(s234,-1) + 3*s12*pow(s13,-1)*
      pow(s24,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 6*s12*pow(
      s13,-1)*pow(s24,-1) + s12*pow(s13,-1)*s34*pow(s134,-2) + 3*s12*
      pow(s13,-1)*s34*pow(s134,-1)*pow(s234,-1) - 2*s12*pow(s13,-1)*
      pow(s234,-1) + s12*s13*pow(s14,-1)*pow(s23,-1)*pow(s24,-1) + 3*
      s12*s13*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) + s12*s13*pow(
      s14,-1)*pow(s24,-1)*pow(s234,-1);
      A +=  + 3*s12*pow(s14,-1)*pow(s23,-1)*s24*pow(s134,-1) - 3*s12*
      pow(s14,-1)*pow(s23,-1)*s34*pow(s134,-1) - 3*s12*pow(s14,-1)*pow(
      s23,-1)*s34*pow(s234,-1) + 3*s12*pow(s14,-1)*pow(s23,-1)*pow(
      s34,2)*pow(s134,-1)*pow(s234,-1) + 6*s12*pow(s14,-1)*pow(s23,-1)
       + s12*pow(s14,-1)*s23*pow(s24,-1)*pow(s134,-1) + s12*pow(s14,-1)
      *pow(s24,-1)*s34*pow(s134,-1) + s12*pow(s14,-1)*pow(s24,-1)*s34*
      pow(s234,-1) + s12*pow(s14,-1)*s34*pow(s134,-2) + 3*s12*pow(
      s14,-1)*s34*pow(s134,-1)*pow(s234,-1) - 2*s12*pow(s14,-1)*pow(
      s234,-1) + 2*s12*pow(s23,-1)*pow(s24,-1) + 3*s12*pow(s23,-1)*s34*
      pow(s134,-1)*pow(s234,-1) + s12*pow(s23,-1)*s34*pow(s234,-2) - 2*
      s12*pow(s23,-1)*pow(s134,-1) + 3*s12*pow(s24,-1)*s34*pow(s134,-1)
      *pow(s234,-1) + s12*pow(s24,-1)*s34*pow(s234,-2) - 2*s12*pow(
      s24,-1)*pow(s134,-1) + 2*s12*pow(s134,-2) + 4*s12*pow(s134,-1)*
      pow(s234,-1) + 2*s12*pow(s234,-2) + 2*pow(s12,2)*pow(s13,-1)*pow(
      s14,-1)*pow(s23,-1) + 2*pow(s12,2)*pow(s13,-1)*pow(s14,-1)*pow(
      s24,-1);
      A +=  + 2*pow(s12,2)*pow(s13,-1)*pow(s23,-1)*pow(s24,-1) + 2*pow(
      s12,2)*pow(s13,-1)*pow(s23,-1)*pow(s134,-1) + 2*pow(s12,2)*pow(
      s13,-1)*pow(s23,-1)*pow(s234,-1) - 4*pow(s12,2)*pow(s13,-1)*pow(
      s24,-1)*s34*pow(s134,-1)*pow(s234,-1) + 4*pow(s12,2)*pow(s13,-1)*
      pow(s24,-1)*pow(s134,-1) + 4*pow(s12,2)*pow(s13,-1)*pow(s24,-1)*
      pow(s234,-1) + 2*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*pow(s24,-1)
       - 4*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*s34*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*pow(s134,-1) + 4*
      pow(s12,2)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) + 2*pow(s12,2)*
      pow(s14,-1)*pow(s24,-1)*pow(s134,-1) + 2*pow(s12,2)*pow(s14,-1)*
      pow(s24,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s13,-1)*pow(s14,-1)*
      pow(s23,-1)*pow(s24,-1) + 2*pow(s12,3)*pow(s13,-1)*pow(s23,-1)*
      pow(s134,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s13,-1)*pow(s24,-1)*
      pow(s134,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s14,-1)*pow(s23,-1)*
      pow(s134,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s14,-1)*pow(s24,-1)*
      pow(s134,-1)*pow(s234,-1);
      A +=  - pow(s13,-1)*s14*pow(s24,-1)*s34*pow(s234,-1) + 3*pow(
      s13,-1)*s14*pow(s24,-1) - 2*pow(s13,-1)*s14*pow(s234,-1) + pow(
      s13,-1)*pow(s14,2)*pow(s24,-1)*pow(s234,-1) - pow(s13,-1)*s23*
      pow(s24,-1)*s34*pow(s134,-1) + 3*pow(s13,-1)*s23*pow(s24,-1) + 
      pow(s13,-1)*s23*s34*pow(s134,-2) - pow(s13,-1)*s23*pow(s134,-1)
       + pow(s13,-1)*pow(s23,2)*pow(s24,-1)*pow(s134,-1) + pow(s13,-1)*
      pow(s24,-1)*pow(s34,2)*pow(s134,-1) + pow(s13,-1)*pow(s24,-1)*
      pow(s34,2)*pow(s234,-1) - pow(s13,-1)*pow(s24,-1)*pow(s34,3)*pow(
      s134,-1)*pow(s234,-1) + pow(s13,-1)*s24*s34*pow(s134,-2) - pow(
      s13,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - pow(s13,-1)*s24*pow(
      s134,-1) + pow(s13,-1)*s24*pow(s234,-1) + 2*pow(s13,-1)*s34*pow(
      s134,-1) + 2*pow(s13,-1)*s34*pow(s234,-1) - 2*pow(s13,-1)*pow(
      s34,2)*pow(s134,-1)*pow(s234,-1) - 2*pow(s13,-1) - s13*pow(
      s14,-1)*pow(s23,-1)*s34*pow(s234,-1) + 3*s13*pow(s14,-1)*pow(
      s23,-1) - 2*s13*pow(s14,-1)*pow(s234,-1) + s13*pow(s23,-1)*s34*
      pow(s234,-2);
      A +=  - s13*pow(s23,-1)*pow(s234,-1) + s13*pow(s24,-1)*s34*pow(
      s234,-2) - s13*pow(s24,-1)*pow(s234,-1) + 2*s13*pow(s234,-2) + 
      pow(s13,2)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) - pow(s14,-1)*
      pow(s23,-1)*s24*s34*pow(s134,-1) + 3*pow(s14,-1)*pow(s23,-1)*s24
       + pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s134,-1) + pow(s14,-1)*
      pow(s23,-1)*pow(s34,2)*pow(s134,-1) + pow(s14,-1)*pow(s23,-1)*
      pow(s34,2)*pow(s234,-1) - pow(s14,-1)*pow(s23,-1)*pow(s34,3)*pow(
      s134,-1)*pow(s234,-1) + pow(s14,-1)*s23*s34*pow(s134,-2) - pow(
      s14,-1)*s23*pow(s134,-1) + pow(s14,-1)*s24*s34*pow(s134,-2) + 
      pow(s14,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - pow(s14,-1)*s24*
      pow(s134,-1) - pow(s14,-1)*s24*pow(s234,-1) + pow(s14,-1)*s34*
      pow(s134,-1) + pow(s14,-1)*s34*pow(s234,-1) - pow(s14,-1)*pow(
      s34,2)*pow(s134,-1)*pow(s234,-1) - pow(s14,-1) - s14*pow(s23,-1)*
      s34*pow(s134,-1)*pow(s234,-1) + s14*pow(s23,-1)*s34*pow(s234,-2)
       + s14*pow(s23,-1)*pow(s134,-1) - s14*pow(s23,-1)*pow(s234,-1) + 
      s14*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1);
      A +=  + s14*pow(s24,-1)*s34*pow(s234,-2) - s14*pow(s24,-1)*pow(
      s134,-1) - s14*pow(s24,-1)*pow(s234,-1) + 2*s14*pow(s234,-2) - 2*
      pow(s23,-1)*s24*pow(s134,-1) + 2*pow(s23,-1)*s34*pow(s134,-1) + 2
      *pow(s23,-1)*s34*pow(s234,-1) - 2*pow(s23,-1)*pow(s34,2)*pow(
      s134,-1)*pow(s234,-1) - 2*pow(s23,-1) - 2*s23*pow(s24,-1)*pow(
      s134,-1) + 2*s23*pow(s134,-2) + pow(s24,-1)*s34*pow(s134,-1) + 
      pow(s24,-1)*s34*pow(s234,-1) - pow(s24,-1)*pow(s34,2)*pow(
      s134,-1)*pow(s234,-1) - pow(s24,-1) + 2*s24*pow(s134,-2) - 4*s34*
      pow(s134,-1)*pow(s234,-1) + 2*pow(s134,-1) + 2*pow(s234,-1);

      A /= s1234;
      
      return A;
  }

  //***********************************************************************
  // B40(q_a,q'_b,qbar'_c,qbar_d)
  // From GGG, JHEP 0509:056,2005
  double Matching::B40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // B40 
    double A = B40aux(pn, a,b,c,d) + B40aux(pn,d,b,c,a) 
      + B40aux(pn,a,c,b,d) + B40aux(pn,d,c,b,a);

    return A;
  }

  // b40 
  // From GGG, JHEP 0509:056,2005
  double Matching::B40aux(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    // b40
    double A = 
      (s12*s13*s14+s13*s14*s23-pow2(s13)*s24)/pow2(s34)/pow2(s134)
      + (-s12*s13*s24+s13*s14*s23-s13*pow2(s24))/pow2(s34)/s134/s234
      + (s12*s13+s13*s23)/s34/pow2(s134)
      + (2.0*s12*s13+pow2(s12))/2.0/s34/s134/s234
      + s12/2.0/s134/s234;

    // Normalize and return
    A /= s1234;
    return A;

  }

  // B40 
  // From GGG Form code, private communication
  double Matching::B40frm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    // B40
    double A =
      4*s12*s14*s24*pow(s34,-2)*pow(s134,-1)*pow(s234,-1) - 2*s12*s14*
      pow(s34,-1)*pow(s134,-2) + 2*s12*s14*pow(s34,-1)*pow(s134,-1)*
      pow(s234,-1) - 2*s12*pow(s14,2)*pow(s34,-2)*pow(s134,-2) + 2*s12*
      s24*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) - 2*s12*s24*pow(s34,-1)
      *pow(s234,-2) - 2*s12*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 2*s12
      *pow(s34,-1)*pow(s134,-1) + 2*s12*pow(s34,-1)*pow(s234,-1) - s12*
      pow(s134,-2) - s12*pow(s234,-2) + 2*pow(s12,2)*pow(s34,-1)*pow(
      s134,-1)*pow(s234,-1) - 2*s13*s24*pow(s34,-1)*pow(s234,-2) - 2*
      s13*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + s13*pow(s34,-1)*pow(
      s234,-1) - s13*pow(s234,-2) - 2*s14*s23*pow(s34,-1)*pow(s134,-2)
       + 4*s14*s24*pow(s34,-2)*pow(s134,-1) + 4*s14*s24*pow(s34,-2)*
      pow(s234,-1) - 2*s14*s24*pow(s34,-1)*pow(s134,-2) - 4*s14*s24*
      pow(s34,-1)*pow(s134,-1)*pow(s234,-1) - 2*s14*s24*pow(s34,-1)*
      pow(s234,-2) - 2*s14*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + s14*
      pow(s34,-1)*pow(s234,-1);
      A +=  - s14*pow(s234,-2) - 2*pow(s14,2)*s23*pow(s34,-2)*pow(
      s134,-2) - 2*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-2) - 2*pow(
      s14,2)*pow(s34,-2)*pow(s134,-1) + 2*pow(s14,2)*pow(s34,-1)*pow(
      s134,-1)*pow(s234,-1) + s23*pow(s34,-1)*pow(s134,-1) - s23*pow(
      s134,-2) + s24*pow(s34,-1)*pow(s134,-1) - s24*pow(s134,-2) - 2*
      pow(s24,2)*pow(s34,-2)*pow(s234,-1) + 2*pow(s24,2)*pow(s34,-1)*
      pow(s134,-1)*pow(s234,-1);

      A /= s1234;
      
      return A;

  }

  // B40 
  // From Ellis, Ross, Terrano (= D in their paper)
  // They then add terms for each possible swapping
  double Matching::B40ert(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    // B40
    double A =
    (1.0/(pow2(s134)*pow2(s34)))*(2*s12*s13*s14 + s13*s14*s23 - pow2(s14)*s23 
				  - pow2(s13)*s24 + s13*s14*s24 + s12*s13*s34 
				  + s12*s14*s34 + s13*s23*s34 + s14*s24*s34) 
    + (1.0/(pow2(s234)*pow2(s34)))*(2*s12*s23*s24 + s13*s23*s24 - s14*pow2(s23)
				    - s13*pow2(s24) + s14*s23*s24 
				    + s12*s23*s34 + s13*s23*s34 + s12*s24*s34 
				    + s14*s24*s34) 
    - (1.0/(s134*s234*pow2(s34))) * (s12*s14*s23 - s13*s14*s23 
				     + s14*pow2(s23) + s12*s13*s24 
				     + pow2(s13)*s24 
				     - s13*s23*s24 - pow2(s12)*s34 
				     - s12*s13*s34 - s12*s23*s34 
				     - s12*pow2(s34))
    - (1.0/(s134*s234*pow2(s34))) * (s12*s14*s23 + pow2(s14)*s23 
				     + s12*s13*s24 - s13*s14*s24 
				     - s14*s23*s24 + s13*pow2(s24) 
				     - pow2(s12)*s34 - s12*s14*s34 
				     - s12*s24*s34 - s12*pow2(s34));
  A /= s1234;
  return A;
}


  //***********************************************************************
  // C40 (q_a,q_b,qbar_c,qbar_d) : identical-flavor-part only
  // From GGG, JHEP 0509:056,2005
  double Matching::C40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // C40 
    double A = C40aux(pn,a,d,b,c) + C40aux(pn,a,c,b,d); 

    return A;
    
  }

  // c40 
  // From GGG, JHEP 0509:056,2005
  double Matching::C40aux(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[b]);
    double s13=m2(pn[a],pn[c]);
    double s14=m2(pn[a],pn[d]);
    double s23=m2(pn[b],pn[c]);
    double s24=m2(pn[b],pn[d]);
    double s34=m2(pn[c],pn[d]);
    double s123=m2(pn[a],pn[b],pn[c]);
    double s134=m2(pn[a],pn[c],pn[d]);
    double s234=m2(pn[b],pn[c],pn[d]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A =  
      - ( s12*s13*s14 )/( 2*s23*s34*s123*s134 )
      + (-s12*s13*s24 + s13*s14*s24 )/( 2*s23*s34*s134*s234 ) 
      - ( s13*pow2(s24) )/( 2*s23*s34*pow2(s234) )
      - ( s12*s13 )/( s23*s123*s134 )
      + (-s12*s14-s12*s34-pow2(s12)+s13*s24 )/( 2*s23*s123*s234 ) 
      + ( s12*s14+s12*s34+pow2(s12)+s13*s24 )/( 2*s23*s134*s234 )  
      - s13/( 2*s123*s134 )
      + ( s12*s24+s14*s24 )/( s23*pow2(s234) ) 
      + (-s12+s14 )/( 2*s123*s234 ); 
    
    A /= s1234;
    return A;
    
  }

  // C40 
  // From GGG form code, private communication
  double Matching::C40frm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s123=m2(pn[a],pn[d],pn[b]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A = 
      1./2.*s12*s14*pow(s23,-1)*pow(s34,-1)*pow(s134,-1) - s12*s14*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) - 1./2.*s12*pow(s23,-1)*pow(
      s34,-1) - s12*pow(s23,-1)*s34*pow(s234,-2) + 1./2.*s12*pow(
      s23,-1)*pow(s134,-1) + s12*pow(s23,-1)*pow(s234,-1) + s12*s24*
      pow(s34,-1)*pow(s234,-2) + 1./2.*s12*pow(s34,-1)*pow(s234,-1) - 
      s12*pow(s234,-2) + 1./2.*pow(s12,2)*pow(s23,-1)*pow(s134,-1)*pow(
      s234,-1) + 1./2.*s13*s14*pow(s23,-1)*pow(s34,-1)*pow(s123,-1) + 
      s13*s14*pow(s23,-1)*pow(s123,-1)*pow(s234,-1) - 1./2.*s13*pow(
      s23,-1)*pow(s34,-1) - s13*pow(s23,-1)*s34*pow(s234,-2) + 3./2.*
      s13*pow(s23,-1)*pow(s123,-1) + 2*s13*pow(s23,-1)*pow(s234,-1) + 
      s13*s24*pow(s34,-1)*pow(s234,-2) + 1./2.*s13*pow(s34,-1)*pow(
      s234,-1) - s13*pow(s123,-1)*pow(s234,-1) - s13*pow(s234,-2) - 1./
      2.*pow(s13,2)*pow(s23,-1)*pow(s34,-1)*pow(s123,-1) + 1./2.*pow(
      s13,2)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 1./2.*s14*pow(
      s23,-1)*pow(s34,-1);
      A +=  + 3*s14*pow(s23,-1)*s34*pow(s123,-1)*pow(s134,-1) + s14*
      pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) - s14*pow(s23,-1)*s34*
      pow(s234,-2) - 2*s14*pow(s23,-1)*pow(s123,-1) + s14*pow(s23,-1)*
      pow(s134,-1) + s14*s24*pow(s34,-1)*pow(s234,-2) - 1./2.*s14*pow(
      s34,-1)*pow(s134,-1) + s14*pow(s123,-1)*pow(s234,-1) + s14*pow(
      s134,-1)*pow(s234,-1) - s14*pow(s234,-2) - pow(s14,2)*pow(s23,-1)
      *pow(s34,-1)*pow(s123,-1) + 1./2.*pow(s14,2)*pow(s23,-1)*pow(
      s34,-1)*pow(s134,-1) + 3*pow(s14,2)*pow(s23,-1)*pow(s123,-1)*pow(
      s134,-1) + 1./2.*pow(s14,2)*pow(s23,-1)*pow(s134,-1)*pow(s234,-1)
       + 1./2.*pow(s14,2)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) + pow(
      s14,3)*pow(s23,-1)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) - pow(
      s23,-1)*s34*pow(s123,-1) + 1./2.*pow(s23,-1)*s34*pow(s134,-1) - 
      pow(s23,-1)*s34*pow(s234,-1) + pow(s23,-1)*pow(s34,2)*pow(
      s123,-1)*pow(s134,-1) + 1./2.*pow(s23,-1)*pow(s34,2)*pow(s134,-1)
      *pow(s234,-1) - 1./2.*pow(s23,-1) + 1./2.*s24*pow(s34,-1)*pow(
      s234,-1);
      A +=  - 1./2.*s24*pow(s134,-1)*pow(s234,-1) + 1./2.*s34*pow(
      s123,-1)*pow(s234,-1) + 1./2.*s34*pow(s134,-1)*pow(s234,-1) - 1./
      2.*pow(s234,-1);

    A /= s1234;

    return A;

  }

  //***********************************************************************
  // D40 (q_a,g_b,g_c,g_d)
  // GGG notation (a = 1, b = 2, c = 3, d = 4)
  // From GGG form code, private communication
  double Matching::D40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[b]);
    double s13=m2(pn[a],pn[c]);
    double s14=m2(pn[a],pn[d]);
    double s23=m2(pn[b],pn[c]);
    double s24=m2(pn[b],pn[d]);
    double s34=m2(pn[c],pn[d]);
    double s123=m2(pn[a],pn[b],pn[c]);
    double s124=m2(pn[a],pn[b],pn[d]);
    double s234=m2(pn[b],pn[c],pn[d]);
    double s134=m2(pn[a],pn[c],pn[d]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A =
      16 + 3*pow(s12,-1)*s13*pow(s14,-1)*pow(s23,-1)*s24*s34 + 3*pow(
      s12,-1)*s13*pow(s14,-1)*pow(s23,-1)*pow(s24,2) + pow(s12,-1)*s13*
      pow(s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s34,-1) + pow(s12,-1)*s13*
      pow(s14,-1)*pow(s23,-1)*pow(s34,2) + 3*pow(s12,-1)*s13*pow(
      s14,-1)*s23*s24*pow(s34,-1) + 3*pow(s12,-1)*s13*pow(s14,-1)*s23
       + pow(s12,-1)*s13*pow(s14,-1)*pow(s23,2)*pow(s34,-1) + 6*pow(
      s12,-1)*s13*pow(s14,-1)*s24 + 3*pow(s12,-1)*s13*pow(s14,-1)*pow(
      s24,2)*pow(s34,-1) + 3*pow(s12,-1)*s13*pow(s14,-1)*s34 + 9*pow(
      s12,-1)*s13*s14*pow(s23,-1)*s24*pow(s34,-1) + pow(s12,-1)*s13*s14
      *pow(s23,-1)*s34*pow(s234,-1) + 17*pow(s12,-1)*s13*s14*pow(
      s23,-1) + 17*pow(s12,-1)*s13*s14*pow(s34,-1) + 2*pow(s12,-1)*s13*
      s14*pow(s234,-1) + 4*pow(s12,-1)*s13*pow(s14,2)*pow(s23,-1)*pow(
      s34,-1) + 4*pow(s12,-1)*s13*pow(s14,2)*pow(s23,-1)*pow(s234,-1)
       + 10*pow(s12,-1)*s13*pow(s14,2)*pow(s34,-1)*pow(s234,-1) + 16*
      pow(s12,-1)*s13*pow(s23,-1)*s24;
    A +=  + 6*pow(s12,-1)*s13*pow(s23,-1)*pow(s24,2)*pow(s34,-1) + 11
      *pow(s12,-1)*s13*pow(s23,-1)*s34 + pow(s12,-1)*s13*pow(s23,-1)*
      pow(s34,2)*pow(s234,-1) + pow(s12,-1)*s13*pow(s23,-1)*pow(s34,3)*
      pow(s124,-1)*pow(s234,-1) + 2*pow(s12,-1)*s13*s23*s24*pow(
      s124,-2) + 8*pow(s12,-1)*s13*s23*pow(s34,-1) + 2*pow(s12,-1)*s13*
      s23*pow(s124,-1) + 4*pow(s12,-1)*s13*pow(s23,2)*pow(s34,-1)*pow(
      s124,-1) + 6*pow(s12,-1)*s13*s24*pow(s34,-1) + 2*pow(s12,-1)*s13*
      s24*s34*pow(s124,-2) - 2*pow(s12,-1)*s13*s24*s34*pow(s124,-1)*
      pow(s234,-1) - pow(s12,-1)*s13*s24*pow(s124,-1) + 2*pow(s12,-1)*
      s13*s24*pow(s234,-1) + 6*pow(s12,-1)*s13*pow(s24,2)*pow(s34,-1)*
      pow(s124,-1) + 6*pow(s12,-1)*s13*pow(s24,2)*pow(s34,-1)*pow(
      s234,-1) - pow(s12,-1)*s13*pow(s24,2)*pow(s124,-1)*pow(s234,-1)
       - 5*pow(s12,-1)*s13*pow(s24,3)*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 2*pow(s12,-1)*s13*s34*pow(s234,-1) + pow(s12,-1)*s13*
      pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 14*pow(s12,-1)*s13 + 6*
      pow(s12,-1)*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*s24;
    A +=  + 3*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*pow(
      s24,2)*pow(s34,-1) + 3*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*pow(
      s23,-1)*s34 + 3*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*s23*pow(
      s34,-1) + 6*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*s24*pow(s34,-1) + 
      6*pow(s12,-1)*pow(s13,2)*pow(s14,-1) + 6*pow(s12,-1)*pow(s13,2)*
      s14*pow(s23,-1)*pow(s34,-1) + 6*pow(s12,-1)*pow(s13,2)*s14*pow(
      s23,-1)*pow(s234,-1) + 10*pow(s12,-1)*pow(s13,2)*s14*pow(s34,-1)*
      pow(s234,-1) + 9*pow(s12,-1)*pow(s13,2)*pow(s23,-1)*s24*pow(
      s34,-1) + 3*pow(s12,-1)*pow(s13,2)*pow(s23,-1)*s34*pow(s234,-1)
       + 3*pow(s12,-1)*pow(s13,2)*pow(s23,-1)*pow(s34,2)*pow(s124,-1)*
      pow(s234,-1) + 15*pow(s12,-1)*pow(s13,2)*pow(s23,-1) + 7*pow(
      s12,-1)*pow(s13,2)*s23*pow(s34,-1)*pow(s124,-1) - 7*pow(s12,-1)*
      pow(s13,2)*s24*pow(s34,-1)*pow(s124,-1) - 7*pow(s12,-1)*pow(
      s13,2)*s24*pow(s34,-1)*pow(s234,-1) + pow(s12,-1)*pow(s13,2)*s24*
      pow(s124,-2) + 9*pow(s12,-1)*pow(s13,2)*pow(s24,2)*pow(s34,-1)*
      pow(s124,-1)*pow(s234,-1);
    A +=  + 18*pow(s12,-1)*pow(s13,2)*pow(s34,-1) + 3*pow(s12,-1)*
      pow(s13,2)*s34*pow(s124,-1)*pow(s234,-1) + 6*pow(s12,-1)*pow(
      s13,2)*pow(s124,-1) + pow(s12,-1)*pow(s13,2)*pow(s234,-1) + 3*
      pow(s12,-1)*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,-1) + 
      4*pow(s12,-1)*pow(s13,3)*pow(s14,-1)*pow(s23,-1) + 4*pow(s12,-1)*
      pow(s13,3)*pow(s14,-1)*pow(s34,-1) + 4*pow(s12,-1)*pow(s13,3)*
      pow(s23,-1)*pow(s34,-1) + 3*pow(s12,-1)*pow(s13,3)*pow(s23,-1)*
      s34*pow(s124,-1)*pow(s234,-1) + pow(s12,-1)*pow(s13,3)*pow(
      s23,-1)*pow(s124,-1) + 4*pow(s12,-1)*pow(s13,3)*pow(s23,-1)*pow(
      s234,-1) - 7*pow(s12,-1)*pow(s13,3)*s24*pow(s34,-1)*pow(s124,-1)*
      pow(s234,-1) + 6*pow(s12,-1)*pow(s13,3)*pow(s34,-1)*pow(s124,-1)
       + 6*pow(s12,-1)*pow(s13,3)*pow(s34,-1)*pow(s234,-1) + 3*pow(
      s12,-1)*pow(s13,3)*pow(s124,-1)*pow(s234,-1) + 2*pow(s12,-1)*pow(
      s13,4)*pow(s14,-1)*pow(s23,-1)*pow(s34,-1) + 2*pow(s12,-1)*pow(
      s13,4)*pow(s23,-1)*pow(s124,-1)*pow(s234,-1) + 2*pow(s12,-1)*pow(
      s13,4)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1);
    A +=  + pow(s12,-1)*pow(s14,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(
      s134,-1) - pow(s12,-1)*pow(s14,-1)*pow(s24,3)*pow(s123,-1) + 12*
      pow(s12,-1)*s14*pow(s23,-1)*s24 + 3*pow(s12,-1)*s14*pow(s23,-1)*
      pow(s24,2)*pow(s34,-1) + 5*pow(s12,-1)*s14*pow(s23,-1)*s34 + pow(
      s12,-1)*s14*pow(s23,-1)*pow(s34,2)*pow(s234,-1) + 2*pow(s12,-1)*
      s14*s23*s24*pow(s34,-1)*pow(s123,-1) + 3*pow(s12,-1)*s14*s23*s24*
      pow(s34,-1)*pow(s134,-1) + 2*pow(s12,-1)*s14*s23*s24*pow(s123,-2)
       + 5*pow(s12,-1)*s14*s23*pow(s34,-1) + 2*pow(s12,-1)*s14*s23*s34*
      pow(s123,-2) + 4*pow(s12,-1)*s14*s23*pow(s123,-1) - 2*pow(s12,-1)
      *s14*s23*pow(s134,-1) + 2*pow(s12,-1)*s14*pow(s23,2)*pow(s34,-1)*
      pow(s134,-1) + 4*pow(s12,-1)*s14*s24*pow(s34,-1) + 3*pow(s12,-1)*
      s14*s24*s34*pow(s123,-1)*pow(s134,-1) + 6*pow(s12,-1)*s14*s24*s34
      *pow(s123,-1)*pow(s234,-1) - 13*pow(s12,-1)*s14*s24*pow(s123,-1)
       - pow(s12,-1)*s14*s24*pow(s134,-1) + 4*pow(s12,-1)*s14*s24*pow(
      s234,-1) - 4*pow(s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s123,-1)
       + 2*pow(s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s134,-1);
    A +=  + 6*pow(s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s234,-1) + 6
      *pow(s12,-1)*s14*pow(s24,2)*pow(s123,-1)*pow(s134,-1) + 9*pow(
      s12,-1)*s14*pow(s24,2)*pow(s123,-1)*pow(s234,-1) + pow(s12,-1)*
      s14*pow(s24,3)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 5*pow(
      s12,-1)*s14*pow(s24,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 6*
      pow(s12,-1)*s14*s34*pow(s123,-1) + 2*pow(s12,-1)*s14*s34*pow(
      s234,-1) + pow(s12,-1)*s14*pow(s34,2)*pow(s123,-1)*pow(s234,-1)
       + 2*pow(s12,-1)*s14 + 4*pow(s12,-1)*pow(s14,2)*pow(s23,-1)*s24*
      pow(s34,-1) - 2*pow(s12,-1)*pow(s14,2)*pow(s23,-1)*s34*pow(
      s234,-1) + 10*pow(s12,-1)*pow(s14,2)*pow(s23,-1) + pow(s12,-1)*
      pow(s14,2)*s23*pow(s34,-1)*pow(s123,-1) - pow(s12,-1)*pow(s14,2)*
      s23*pow(s34,-1)*pow(s134,-1) + pow(s12,-1)*pow(s14,2)*s23*pow(
      s123,-2) - 5*pow(s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s123,-1)
       + pow(s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s134,-1) + 7*pow(
      s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s234,-1) + 6*pow(s12,-1)*
      pow(s14,2)*s24*pow(s123,-1)*pow(s134,-1);
    A +=  + 9*pow(s12,-1)*pow(s14,2)*s24*pow(s123,-1)*pow(s234,-1) + 
      3*pow(s12,-1)*pow(s14,2)*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*pow(
      s134,-1) + 9*pow(s12,-1)*pow(s14,2)*pow(s24,2)*pow(s34,-1)*pow(
      s123,-1)*pow(s234,-1) + 5*pow(s12,-1)*pow(s14,2)*pow(s34,-1) + 
      pow(s12,-1)*pow(s14,2)*s34*pow(s123,-1)*pow(s134,-1) + 3*pow(
      s12,-1)*pow(s14,2)*s34*pow(s123,-1)*pow(s234,-1) - 7*pow(s12,-1)*
      pow(s14,2)*pow(s123,-1) + pow(s12,-1)*pow(s14,2)*pow(s134,-1) + 
      pow(s12,-1)*pow(s14,2)*pow(s234,-1) + 2*pow(s12,-1)*pow(s14,3)*
      pow(s23,-1)*pow(s34,-1) + 2*pow(s12,-1)*pow(s14,3)*pow(s23,-1)*
      pow(s234,-1) + 3*pow(s12,-1)*pow(s14,3)*s24*pow(s34,-1)*pow(
      s123,-1)*pow(s134,-1) + 7*pow(s12,-1)*pow(s14,3)*s24*pow(s34,-1)*
      pow(s123,-1)*pow(s234,-1) - 2*pow(s12,-1)*pow(s14,3)*pow(s34,-1)*
      pow(s123,-1) + pow(s12,-1)*pow(s14,3)*pow(s34,-1)*pow(s134,-1) + 
      6*pow(s12,-1)*pow(s14,3)*pow(s34,-1)*pow(s234,-1) + 2*pow(s12,-1)
      *pow(s14,3)*pow(s123,-1)*pow(s134,-1) + 3*pow(s12,-1)*pow(s14,3)*
      pow(s123,-1)*pow(s234,-1);
    A +=  + pow(s12,-1)*pow(s14,4)*pow(s34,-1)*pow(s123,-1)*pow(
      s134,-1) + 2*pow(s12,-1)*pow(s14,4)*pow(s34,-1)*pow(s123,-1)*pow(
      s234,-1) + 5*pow(s12,-1)*pow(s23,-1)*s24*s34 + 5*pow(s12,-1)*pow(
      s23,-1)*pow(s24,2) + pow(s12,-1)*pow(s23,-1)*pow(s24,3)*pow(
      s34,-1) + 2*pow(s12,-1)*pow(s23,-1)*pow(s34,2) - pow(s12,-1)*s23*
      s24*pow(s34,-1) + 2*pow(s12,-1)*s23*s24*s34*pow(s123,-2) + 2*pow(
      s12,-1)*s23*s24*s34*pow(s124,-2) + 3*pow(s12,-1)*s23*s24*s34*pow(
      s124,-1)*pow(s134,-1) + 4*pow(s12,-1)*s23*s24*pow(s123,-1) - pow(
      s12,-1)*s23*s24*pow(s124,-1) + 6*pow(s12,-1)*s23*s24*pow(s134,-1)
       + pow(s12,-1)*s23*pow(s24,2)*pow(s34,-1)*pow(s123,-1) + 4*pow(
      s12,-1)*s23*pow(s24,2)*pow(s34,-1)*pow(s124,-1) - 3*pow(s12,-1)*
      s23*pow(s24,2)*pow(s34,-1)*pow(s134,-1) + pow(s12,-1)*s23*pow(
      s24,2)*pow(s123,-2) - 6*pow(s12,-1)*s23*pow(s24,2)*pow(s124,-1)*
      pow(s134,-1) + 3*pow(s12,-1)*s23*pow(s24,3)*pow(s34,-1)*pow(
      s124,-1)*pow(s134,-1) + 3*pow(s12,-1)*s23*s34*pow(s123,-1) - 2*
      pow(s12,-1)*s23*s34*pow(s124,-1);
    A +=  - pow(s12,-1)*s23*s34*pow(s134,-1) + pow(s12,-1)*s23*pow(
      s34,2)*pow(s123,-2) + 2*pow(s12,-1)*s23 + 3*pow(s12,-1)*pow(
      s23,2)*s24*pow(s34,-1)*pow(s124,-1) - 3*pow(s12,-1)*pow(s23,2)*
      s24*pow(s34,-1)*pow(s134,-1) + pow(s12,-1)*pow(s23,2)*s24*pow(
      s124,-2) - 3*pow(s12,-1)*pow(s23,2)*s24*pow(s124,-1)*pow(s134,-1)
       + 3*pow(s12,-1)*pow(s23,2)*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*
      pow(s134,-1) - pow(s12,-1)*pow(s23,2)*pow(s34,-1) + 2*pow(s12,-1)
      *pow(s23,2)*pow(s134,-1) + pow(s12,-1)*pow(s23,3)*s24*pow(s34,-1)
      *pow(s124,-1)*pow(s134,-1) + pow(s12,-1)*pow(s23,3)*pow(s34,-1)*
      pow(s124,-1) - pow(s12,-1)*pow(s23,3)*pow(s34,-1)*pow(s134,-1) - 
      5*pow(s12,-1)*s24*s34*pow(s123,-1) + 3*pow(s12,-1)*s24*s34*pow(
      s124,-1) - 3*pow(s12,-1)*s24*s34*pow(s134,-1) + pow(s12,-1)*s24*
      s34*pow(s234,-1) + pow(s12,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(
      s234,-1) + pow(s12,-1)*s24*pow(s34,2)*pow(s124,-2) - pow(s12,-1)*
      s24*pow(s34,2)*pow(s124,-1)*pow(s134,-1) + 3*pow(s12,-1)*s24 + 2*
      pow(s12,-1)*pow(s24,2)*pow(s34,-1);
    A +=  + 3*pow(s12,-1)*pow(s24,2)*s34*pow(s123,-1)*pow(s134,-1) + 
      3*pow(s12,-1)*pow(s24,2)*s34*pow(s123,-1)*pow(s234,-1) + 3*pow(
      s12,-1)*pow(s24,2)*s34*pow(s124,-1)*pow(s134,-1) - 7*pow(s12,-1)*
      pow(s24,2)*pow(s123,-1) - pow(s12,-1)*pow(s24,2)*pow(s124,-1) + 4
      *pow(s12,-1)*pow(s24,2)*pow(s134,-1) + 2*pow(s12,-1)*pow(s24,2)*
      pow(s234,-1) - pow(s12,-1)*pow(s24,3)*pow(s34,-1)*pow(s123,-1) - 
      pow(s12,-1)*pow(s24,3)*pow(s34,-1)*pow(s134,-1) + 2*pow(s12,-1)*
      pow(s24,3)*pow(s123,-1)*pow(s134,-1) + 3*pow(s12,-1)*pow(s24,3)*
      pow(s123,-1)*pow(s234,-1) - 3*pow(s12,-1)*pow(s24,3)*pow(s124,-1)
      *pow(s134,-1) + pow(s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s123,-1)*
      pow(s234,-1) + pow(s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s124,-1)*
      pow(s134,-1) + pow(s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s124,-1)*
      pow(s234,-1) - 2*pow(s12,-1)*s34 - 2*pow(s12,-1)*pow(s34,2)*pow(
      s123,-1) - pow(s12,-1)*pow(s34,2)*pow(s124,-1) + 9*s12*s13*pow(
      s14,-1)*pow(s23,-1)*s24*pow(s34,-1) + 17*s12*s13*pow(s14,-1)*pow(
      s23,-1);
    A +=  - s12*s13*pow(s14,-1)*s24*pow(s34,-1)*pow(s234,-1) + 18*s12
      *s13*pow(s14,-1)*pow(s34,-1) + s12*s13*pow(s14,-1)*pow(s234,-1)
       - 4*s12*s13*pow(s23,-2)*s34*pow(s234,-1) + 4*s12*s13*pow(s23,-2)
      *pow(s34,2)*pow(s234,-2) + 18*s12*s13*pow(s23,-1)*pow(s34,-1) + 8
      *s12*s13*pow(s23,-1)*s34*pow(s234,-2) - 12*s12*s13*pow(s23,-1)*
      pow(s234,-1) + 4*s12*s13*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 12
      *s12*s13*pow(s34,-1)*pow(s234,-1) + 8*s12*s13*pow(s234,-2) + 6*
      s12*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*pow(s34,-1) + 10*s12*pow(
      s13,2)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) + 6*s12*pow(s13,2)*
      pow(s14,-1)*pow(s34,-1)*pow(s234,-1) + 8*s12*pow(s14,-1)*pow(
      s23,-1)*s24 + 3*s12*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(
      s34,-1) + 4*s12*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s134,-1)
       - 3*s12*pow(s14,-1)*pow(s23,-1)*s34 + 6*s12*pow(s14,-1)*pow(
      s23,-1)*pow(s34,2)*pow(s134,-1) + 6*s12*pow(s14,-1)*pow(s23,-1)*
      pow(s34,2)*pow(s234,-1) - 5*s12*pow(s14,-1)*pow(s23,-1)*pow(
      s34,3)*pow(s134,-1)*pow(s234,-1);
    A +=  + 6*s12*pow(s14,-1)*s23*pow(s34,-1) + 2*s12*pow(s14,-1)*s23
      *s34*pow(s134,-2) - 5*s12*pow(s14,-1)*s23*pow(s134,-1) + 11*s12*
      pow(s14,-1)*s24*pow(s34,-1) + 2*s12*pow(s14,-1)*s24*s34*pow(
      s134,-2) + s12*pow(s14,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - 5*
      s12*pow(s14,-1)*s24*pow(s134,-1) - 2*s12*pow(s14,-1)*s24*pow(
      s234,-1) + s12*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s234,-1) - 
      s12*pow(s14,-1)*pow(s24,2)*pow(s134,-1)*pow(s234,-1) + 3*s12*pow(
      s14,-1)*s34*pow(s134,-1) + 5*s12*pow(s14,-1)*s34*pow(s234,-1) - 4
      *s12*pow(s14,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 3*s12*
      pow(s14,-1) + 4*s12*s14*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + s12
      *s14*pow(s23,-1)*s24*pow(s34,-1)*pow(s134,-1) + 8*s12*s14*pow(
      s23,-1)*pow(s34,-1) - 12*s12*s14*pow(s23,-1)*s34*pow(s134,-1)*
      pow(s234,-1) + 8*s12*s14*pow(s23,-1)*s34*pow(s234,-2) + 13*s12*
      s14*pow(s23,-1)*pow(s134,-1) - 10*s12*s14*pow(s23,-1)*pow(
      s234,-1) + 8*s12*s14*s23*pow(s34,-1)*pow(s134,-2) - 8*s12*s14*s24
      *pow(s34,-2)*pow(s134,-1);
    A +=  - 8*s12*s14*s24*pow(s34,-2)*pow(s234,-1) + 8*s12*s14*s24*
      pow(s34,-1)*pow(s134,-2) - 2*s12*s14*s24*pow(s34,-1)*pow(s134,-1)
      *pow(s234,-1) + 4*s12*s14*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 9
      *s12*s14*pow(s34,-1)*pow(s134,-1) - 6*s12*s14*pow(s34,-1)*pow(
      s234,-1) - 12*s12*s14*pow(s134,-1)*pow(s234,-1) + 8*s12*s14*pow(
      s234,-2) + 4*s12*pow(s14,2)*pow(s23,-1)*pow(s34,-1)*pow(s134,-1)
       - 4*s12*pow(s14,2)*pow(s23,-1)*pow(s134,-1)*pow(s234,-1) + 4*s12
      *pow(s14,2)*s23*pow(s34,-2)*pow(s134,-2) + 4*s12*pow(s14,2)*s24*
      pow(s34,-2)*pow(s134,-2) + 4*s12*pow(s14,2)*pow(s34,-2)*pow(
      s134,-1) - 8*s12*pow(s14,2)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1)
       + 4*s12*pow(s23,-2)*pow(s34,2)*pow(s234,-1) + 17*s12*pow(s23,-1)
      *s24*pow(s34,-1) + 2*s12*pow(s23,-1)*s24*pow(s134,-1) + 3*s12*
      pow(s23,-1)*pow(s24,2)*pow(s34,-1)*pow(s134,-1) + 14*s12*pow(
      s23,-1)*s34*pow(s134,-1) + 17*s12*pow(s23,-1)*s34*pow(s234,-1) - 
      12*s12*pow(s23,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) - 9*s12*
      pow(s23,-1);
    A +=  - 11*s12*s23*pow(s34,-1)*pow(s134,-1) + 6*s12*s23*pow(
      s134,-2) - 15*s12*s24*pow(s34,-1)*pow(s134,-1) + s12*s24*pow(
      s34,-1)*pow(s234,-1) + 6*s12*s24*pow(s134,-2) - 3*s12*s24*pow(
      s134,-1)*pow(s234,-1) + 4*s12*pow(s24,2)*pow(s34,-2)*pow(s234,-1)
       - 2*s12*pow(s24,2)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) - 8*s12
      *pow(s34,-1) - 12*s12*s34*pow(s134,-1)*pow(s234,-1) + 14*s12*pow(
      s134,-1) + 19*s12*pow(s234,-1) + 4*pow(s12,2)*s13*pow(s14,-1)*
      pow(s23,-1)*pow(s34,-1) + 10*pow(s12,2)*s13*pow(s14,-1)*pow(
      s23,-1)*pow(s234,-1) + 4*pow(s12,2)*s13*pow(s14,-1)*pow(s34,-1)*
      pow(s234,-1) + 4*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*s24*pow(
      s34,-1) + 7*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*s24*pow(s134,-1)
       - 7*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*s34*pow(s134,-1) - 7*pow(
      s12,2)*pow(s14,-1)*pow(s23,-1)*s34*pow(s234,-1) + 9*pow(s12,2)*
      pow(s14,-1)*pow(s23,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 12
      *pow(s12,2)*pow(s14,-1)*pow(s23,-1) + 2*pow(s12,2)*pow(s14,-1)*
      s24*pow(s34,-1)*pow(s234,-1);
    A +=  - 3*pow(s12,2)*pow(s14,-1)*s24*pow(s134,-1)*pow(s234,-1) + 
      8*pow(s12,2)*pow(s14,-1)*pow(s34,-1) + pow(s12,2)*pow(s14,-1)*s34
      *pow(s134,-2) + 6*pow(s12,2)*pow(s14,-1)*s34*pow(s134,-1)*pow(
      s234,-1) - 3*pow(s12,2)*pow(s14,-1)*pow(s134,-1) - 4*pow(s12,2)*
      pow(s14,-1)*pow(s234,-1) - 2*pow(s12,2)*s14*pow(s23,-1)*pow(
      s34,-1)*pow(s134,-1) + 6*pow(s12,2)*s14*pow(s23,-1)*pow(s134,-1)*
      pow(s234,-1) - 4*pow(s12,2)*s14*s24*pow(s34,-2)*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*s14*pow(s34,-1)*pow(s134,-2) + 2*pow(
      s12,2)*pow(s14,2)*pow(s34,-2)*pow(s134,-2) + 2*pow(s12,2)*pow(
      s23,-2)*pow(s34,2)*pow(s234,-2) + 4*pow(s12,2)*pow(s23,-1)*s24*
      pow(s34,-1)*pow(s134,-1) + 10*pow(s12,2)*pow(s23,-1)*pow(s34,-1)
       + 12*pow(s12,2)*pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) + 4*
      pow(s12,2)*pow(s23,-1)*s34*pow(s234,-2) - 6*pow(s12,2)*pow(
      s23,-1)*pow(s134,-1) - 12*pow(s12,2)*pow(s23,-1)*pow(s234,-1) - 4
      *pow(s12,2)*s24*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 2*pow(
      s12,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2);
    A +=  - 14*pow(s12,2)*pow(s34,-1)*pow(s134,-1) - 14*pow(s12,2)*
      pow(s34,-1)*pow(s234,-1) + 3*pow(s12,2)*pow(s134,-2) + 12*pow(
      s12,2)*pow(s134,-1)*pow(s234,-1) + 4*pow(s12,2)*pow(s234,-2) + 2*
      pow(s12,3)*pow(s14,-1)*pow(s23,-1)*pow(s34,-1) - 7*pow(s12,3)*
      pow(s14,-1)*pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) + 6*pow(
      s12,3)*pow(s14,-1)*pow(s23,-1)*pow(s134,-1) + 6*pow(s12,3)*pow(
      s14,-1)*pow(s23,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s14,-1)*pow(
      s34,-1)*pow(s234,-1) - 4*pow(s12,3)*pow(s14,-1)*pow(s134,-1)*pow(
      s234,-1) + 2*pow(s12,3)*pow(s23,-1)*pow(s34,-1)*pow(s134,-1) - 4*
      pow(s12,3)*pow(s23,-1)*pow(s134,-1)*pow(s234,-1) - 8*pow(s12,3)*
      pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 2*pow(s12,4)*pow(s14,-1)*
      pow(s23,-1)*pow(s134,-1)*pow(s234,-1) + 5*s13*pow(s14,-1)*pow(
      s23,-1)*s24*s34*pow(s124,-1) + 14*s13*pow(s14,-1)*pow(s23,-1)*s24
       + 6*s13*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,-1) + s13*
      pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s123,-1) + s13*pow(s14,-1)
      *pow(s23,-1)*pow(s24,2)*pow(s124,-1);
    A +=  + 4*s13*pow(s14,-1)*pow(s23,-1)*s34 - s13*pow(s14,-1)*pow(
      s23,-1)*pow(s34,2)*pow(s123,-1) - s13*pow(s14,-1)*pow(s23,-1)*
      pow(s34,2)*pow(s124,-1) + 6*s13*pow(s14,-1)*pow(s23,-1)*pow(
      s34,2)*pow(s234,-1) + 5*s13*pow(s14,-1)*pow(s23,-1)*pow(s34,3)*
      pow(s124,-1)*pow(s234,-1) - s13*pow(s14,-1)*s23*s24*pow(s34,-1)*
      pow(s124,-1) + 2*s13*pow(s14,-1)*s23*s24*pow(s124,-2) + 12*s13*
      pow(s14,-1)*s23*pow(s34,-1) + s13*pow(s14,-1)*pow(s23,2)*pow(
      s34,-1)*pow(s124,-1) + 15*s13*pow(s14,-1)*s24*pow(s34,-1) + 2*s13
      *pow(s14,-1)*s24*s34*pow(s124,-2) - 4*s13*pow(s14,-1)*s24*s34*
      pow(s124,-1)*pow(s234,-1) - 2*s13*pow(s14,-1)*s24*pow(s124,-1) - 
      4*s13*pow(s14,-1)*s24*pow(s234,-1) + s13*pow(s14,-1)*pow(s24,2)*
      pow(s34,-1)*pow(s124,-1) + s13*pow(s14,-1)*pow(s24,2)*pow(s34,-1)
      *pow(s234,-1) + 4*s13*pow(s14,-1)*pow(s24,2)*pow(s124,-1)*pow(
      s234,-1) - s13*pow(s14,-1)*pow(s24,3)*pow(s34,-1)*pow(s124,-1)*
      pow(s234,-1) - 2*s13*pow(s14,-1)*s34*pow(s123,-1) + 2*s13*pow(
      s14,-1)*s34*pow(s124,-1);
    A +=  + 5*s13*pow(s14,-1)*s34*pow(s234,-1) + 5*s13*pow(s14,-1)*
      pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 15*s13*pow(s14,-1) - 8*s13
      *s14*pow(s23,-2)*s34*pow(s123,-1) - 8*s13*s14*pow(s23,-2)*s34*
      pow(s234,-1) + 4*s13*s14*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + 2*
      s13*s14*pow(s23,-1)*s24*pow(s34,-1)*pow(s123,-1) + 14*s13*s14*
      pow(s23,-1)*pow(s34,-1) - 2*s13*s14*pow(s23,-1)*s34*pow(s123,-1)*
      pow(s234,-1) + 4*s13*s14*pow(s23,-1)*s34*pow(s124,-1)*pow(
      s234,-1) + 8*s13*s14*pow(s23,-1)*s34*pow(s234,-2) + 2*s13*s14*
      pow(s23,-1)*pow(s123,-1) - 10*s13*s14*pow(s23,-1)*pow(s234,-1) - 
      4*s13*s14*s24*pow(s34,-2)*pow(s234,-1) - 4*s13*s14*s24*pow(
      s34,-1)*pow(s123,-1)*pow(s234,-1) + 4*s13*s14*s24*pow(s34,-1)*
      pow(s124,-1)*pow(s234,-1) + 4*s13*s14*pow(s24,2)*pow(s34,-2)*pow(
      s234,-2) - 2*s13*s14*pow(s34,-1)*pow(s123,-1) - 4*s13*s14*pow(
      s34,-1)*pow(s124,-1) - 4*s13*s14*pow(s34,-1)*pow(s234,-1) - 10*
      s13*s14*pow(s123,-1)*pow(s234,-1) + 4*s13*s14*pow(s124,-1)*pow(
      s234,-1);
    A +=  + 8*s13*s14*pow(s234,-2) - 4*s13*pow(s14,2)*pow(s23,-2)*s34
      *pow(s123,-1)*pow(s234,-1) + 4*s13*pow(s14,2)*pow(s23,-1)*pow(
      s34,-1)*pow(s123,-1) - 4*s13*pow(s14,2)*pow(s23,-1)*pow(s123,-1)*
      pow(s234,-1) - 6*s13*pow(s14,2)*pow(s34,-1)*pow(s123,-1)*pow(
      s234,-1) - 4*s13*pow(s23,-2)*s24*s34*pow(s123,-1) - 8*s13*pow(
      s23,-2)*s34 - 4*s13*pow(s23,-2)*pow(s34,2)*pow(s123,-1) + 4*s13*
      pow(s23,-2)*pow(s34,2)*pow(s234,-1) + 22*s13*pow(s23,-1)*s24*pow(
      s34,-1) + 5*s13*pow(s23,-1)*s24*pow(s123,-1) + s13*pow(s23,-1)*
      s24*pow(s124,-1) - 3*s13*pow(s23,-1)*s34*pow(s123,-1) + 9*s13*
      pow(s23,-1)*s34*pow(s124,-1) + 5*s13*pow(s23,-1)*s34*pow(s234,-1)
       - 8*s13*pow(s23,-1)*pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 14*
      s13*pow(s23,-1) + s13*s23*pow(s34,-1)*pow(s124,-1) + 4*s13*s23*
      pow(s124,-2) + s13*s24*pow(s34,-1)*pow(s123,-1) + 5*s13*s24*pow(
      s34,-1)*pow(s124,-1) - 5*s13*s24*pow(s123,-1)*pow(s234,-1) + 10*
      s13*s24*pow(s124,-1)*pow(s234,-1) + 4*s13*pow(s24,2)*pow(s34,-2)*
      pow(s234,-1);
    A +=  - 2*s13*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 
      4*s13*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 12*s13*
      pow(s34,-1) - 3*s13*s34*pow(s123,-1)*pow(s234,-1) + 4*s13*s34*
      pow(s124,-2) - 8*s13*s34*pow(s124,-1)*pow(s234,-1) - 3*s13*pow(
      s123,-1) + 4*s13*pow(s124,-1) + 4*s13*pow(s234,-1) + 9*pow(s13,2)
      *pow(s14,-1)*pow(s23,-1)*s24*pow(s34,-1) - 2*pow(s13,2)*pow(
      s14,-1)*pow(s23,-1)*s24*pow(s123,-1) + 2*pow(s13,2)*pow(s14,-1)*
      pow(s23,-1)*s24*pow(s124,-1) - 2*pow(s13,2)*pow(s14,-1)*pow(
      s23,-1)*s34*pow(s123,-1) - 2*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*
      s34*pow(s124,-1) + 7*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*s34*pow(
      s234,-1) + 9*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*pow(s34,2)*pow(
      s124,-1)*pow(s234,-1) + 11*pow(s13,2)*pow(s14,-1)*pow(s23,-1) + 3
      *pow(s13,2)*pow(s14,-1)*s23*pow(s34,-1)*pow(s124,-1) - 3*pow(
      s13,2)*pow(s14,-1)*s24*pow(s34,-1)*pow(s124,-1) - 3*pow(s13,2)*
      pow(s14,-1)*s24*pow(s34,-1)*pow(s234,-1) + pow(s13,2)*pow(s14,-1)
      *s24*pow(s124,-2);
    A +=  - 6*pow(s13,2)*pow(s14,-1)*s24*pow(s124,-1)*pow(s234,-1) + 
      3*pow(s13,2)*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 18*pow(s13,2)*pow(s14,-1)*pow(s34,-1) + 9*pow(s13,2)*
      pow(s14,-1)*s34*pow(s124,-1)*pow(s234,-1) + 6*pow(s13,2)*pow(
      s14,-1)*pow(s124,-1) + 5*pow(s13,2)*pow(s14,-1)*pow(s234,-1) + 4*
      pow(s13,2)*s14*pow(s23,-2)*s24*pow(s123,-2) + 4*pow(s13,2)*s14*
      pow(s23,-2)*s34*pow(s123,-2) + 4*pow(s13,2)*s14*pow(s23,-2)*pow(
      s123,-1) + 2*pow(s13,2)*s14*pow(s23,-1)*pow(s34,-1)*pow(s123,-1)
       - 8*pow(s13,2)*s14*pow(s23,-1)*pow(s123,-1)*pow(s234,-1) - 4*
      pow(s13,2)*s14*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) + 2*pow(
      s13,2)*pow(s14,2)*pow(s23,-2)*pow(s123,-2) + 4*pow(s13,2)*pow(
      s23,-2)*s24*s34*pow(s123,-2) + 4*pow(s13,2)*pow(s23,-2)*s24*pow(
      s123,-1) + 2*pow(s13,2)*pow(s23,-2)*pow(s24,2)*pow(s123,-2) + 4*
      pow(s13,2)*pow(s23,-2)*s34*pow(s123,-1) - 4*pow(s13,2)*pow(
      s23,-2)*s34*pow(s234,-1) + 2*pow(s13,2)*pow(s23,-2)*pow(s34,2)*
      pow(s123,-2);
    A +=  + 2*pow(s13,2)*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + 2*pow(
      s13,2)*pow(s23,-2) + pow(s13,2)*pow(s23,-1)*s24*pow(s34,-1)*pow(
      s123,-1) + 14*pow(s13,2)*pow(s23,-1)*pow(s34,-1) - 6*pow(s13,2)*
      pow(s23,-1)*s34*pow(s124,-1)*pow(s234,-1) + 4*pow(s13,2)*pow(
      s23,-1)*s34*pow(s234,-2) - 7*pow(s13,2)*pow(s23,-1)*pow(s123,-1)
       + 4*pow(s13,2)*pow(s23,-1)*pow(s124,-1) - 8*pow(s13,2)*pow(
      s23,-1)*pow(s234,-1) + pow(s13,2)*s24*pow(s34,-1)*pow(s123,-1)*
      pow(s234,-1) + 6*pow(s13,2)*s24*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 2*pow(s13,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 2*
      pow(s13,2)*pow(s34,-1)*pow(s124,-1) - 2*pow(s13,2)*pow(s34,-1)*
      pow(s234,-1) + 3*pow(s13,2)*pow(s123,-1)*pow(s234,-1) + 2*pow(
      s13,2)*pow(s124,-2) - 2*pow(s13,2)*pow(s124,-1)*pow(s234,-1) + 4*
      pow(s13,2)*pow(s234,-2) + 4*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*
      pow(s34,-1) + 7*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*s34*pow(
      s124,-1)*pow(s234,-1) - pow(s13,3)*pow(s14,-1)*pow(s23,-1)*pow(
      s124,-1);
    A +=  + 6*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) - 3*
      pow(s13,3)*pow(s14,-1)*s24*pow(s34,-1)*pow(s124,-1)*pow(s234,-1)
       + 4*pow(s13,3)*pow(s14,-1)*pow(s34,-1)*pow(s124,-1) + 4*pow(
      s13,3)*pow(s14,-1)*pow(s34,-1)*pow(s234,-1) + 7*pow(s13,3)*pow(
      s14,-1)*pow(s124,-1)*pow(s234,-1) + 2*pow(s13,3)*pow(s23,-1)*pow(
      s34,-1)*pow(s123,-1) - 2*pow(s13,3)*pow(s34,-1)*pow(s123,-1)*pow(
      s234,-1) + 2*pow(s13,4)*pow(s14,-1)*pow(s23,-1)*pow(s124,-1)*pow(
      s234,-1) + 2*pow(s13,4)*pow(s14,-1)*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) - pow(s14,-1)*pow(s23,-1)*s24*s34 - 6*pow(s14,-1)*pow(
      s23,-1)*s24*pow(s34,2)*pow(s123,-1) + 4*pow(s14,-1)*pow(s23,-1)*
      s24*pow(s34,2)*pow(s124,-1) + 4*pow(s14,-1)*pow(s23,-1)*s24*pow(
      s34,2)*pow(s134,-1) + pow(s14,-1)*pow(s23,-1)*s24*pow(s34,3)*pow(
      s123,-1)*pow(s124,-1) + 3*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,3)*
      pow(s123,-1)*pow(s134,-1) - 6*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*
      s34*pow(s123,-1) + 3*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*s34*pow(
      s124,-1);
    A +=  + 3*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*s34*pow(s134,-1) + 3
      *pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,2)*pow(s123,-1)*pow(
      s124,-1) + 3*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,2)*pow(
      s123,-1)*pow(s134,-1) + pow(s14,-1)*pow(s23,-1)*pow(s24,2) + pow(
      s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s34,-1) + 3*pow(s14,-1)*pow(
      s23,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(s124,-1) + pow(s14,-1)*
      pow(s23,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(s134,-1) - 2*pow(
      s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s123,-1) + pow(s14,-1)*pow(
      s23,-1)*pow(s24,3)*pow(s124,-1) + pow(s14,-1)*pow(s23,-1)*pow(
      s24,3)*pow(s134,-1) + pow(s14,-1)*pow(s23,-1)*pow(s24,4)*pow(
      s123,-1)*pow(s124,-1) - 2*pow(s14,-1)*pow(s23,-1)*pow(s34,3)*pow(
      s123,-1) + pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(s123,-1)*pow(
      s134,-1) + pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(s124,-1)*pow(
      s234,-1) + pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(s134,-1)*pow(
      s234,-1) + 5*pow(s14,-1)*s23*s24*pow(s34,-1) + 3*pow(s14,-1)*s23*
      s24*s34*pow(s123,-1)*pow(s124,-1);
    A +=  + 2*pow(s14,-1)*s23*s24*s34*pow(s124,-2) + 2*pow(s14,-1)*
      s23*s24*s34*pow(s134,-2) - pow(s14,-1)*s23*s24*pow(s123,-1) + 3*
      pow(s14,-1)*s23*s24*pow(s124,-1) - 4*pow(s14,-1)*s23*s24*pow(
      s134,-1) + 3*pow(s14,-1)*s23*pow(s24,2)*pow(s123,-1)*pow(s124,-1)
       - pow(s14,-1)*s23*s34*pow(s123,-1) - 2*pow(s14,-1)*s23*s34*pow(
      s124,-1) + 3*pow(s14,-1)*s23*s34*pow(s134,-1) + pow(s14,-1)*s23*
      pow(s34,2)*pow(s123,-1)*pow(s134,-1) - 2*pow(s14,-1)*s23 - pow(
      s14,-1)*pow(s23,2)*s24*pow(s123,-1)*pow(s124,-1) + pow(s14,-1)*
      pow(s23,2)*s24*pow(s124,-2) + 2*pow(s14,-1)*pow(s23,2)*pow(
      s34,-1) + pow(s14,-1)*pow(s23,2)*s34*pow(s134,-2) - pow(s14,-1)*
      pow(s23,2)*pow(s124,-1) - 2*pow(s14,-1)*pow(s23,2)*pow(s134,-1)
       + 6*pow(s14,-1)*s24*s34*pow(s123,-1) - pow(s14,-1)*s24*s34*pow(
      s124,-1) + 3*pow(s14,-1)*s24*s34*pow(s134,-1) - pow(s14,-1)*s24*
      s34*pow(s234,-1) - 3*pow(s14,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(
      s124,-1) - 3*pow(s14,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s134,-1)
       + pow(s14,-1)*s24*pow(s34,2)*pow(s124,-2);
    A +=  - pow(s14,-1)*s24*pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 2*
      pow(s14,-1)*s24 + 5*pow(s14,-1)*pow(s24,2)*pow(s34,-1) - 6*pow(
      s14,-1)*pow(s24,2)*s34*pow(s123,-1)*pow(s124,-1) + pow(s14,-1)*
      pow(s24,2)*s34*pow(s124,-1)*pow(s234,-1) + pow(s14,-1)*pow(s24,2)
      *s34*pow(s134,-2) + 2*pow(s14,-1)*pow(s24,2)*pow(s123,-1) - pow(
      s14,-1)*pow(s24,2)*pow(s124,-1) - 2*pow(s14,-1)*pow(s24,2)*pow(
      s134,-1) + pow(s14,-1)*pow(s24,2)*pow(s234,-1) - 3*pow(s14,-1)*
      pow(s24,3)*pow(s123,-1)*pow(s124,-1) - pow(s14,-1)*pow(s24,3)*
      pow(s124,-1)*pow(s234,-1) + 2*pow(s14,-1)*s34 + 2*pow(s14,-1)*
      pow(s34,2)*pow(s123,-1) - pow(s14,-1)*pow(s34,2)*pow(s134,-1) - 2
      *pow(s14,-1)*pow(s34,3)*pow(s123,-1)*pow(s134,-1) + pow(s14,-1)*
      pow(s34,3)*pow(s124,-1)*pow(s234,-1) + pow(s14,-1)*pow(s34,3)*
      pow(s134,-1)*pow(s234,-1) + 4*s14*pow(s23,-2)*pow(s34,2)*pow(
      s234,-1) + 14*s14*pow(s23,-1)*s24*pow(s34,-1) + 9*s14*pow(s23,-1)
      *s24*s34*pow(s123,-1)*pow(s124,-1) + 18*s14*pow(s23,-1)*s24*s34*
      pow(s123,-1)*pow(s134,-1);
    A +=  - 21*s14*pow(s23,-1)*s24*pow(s123,-1) + 3*s14*pow(s23,-1)*
      s24*pow(s124,-1) + 12*s14*pow(s23,-1)*s24*pow(s134,-1) + 3*s14*
      pow(s23,-1)*pow(s24,2)*pow(s34,-1)*pow(s134,-1) + 6*s14*pow(
      s23,-1)*pow(s24,2)*pow(s123,-1)*pow(s124,-1) + 9*s14*pow(s23,-1)*
      pow(s24,2)*pow(s123,-1)*pow(s134,-1) + s14*pow(s23,-1)*pow(s24,3)
      *pow(s34,-1)*pow(s123,-1)*pow(s134,-1) - 18*s14*pow(s23,-1)*s34*
      pow(s123,-1) + s14*pow(s23,-1)*s34*pow(s124,-1) + 8*s14*pow(
      s23,-1)*s34*pow(s234,-1) + 3*s14*pow(s23,-1)*pow(s34,2)*pow(
      s123,-1)*pow(s124,-1) + 10*s14*pow(s23,-1)*pow(s34,2)*pow(
      s123,-1)*pow(s134,-1) - 2*s14*pow(s23,-1)*pow(s34,2)*pow(s123,-1)
      *pow(s234,-1) + 3*s14*pow(s23,-1)*pow(s34,2)*pow(s124,-1)*pow(
      s234,-1) + 6*s14*pow(s23,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1)
       - 8*s14*pow(s23,-1) - 4*s14*s23*s24*pow(s34,-2)*pow(s134,-1) - 3
      *s14*s23*s24*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 8*s14*s23*
      s24*pow(s34,-1)*pow(s134,-2) - s14*s23*pow(s34,-1)*pow(s123,-1)
       - 4*s14*s23*pow(s34,-1)*pow(s124,-1);
    A +=  + 11*s14*s23*pow(s34,-1)*pow(s134,-1) + 3*s14*s23*pow(
      s123,-1)*pow(s124,-1) + 3*s14*s23*pow(s123,-1)*pow(s134,-1) - 3*
      s14*pow(s23,2)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 4*s14*pow(
      s23,2)*pow(s34,-1)*pow(s134,-2) - 8*s14*s24*pow(s34,-2) + 2*s14*
      s24*pow(s34,-1)*pow(s123,-1) + 8*s14*s24*pow(s34,-1)*pow(s134,-1)
       - 3*s14*s24*pow(s34,-1)*pow(s234,-1) + 2*s14*s24*pow(s123,-2) - 
      9*s14*s24*pow(s123,-1)*pow(s124,-1) - 3*s14*s24*pow(s123,-1)*pow(
      s134,-1) + s14*s24*pow(s123,-1)*pow(s234,-1) + 3*s14*s24*pow(
      s124,-1)*pow(s134,-1) - 3*s14*s24*pow(s124,-1)*pow(s234,-1) - 4*
      s14*pow(s24,2)*pow(s34,-2)*pow(s134,-1) + 4*s14*pow(s24,2)*pow(
      s34,-2)*pow(s234,-1) - 4*s14*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*
      pow(s234,-1) - 2*s14*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(
      s134,-1) + s14*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1)
       + 4*s14*pow(s24,2)*pow(s34,-1)*pow(s134,-2) + 11*s14*pow(s34,-1)
       + 2*s14*s34*pow(s123,-2) - 12*s14*s34*pow(s123,-1)*pow(s134,-1)
       + 2*s14*s34*pow(s123,-1)*pow(s234,-1);
    A +=  - 3*s14*s34*pow(s124,-1)*pow(s134,-1) + 3*s14*s34*pow(
      s124,-1)*pow(s234,-1) + 3*s14*s34*pow(s134,-1)*pow(s234,-1) + 12*
      s14*pow(s123,-1) - 8*s14*pow(s134,-1) + 15*s14*pow(s234,-1) + 2*
      pow(s14,2)*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + pow(s14,2)*pow(
      s23,-1)*s24*pow(s34,-1)*pow(s123,-1) + 4*pow(s14,2)*pow(s23,-1)*
      s24*pow(s34,-1)*pow(s134,-1) + 5*pow(s14,2)*pow(s23,-1)*s24*pow(
      s123,-1)*pow(s124,-1) + 12*pow(s14,2)*pow(s23,-1)*s24*pow(
      s123,-1)*pow(s134,-1) + 3*pow(s14,2)*pow(s23,-1)*pow(s24,2)*pow(
      s34,-1)*pow(s123,-1)*pow(s134,-1) + 8*pow(s14,2)*pow(s23,-1)*pow(
      s34,-1) + 3*pow(s14,2)*pow(s23,-1)*s34*pow(s123,-1)*pow(s124,-1)
       + 11*pow(s14,2)*pow(s23,-1)*s34*pow(s123,-1)*pow(s134,-1) - 2*
      pow(s14,2)*pow(s23,-1)*s34*pow(s124,-1)*pow(s234,-1) + 5*pow(
      s14,2)*pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) + 4*pow(s14,2)*
      pow(s23,-1)*s34*pow(s234,-2) - 21*pow(s14,2)*pow(s23,-1)*pow(
      s123,-1) + 2*pow(s14,2)*pow(s23,-1)*pow(s124,-1) - 16*pow(s14,2)*
      pow(s23,-1)*pow(s234,-1);
    A +=  + 4*pow(s14,2)*s23*s24*pow(s34,-2)*pow(s134,-2) + 4*pow(
      s14,2)*s23*pow(s34,-2)*pow(s134,-1) + pow(s14,2)*s23*pow(s34,-1)*
      pow(s123,-1)*pow(s134,-1) + 3*pow(s14,2)*s23*pow(s34,-1)*pow(
      s124,-1)*pow(s134,-1) + 2*pow(s14,2)*pow(s23,2)*pow(s34,-2)*pow(
      s134,-2) + 4*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-1) - 4*pow(
      s14,2)*s24*pow(s34,-2)*pow(s234,-1) - 6*pow(s14,2)*s24*pow(
      s34,-1)*pow(s123,-1)*pow(s234,-1) - pow(s14,2)*s24*pow(s34,-1)*
      pow(s124,-1)*pow(s134,-1) + 2*pow(s14,2)*s24*pow(s34,-1)*pow(
      s124,-1)*pow(s234,-1) + 2*pow(s14,2)*pow(s24,2)*pow(s34,-2)*pow(
      s134,-2) + 2*pow(s14,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 2*
      pow(s14,2)*pow(s34,-2) + 3*pow(s14,2)*pow(s34,-1)*pow(s123,-1) - 
      4*pow(s14,2)*pow(s34,-1)*pow(s134,-1) - 6*pow(s14,2)*pow(s34,-1)*
      pow(s234,-1) + pow(s14,2)*pow(s123,-2) - 3*pow(s14,2)*pow(
      s123,-1)*pow(s124,-1) - 8*pow(s14,2)*pow(s123,-1)*pow(s134,-1) + 
      6*pow(s14,2)*pow(s123,-1)*pow(s234,-1) - 3*pow(s14,2)*pow(
      s124,-1)*pow(s134,-1);
    A +=  - 2*pow(s14,2)*pow(s124,-1)*pow(s234,-1) + 2*pow(s14,2)*
      pow(s134,-1)*pow(s234,-1) + 4*pow(s14,2)*pow(s234,-2) + 3*pow(
      s14,3)*pow(s23,-1)*s24*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*
      pow(s14,3)*pow(s23,-1)*pow(s123,-1)*pow(s124,-1) + 7*pow(s14,3)*
      pow(s23,-1)*pow(s123,-1)*pow(s134,-1) - 8*pow(s14,3)*pow(s23,-1)*
      pow(s123,-1)*pow(s234,-1) + 2*pow(s14,3)*pow(s23,-1)*pow(s134,-1)
      *pow(s234,-1) - 2*pow(s14,3)*pow(s34,-1)*pow(s123,-1)*pow(
      s134,-1) - 4*pow(s14,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 2
      *pow(s14,3)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(s14,4)*
      pow(s23,-1)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*pow(s23,-2)
      *pow(s34,2) - 19*pow(s23,-1)*s24*s34*pow(s123,-1) + 5*pow(s23,-1)
      *s24*s34*pow(s124,-1) + 12*pow(s23,-1)*s24*s34*pow(s134,-1) + 6*
      pow(s23,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s124,-1) + 12*pow(
      s23,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s134,-1) + pow(s23,-1)*
      s24 + 9*pow(s23,-1)*pow(s24,2)*pow(s34,-1) + 9*pow(s23,-1)*pow(
      s24,2)*s34*pow(s123,-1)*pow(s124,-1);
    A +=  + 9*pow(s23,-1)*pow(s24,2)*s34*pow(s123,-1)*pow(s134,-1) - 
      11*pow(s23,-1)*pow(s24,2)*pow(s123,-1) + 3*pow(s23,-1)*pow(s24,2)
      *pow(s124,-1) + 6*pow(s23,-1)*pow(s24,2)*pow(s134,-1) + pow(
      s23,-1)*pow(s24,3)*pow(s34,-1)*pow(s134,-1) + 4*pow(s23,-1)*pow(
      s24,3)*pow(s123,-1)*pow(s124,-1) + 2*pow(s23,-1)*pow(s24,3)*pow(
      s123,-1)*pow(s134,-1) + 5*pow(s23,-1)*s34 - 10*pow(s23,-1)*pow(
      s34,2)*pow(s123,-1) + 6*pow(s23,-1)*pow(s34,2)*pow(s124,-1) - 5*
      pow(s23,-1)*pow(s34,2)*pow(s234,-1) + pow(s23,-1)*pow(s34,3)*pow(
      s123,-1)*pow(s124,-1) + 5*pow(s23,-1)*pow(s34,3)*pow(s123,-1)*
      pow(s134,-1) - 3*pow(s23,-1)*pow(s34,3)*pow(s124,-1)*pow(s234,-1)
       + 4*pow(s23,-1)*pow(s34,3)*pow(s134,-1)*pow(s234,-1) + 4*s23*s24
      *pow(s34,-1)*pow(s124,-1) - 8*s23*s24*pow(s34,-1)*pow(s134,-1) + 
      6*s23*s24*pow(s123,-1)*pow(s124,-1) - 9*s23*s24*pow(s124,-1)*pow(
      s134,-1) + 6*s23*s24*pow(s134,-2) + 3*s23*pow(s24,2)*pow(s34,-1)*
      pow(s124,-1)*pow(s134,-1) - 3*s23*pow(s34,-1) + 3*s23*s34*pow(
      s123,-1)*pow(s134,-1);
    A +=  + 4*s23*s34*pow(s124,-2) - 3*s23*pow(s123,-1) + 7*s23*pow(
      s124,-1) + 14*s23*pow(s134,-1) + 3*pow(s23,2)*s24*pow(s34,-1)*
      pow(s124,-1)*pow(s134,-1) + 3*pow(s23,2)*pow(s34,-1)*pow(s124,-1)
       - 3*pow(s23,2)*pow(s34,-1)*pow(s134,-1) - pow(s23,2)*pow(
      s123,-1)*pow(s124,-1) + 2*pow(s23,2)*pow(s124,-2) + 3*pow(s23,2)*
      pow(s134,-2) + pow(s23,3)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1)
       + 6*s24*pow(s34,-1) + 2*s24*s34*pow(s123,-2) - 9*s24*s34*pow(
      s123,-1)*pow(s124,-1) - 6*s24*s34*pow(s123,-1)*pow(s134,-1) + s24
      *s34*pow(s123,-1)*pow(s234,-1) + 3*s24*s34*pow(s124,-1)*pow(
      s134,-1) + 3*s24*s34*pow(s124,-1)*pow(s234,-1) + 11*s24*pow(
      s123,-1) - 4*s24*pow(s124,-1) + 7*s24*pow(s134,-1) + 3*s24*pow(
      s234,-1) + 2*pow(s24,2)*pow(s34,-2) + pow(s24,2)*pow(s34,-1)*pow(
      s123,-1) - 6*pow(s24,2)*pow(s34,-1)*pow(s134,-1) - pow(s24,2)*
      pow(s34,-1)*pow(s234,-1) + pow(s24,2)*pow(s123,-2) - 9*pow(s24,2)
      *pow(s123,-1)*pow(s124,-1) - pow(s24,2)*pow(s123,-1)*pow(s234,-1)
       - 3*pow(s24,2)*pow(s124,-1)*pow(s134,-1);
    A +=  - 3*pow(s24,2)*pow(s124,-1)*pow(s234,-1) + 3*pow(s24,2)*
      pow(s134,-2) - pow(s24,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1)
       + pow(s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(
      s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 9*s34*pow(s123,-1)
       + 7*s34*pow(s124,-1) - 5*s34*pow(s134,-1) - 3*s34*pow(s234,-1)
       + pow(s34,2)*pow(s123,-2) - 8*pow(s34,2)*pow(s123,-1)*pow(
      s134,-1) + pow(s34,2)*pow(s123,-1)*pow(s234,-1) + 2*pow(s34,2)*
      pow(s124,-2) - pow(s34,2)*pow(s124,-1)*pow(s134,-1) - 3*pow(
      s34,2)*pow(s124,-1)*pow(s234,-1) + 3*pow(s34,2)*pow(s134,-1)*pow(
      s234,-1);

    //Normalize
    A /= pow2(s1234);

    // Return
    return A;

  }

  //***********************************************************************
  // E40a (q_a,qbar'_b,q'_c,g_d)
  // GGG notation (a = 1, b = 4, c = 3, d = 5)
  // From GGGH, JHEP 0711:058,2007
  double Matching::E40a(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants for E30(5,4,3) = E30(a,b,r)
    double sar   = m2(pn[d],pn[c]);
    double srb   = m2(pn[c],pn[b]);
    double sab   = m2(pn[d],pn[b]);
    double sarb  = m2(pn[b],pn[c],pn[d]);
    double Eterm = E30(sar,srb,sab,sarb);

    // Compute clustering and associated invariants for Q30(1,hat(34),hat(54))
    Vec4 p1(pn[a]);
    Vec4 p4(pn[b]);
    Vec4 p3(pn[c]);
    Vec4 p5(pn[d]);
    Vec4 pClu34;
    Vec4 pClu54;
    vinCluPtr->ks3to2(p3,p4,p5,pClu34,pClu54);
    sar          = m2(p1,pClu34);
    srb          = m2(pClu34,pClu54);
    sab          = m2(p1,pClu54);
    sarb         = m2(p1,pClu34,pClu54);
    double Qterm = Q30(sar,srb,sarb);

    // E40a
    double A = B40(pn,a,b,c,d) + Eterm*Qterm;
    
    // Optional: add finite correction to ensure positivity
    // A += 0.0;

    if (A < 0.0) {
      cout << " B4  = "<<B40(pn,a,b,c,d)
	   <<"   Eterm = "<<Eterm<<"   Qterm = "<<Qterm
	   <<"  s{ar,rb,ab,arb} = "<<sar<<" "<<srb<<" "<<sab<<" "<<sarb<<endl; 
    }

    // Return
    return A;

  }

  //***********************************************************************
  // E40b (q_a,qbar'_b,q'_c,g_d)
  // GGG notation (a = 1, b = 4, c = 3, d = 5)
  // From GGGH, JHEP 0711:058,2007
  double Matching::E40b(vector<Vec4>& pn, int a, int b, int c, int d) {

    // E40 
    double A = E40(pn,a,b,c,d); 

    // subtract E40a
    A -= E40a(pn,a,b,c,d); 

    return A;
    
  }

  //***********************************************************************
  // E40 (q_a,q'_b,qbar'_c,g_d)
  // GGG notation (a = 1, b = 4, c = 3, d = 5)
  // From GGG, JHEP 0509:056,2005
  double Matching::E40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate Vincia ordering to GGG notation
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s15=m2(pn[a],pn[d]);
    double s34=m2(pn[b],pn[c]);
    double s35=m2(pn[b],pn[d]);
    double s45=m2(pn[c],pn[d]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s345=m2(pn[b],pn[c],pn[d]);
    double s1345=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A=0.0;

    A +=  
      - (s13 * s14)/(s15 * s345)   
      + s14/( s15 * s34 * s45) * (
				  s13 * s35
				  + pow2(s13)
				  + pow2(s14)
				  )
      + s14/( s15 * s34 * s134) * (
				   2 * s14 * s35
				   + 2 * s14 * s45
				   - pow2(s35)
				   + pow2(s45)
				   )
      + s14/( s15 * s45 * s345) * (
				   s13 * s35
				   + pow2(s13)
				   + pow2(s14)
				   )
      + s14/( s15 * s45) * (
			    2 * s13
			    + 2 * s14
			    + s35
			    )
      - (4 * s14 * pow2(s15)*s35)/( pow2(s34) * s134 * s345 ) 
      + s14/( pow2(s34) * s134) * (
				   - 4 * s14 * s15
				   - 4 * s14 * s35
				   - 4 * s14 * s45
				   + 8 * s15 * s45
				   + 4 * pow2(s15)
				   + 4 * s35 * s45
				   + 4 * pow2(s45)
				   )
      + s14/( s34 * s45 * s134) * (
				    - 2 * s14 * s15
				    - s14 * s35
				    - 2 * pow2(s14)
				    - s15 * s35
				    - pow2(s15)
				    )
      + s14/( s34 * pow2(s134) ) * (
				    - 4 * s15 * s35
				    - 4 * s15 * s45
				    - 2 * pow2(s15)
				    - 4 * s35 * s45
				    - 2 * pow2(s35)
				    - 2 * pow2(s45)
				    )
      - s14/s345 
      + pow2(s14)/( pow2(s34) * pow2(s134) )  * (
						 - 4 * s15 * s35
						 - 4 * s15 * s45
						 - 2 * pow2(s15)
						 - 4 * s35 * s45
						 - 2 * pow2(s35)
						 - 2 * pow2(s45)
						 )
      + 1.0/(s15 * s34 * s345)  * (
				   s13 * pow2(s14)
				   + pow2(s13) * s14
				   + pow2(s13) * s35
				   + pow(s13,3)
				   - pow2(s14) * s35
				   + pow(s14,3)
				   )
      + 1.0/(s15 * s34) * (
			   s13 * s14
			   + 3 * s13 * s35
			   + s13 * s45
			   + 2 * pow2(s13)
			   - s14 * s35
			   + s14 * s45
			   + 4 * pow2(s14)
			   + s35 * s45
			   + pow2(s35)
			   )
      + 1.0/(s15 * s134) * (
			    - s13 * s45
			    + s14 * s35
			    + 2 * s14 * s45
			    - s35 * s45
			    + pow2(s45)
			    )
      + 1.0/s15 * (
		   s13
		   + 2 * s14
		   + s35
		   + s45
		   )
      + s15/( s34 * s134 * s345) * (
				    - 2 * s14 * s35
				    + 8 * pow2(s14)
				    - 2 * s15 * s35
				    + 2 * pow2(s15)
				    + 2 * pow2(s35)
				    )
      + s35/(pow2(s34) * s345 ) * (
				  4 * s13 * s14
				  + 8 * s13 * s15
				  - 4 * s13 * s35
				  + 4 * pow2(s13)
				  - 4 * s14 * s35
				  - 4 * s15 * s35
				  + 4 * pow2(s15)
				  )
      + pow2(s35)/(pow2(s34) * pow2(s345)) * (
					      - 4 * s13 * s14
					      - 4 * s13 * s15
					      - 2 * pow2(s13)
					      - 4 * s14 * s15
					      - 2 * pow2(s14)
					      - 2 * pow2(s15)
					      )
      + 1.0/(pow2(s34)) * (
			   - 4 * s13 * s15
			   + 4 * s13 * s35
			   - 4 * s13 * s45
			   - 2 * pow2(s13)
			   + 4 * s14 * s15
			   + 4 * s14 * s35
			   + 4 * s14 * s45
			   + 4 * s15 * s35
			   - 4 * s15 * s45
			   - 2 * pow2(s15)
			   - 2 * pow2(s45)
			   )
      + 1.0/(s34 * s134) * (
			    - s14 * s15
			    - 7 * s14 * s35
			    + s14 * s45
			    + 7 * pow2(s14)
			    + s15 * s35
			    + 7 * s15 * s45
			    + 6 * pow2(s15)
			    + 2 * s35 * s45
			    + pow2(s35)
			    + 3 * pow2(s45)
			    )
      + 1.0/(s34 * s345) * (
			    6 * s13 * s15
			    + 2 * s13 * s35
			    + 4 * pow2(s13)
			    - 4 * s14 * s35
			    + 6 * pow2(s14)
			    + 5 * pow2(s15)
			    )
      + 1.0/s34  * (
		    3 * s13
		    + 2 * s14
		    + 6 * s15
		    + 6 * s35
		    + s45
		    )
      + s35/( s45 * pow2(s345) ) * (
				    2 * s13 * s14
				    + 2 * s13 * s15
				    + pow2(s13)
				    + 2 * s14 * s15
				    + pow2(s14)
				    + pow2(s15)
				    )
      + 1.0/(s45 * s134 * s345) * (
				   - s13 * s14 * s15
				   + s13 * s14 * s35
				   - s13 * pow2(s14)
				   + s13 * s15 * s35
				   + 3 * s14 * pow2(s15)
				   + 3 * pow2(s14) * s15
				   + pow(s14,3)
				   + pow(s15,3)
				   )
      + 1.0/(s45 * s134) * (
			    - s13 * s15
			    + s14 * s15
			    - pow2(s14)
			    + pow2(s15)
			    )
      + 1.0/(s45 * s345 ) * (
			     - 2 * s13 * s15
			     + 2 * s13 * s35
			     - pow2(s13)
			     + 3 * s14 * s15
			     + s14 * s35
			     + 4 * pow2(s14)
			     + s15 * s35
			     )
      + 1.0/s45 * (
		   - 2 * s13
		   + 2 * s14
		   - s34 
		   )
      + 1.0/pow2(s134) * (
			  - 2 * s15 * s35
			  - 2 * s15 * s45
			  - pow2(s15)
			  - 2 * s35 * s45
			  - pow2(s35)
			  - pow2(s45)
			  )
      + 1.0/(s134 * s345) * (
			     s13 * s14
			     - 3 * s13 * s15
			     - pow2(s13)
			     + 4 * s14 * s15
			     - pow2(s14)
			     - 3 * pow2(s15)
			     )
      + 1.0/s134 * (
		    - 3 * s13
		    + 6 * s14 
		    - 3 * s35
		    + 2 * s45
		    )
      + 2 ;
    
    //Normalize
    A /= pow2(s1345);

    // Optional: add finite correction to ensure positivity
    // A += 0.0;
    
    if (A < 0.0) { 
      cout << " E40  = "<<A<<endl;
    }

    // Return
    return A;

  }
  
  //***********************************************************************
  // E40 (q_a,q'_b,qbar'_c,g_d)
  // GGG notation (a = 1, b = 3, c = 4, d = 5)
  // From GGG Form code, private communication
  double Matching::E40frm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate Vincia ordering to GGG notation
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s15=m2(pn[a],pn[d]);
    double s34=m2(pn[b],pn[c]);
    double s35=m2(pn[b],pn[d]);
    double s45=m2(pn[c],pn[d]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s345=m2(pn[b],pn[c],pn[d]);
    double s1345=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A=0.0;

    A +=  
      2 + 4*s13*s14*pow(s34,-2)*s35*pow(s345,-1) - 4*s13*s14*pow(
      s34,-2)*pow(s35,2)*pow(s345,-2) + s13*s14*pow(s34,-1)*pow(s15,-1)
      *pow(s45,-1)*s35 + s13*s14*pow(s34,-1)*pow(s15,-1) - s13*s14*pow(
      s134,-1)*s15*pow(s45,-1)*pow(s345,-1) + s13*s14*pow(s134,-1)*pow(
      s45,-1)*s35*pow(s345,-1) + s13*s14*pow(s134,-1)*pow(s345,-1) + 
      s13*s14*pow(s15,-1)*pow(s45,-1)*s35*pow(s345,-1) + 2*s13*s14*pow(
      s15,-1)*pow(s45,-1) - s13*s14*pow(s15,-1)*pow(s345,-1) + 2*s13*
      s14*pow(s45,-1)*s35*pow(s345,-2) + s13*pow(s14,2)*pow(s34,-1)*
      pow(s15,-1)*pow(s345,-1) - s13*pow(s14,2)*pow(s134,-1)*pow(
      s45,-1)*pow(s345,-1) + 8*s13*pow(s34,-2)*s15*s35*pow(s345,-1) - 4
      *s13*pow(s34,-2)*s15*pow(s35,2)*pow(s345,-2) - 4*s13*pow(s34,-2)*
      s15 - 4*s13*pow(s34,-2)*s45 + 4*s13*pow(s34,-2)*s35 - 4*s13*pow(
      s34,-2)*pow(s35,2)*pow(s345,-1) + s13*pow(s34,-1)*pow(s15,-1)*s45
       + 3*s13*pow(s34,-1)*pow(s15,-1)*s35 + 6*s13*pow(s34,-1)*s15*pow(
      s345,-1);
      A +=  + 2*s13*pow(s34,-1)*s35*pow(s345,-1) + 3*s13*pow(s34,-1) - 
      s13*pow(s134,-1)*pow(s15,-1)*s45 + s13*pow(s134,-1)*s15*pow(
      s45,-1)*s35*pow(s345,-1) - s13*pow(s134,-1)*s15*pow(s45,-1) - 3*
      s13*pow(s134,-1)*s15*pow(s345,-1) - 3*s13*pow(s134,-1) + s13*pow(
      s15,-1) + 2*s13*s15*pow(s45,-1)*s35*pow(s345,-2) - 2*s13*s15*pow(
      s45,-1)*pow(s345,-1) + 2*s13*pow(s45,-1)*s35*pow(s345,-1) - 2*s13
      *pow(s45,-1) + pow(s13,2)*s14*pow(s34,-1)*pow(s15,-1)*pow(s45,-1)
       + pow(s13,2)*s14*pow(s34,-1)*pow(s15,-1)*pow(s345,-1) + pow(
      s13,2)*s14*pow(s15,-1)*pow(s45,-1)*pow(s345,-1) + 4*pow(s13,2)*
      pow(s34,-2)*s35*pow(s345,-1) - 2*pow(s13,2)*pow(s34,-2)*pow(
      s35,2)*pow(s345,-2) - 2*pow(s13,2)*pow(s34,-2) + pow(s13,2)*pow(
      s34,-1)*pow(s15,-1)*s35*pow(s345,-1) + 2*pow(s13,2)*pow(s34,-1)*
      pow(s15,-1) + 4*pow(s13,2)*pow(s34,-1)*pow(s345,-1) - pow(s13,2)*
      pow(s134,-1)*pow(s345,-1) + pow(s13,2)*pow(s45,-1)*s35*pow(
      s345,-2) - pow(s13,2)*pow(s45,-1)*pow(s345,-1) + pow(s13,3)*pow(
      s34,-1)*pow(s15,-1)*pow(s345,-1);
      A +=  + 8*s14*pow(s34,-2)*pow(s134,-1)*s15*s45 - 4*s14*pow(
      s34,-2)*pow(s134,-1)*pow(s15,2)*s35*pow(s345,-1) + 4*s14*pow(
      s34,-2)*pow(s134,-1)*pow(s15,2) + 4*s14*pow(s34,-2)*pow(s134,-1)*
      s45*s35 + 4*s14*pow(s34,-2)*pow(s134,-1)*pow(s45,2) - 4*s14*pow(
      s34,-2)*s15*pow(s35,2)*pow(s345,-2) + 4*s14*pow(s34,-2)*s15 + 4*
      s14*pow(s34,-2)*s45 + 4*s14*pow(s34,-2)*s35 - 4*s14*pow(s34,-2)*
      pow(s35,2)*pow(s345,-1) - 4*s14*pow(s34,-1)*pow(s134,-2)*s15*s45
       - 4*s14*pow(s34,-1)*pow(s134,-2)*s15*s35 - 2*s14*pow(s34,-1)*
      pow(s134,-2)*pow(s15,2) - 4*s14*pow(s34,-1)*pow(s134,-2)*s45*s35
       - 2*s14*pow(s34,-1)*pow(s134,-2)*pow(s45,2) - 2*s14*pow(s34,-1)*
      pow(s134,-2)*pow(s35,2) + s14*pow(s34,-1)*pow(s134,-1)*pow(
      s15,-1)*pow(s45,2) - s14*pow(s34,-1)*pow(s134,-1)*pow(s15,-1)*
      pow(s35,2) - s14*pow(s34,-1)*pow(s134,-1)*s15*pow(s45,-1)*s35 - 2
      *s14*pow(s34,-1)*pow(s134,-1)*s15*s35*pow(s345,-1) - s14*pow(
      s34,-1)*pow(s134,-1)*s15 - s14*pow(s34,-1)*pow(s134,-1)*pow(
      s15,2)*pow(s45,-1);
      A +=  + s14*pow(s34,-1)*pow(s134,-1)*s45 - 7*s14*pow(s34,-1)*pow(
      s134,-1)*s35 + s14*pow(s34,-1)*pow(s15,-1)*s45 - s14*pow(s34,-1)*
      pow(s15,-1)*s35 - 4*s14*pow(s34,-1)*s35*pow(s345,-1) + 2*s14*pow(
      s34,-1) + 2*s14*pow(s134,-1)*pow(s15,-1)*s45 + s14*pow(s134,-1)*
      pow(s15,-1)*s35 + s14*pow(s134,-1)*s15*pow(s45,-1) + 4*s14*pow(
      s134,-1)*s15*pow(s345,-1) + 3*s14*pow(s134,-1)*pow(s15,2)*pow(
      s45,-1)*pow(s345,-1) + 6*s14*pow(s134,-1) + s14*pow(s15,-1)*pow(
      s45,-1)*s35 + 2*s14*pow(s15,-1) + 2*s14*s15*pow(s45,-1)*s35*pow(
      s345,-2) + 3*s14*s15*pow(s45,-1)*pow(s345,-1) + s14*pow(s45,-1)*
      s35*pow(s345,-1) + 2*s14*pow(s45,-1) - s14*pow(s345,-1) - 4*pow(
      s14,2)*pow(s34,-2)*pow(s134,-2)*s15*s45 - 4*pow(s14,2)*pow(
      s34,-2)*pow(s134,-2)*s15*s35 - 2*pow(s14,2)*pow(s34,-2)*pow(
      s134,-2)*pow(s15,2) - 4*pow(s14,2)*pow(s34,-2)*pow(s134,-2)*s45*
      s35 - 2*pow(s14,2)*pow(s34,-2)*pow(s134,-2)*pow(s45,2) - 2*pow(
      s14,2)*pow(s34,-2)*pow(s134,-2)*pow(s35,2) - 4*pow(s14,2)*pow(
      s34,-2)*pow(s134,-1)*s15;
      A +=  - 4*pow(s14,2)*pow(s34,-2)*pow(s134,-1)*s45 - 4*pow(s14,2)*
      pow(s34,-2)*pow(s134,-1)*s35 - 2*pow(s14,2)*pow(s34,-2)*pow(
      s35,2)*pow(s345,-2) + 2*pow(s14,2)*pow(s34,-1)*pow(s134,-1)*pow(
      s15,-1)*s45 + 2*pow(s14,2)*pow(s34,-1)*pow(s134,-1)*pow(s15,-1)*
      s35 - 2*pow(s14,2)*pow(s34,-1)*pow(s134,-1)*s15*pow(s45,-1) + 8*
      pow(s14,2)*pow(s34,-1)*pow(s134,-1)*s15*pow(s345,-1) - pow(s14,2)
      *pow(s34,-1)*pow(s134,-1)*pow(s45,-1)*s35 + 7*pow(s14,2)*pow(
      s34,-1)*pow(s134,-1) - pow(s14,2)*pow(s34,-1)*pow(s15,-1)*s35*
      pow(s345,-1) + 4*pow(s14,2)*pow(s34,-1)*pow(s15,-1) + 6*pow(
      s14,2)*pow(s34,-1)*pow(s345,-1) + 3*pow(s14,2)*pow(s134,-1)*s15*
      pow(s45,-1)*pow(s345,-1) - pow(s14,2)*pow(s134,-1)*pow(s45,-1) - 
      pow(s14,2)*pow(s134,-1)*pow(s345,-1) + 2*pow(s14,2)*pow(s15,-1)*
      pow(s45,-1) + pow(s14,2)*pow(s45,-1)*s35*pow(s345,-2) + 4*pow(
      s14,2)*pow(s45,-1)*pow(s345,-1) - 2*pow(s14,3)*pow(s34,-1)*pow(
      s134,-1)*pow(s45,-1) + pow(s14,3)*pow(s34,-1)*pow(s15,-1)*pow(
      s45,-1);
      A +=  + pow(s14,3)*pow(s34,-1)*pow(s15,-1)*pow(s345,-1) + pow(
      s14,3)*pow(s134,-1)*pow(s45,-1)*pow(s345,-1) + pow(s14,3)*pow(
      s15,-1)*pow(s45,-1)*pow(s345,-1) - 4*pow(s34,-2)*s15*s45 + 4*pow(
      s34,-2)*s15*s35 - 4*pow(s34,-2)*s15*pow(s35,2)*pow(s345,-1) + 4*
      pow(s34,-2)*pow(s15,2)*s35*pow(s345,-1) - 2*pow(s34,-2)*pow(
      s15,2)*pow(s35,2)*pow(s345,-2) - 2*pow(s34,-2)*pow(s15,2) - 2*
      pow(s34,-2)*pow(s45,2) + 7*pow(s34,-1)*pow(s134,-1)*s15*s45 + 
      pow(s34,-1)*pow(s134,-1)*s15*s35 + 2*pow(s34,-1)*pow(s134,-1)*s15
      *pow(s35,2)*pow(s345,-1) - 2*pow(s34,-1)*pow(s134,-1)*pow(s15,2)*
      s35*pow(s345,-1) + 6*pow(s34,-1)*pow(s134,-1)*pow(s15,2) + 2*pow(
      s34,-1)*pow(s134,-1)*pow(s15,3)*pow(s345,-1) + 2*pow(s34,-1)*pow(
      s134,-1)*s45*s35 + 3*pow(s34,-1)*pow(s134,-1)*pow(s45,2) + pow(
      s34,-1)*pow(s134,-1)*pow(s35,2) + pow(s34,-1)*pow(s15,-1)*s45*s35
       + pow(s34,-1)*pow(s15,-1)*pow(s35,2) + 6*pow(s34,-1)*s15 + 5*
      pow(s34,-1)*pow(s15,2)*pow(s345,-1) + pow(s34,-1)*s45 + 6*pow(
      s34,-1)*s35;
      A +=  - s34*pow(s45,-1) - 2*pow(s134,-2)*s15*s45 - 2*pow(s134,-2)
      *s15*s35 - pow(s134,-2)*pow(s15,2) - 2*pow(s134,-2)*s45*s35 - 
      pow(s134,-2)*pow(s45,2) - pow(s134,-2)*pow(s35,2) - pow(s134,-1)*
      pow(s15,-1)*s45*s35 + pow(s134,-1)*pow(s15,-1)*pow(s45,2) + pow(
      s134,-1)*pow(s15,2)*pow(s45,-1) - 3*pow(s134,-1)*pow(s15,2)*pow(
      s345,-1) + pow(s134,-1)*pow(s15,3)*pow(s45,-1)*pow(s345,-1) + 2*
      pow(s134,-1)*s45 - 3*pow(s134,-1)*s35 + pow(s15,-1)*s45 + pow(
      s15,-1)*s35 + s15*pow(s45,-1)*s35*pow(s345,-1) + pow(s15,2)*pow(
      s45,-1)*s35*pow(s345,-2);

    //Normalize
    A /= pow2(s1345);

    return A;
  }
  
  //***********************************************************************
  // ~E40 (q_a,q'_b,qbar'_c,g_d)
  // GGG notation (a = 1, b = 3, c = 4, d = 5)
  // From GGG Form code, private communication.
  double Matching::E40NLCfrm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate Vincia ordering to GGG notation
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s15=m2(pn[a],pn[d]);
    double s34=m2(pn[b],pn[c]);
    double s35=m2(pn[b],pn[d]);
    double s45=m2(pn[c],pn[d]);
    //    double s134=m2(pn[a],pn[b],pn[c]);
    double s345=m2(pn[b],pn[c],pn[d]);
    double s1345=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A=0.0;

    A =
      2*s13*s14*pow(s45,-1)*s35*pow(s345,-2) - 2*s13*s14*pow(s45,-1)*
      pow(s345,-1) + 2*s13*s14*s45*pow(s35,-1)*pow(s345,-2) - 2*s13*s14
      *pow(s35,-1)*pow(s345,-1) + 2*s13*s34*pow(s45,-1)*pow(s35,-1) + 2
      *s13*s15*pow(s45,-1)*pow(s35,-1) + 2*s13*s15*pow(s45,-1)*s35*pow(
      s345,-2) - 4*s13*s15*pow(s45,-1)*pow(s345,-1) + 2*s13*s15*s45*
      pow(s35,-1)*pow(s345,-2) - 2*s13*s15*pow(s35,-1)*pow(s345,-1) + 
      s13*pow(s45,-1)*s35*pow(s345,-1) + s13*s45*pow(s35,-1)*pow(
      s345,-1) + 2*pow(s13,2)*pow(s45,-1)*pow(s35,-1) + pow(s13,2)*pow(
      s45,-1)*s35*pow(s345,-2) - 2*pow(s13,2)*pow(s45,-1)*pow(s345,-1)
       + pow(s13,2)*s45*pow(s35,-1)*pow(s345,-2) + 2*s14*s34*pow(
      s45,-1)*pow(s35,-1) + 2*s14*s15*pow(s45,-1)*pow(s35,-1) + 2*s14*
      s15*pow(s45,-1)*s35*pow(s345,-2) - 2*s14*s15*pow(s45,-1)*pow(
      s345,-1) + 2*s14*s15*s45*pow(s35,-1)*pow(s345,-2) - 4*s14*s15*
      pow(s35,-1)*pow(s345,-1) + s14*pow(s45,-1)*s35*pow(s345,-1) + s14
      *s45*pow(s35,-1)*pow(s345,-1);
    A +=  + 2*pow(s14,2)*pow(s45,-1)*pow(s35,-1) + pow(s14,2)*pow(
      s45,-1)*s35*pow(s345,-2) + pow(s14,2)*s45*pow(s35,-1)*pow(
      s345,-2) - 2*pow(s14,2)*pow(s35,-1)*pow(s345,-1) + 2*s34*s15*pow(
      s45,-1)*pow(s35,-1) + s15*pow(s45,-1)*s35*pow(s345,-1) + s15*s45*
      pow(s35,-1)*pow(s345,-1) + 2*pow(s15,2)*pow(s45,-1)*pow(s35,-1)
      + pow(s15,2)*pow(s45,-1)*s35*pow(s345,-2) - 2*pow(s15,2)*pow(
      s45,-1)*pow(s345,-1) + pow(s15,2)*s45*pow(s35,-1)*pow(s345,-2) - 
      2*pow(s15,2)*pow(s35,-1)*pow(s345,-1);
    
    //Normalize
    A /= pow2(s1345);
    
    return A;
  }

  //***********************************************************************
  // F40 (g_a,g_b,g_c,g_d)
  // GGG notation (a = 1, b = 2, c = 3, d = 4)
  // From GGG form code, private communication
  double Matching::F40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[b]);
    double s13=m2(pn[a],pn[c]);
    double s14=m2(pn[a],pn[d]);
    double s23=m2(pn[b],pn[c]);
    double s24=m2(pn[b],pn[d]);
    double s34=m2(pn[c],pn[d]);
    double s123=m2(pn[a],pn[b],pn[c]);
    double s124=m2(pn[a],pn[b],pn[d]);
    double s234=m2(pn[b],pn[c],pn[d]);
    double s134=m2(pn[a],pn[c],pn[d]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A =
      42 - 4*pow(s12,-2)*s13*s14*s23*s24*pow(s34,-2) - 4*pow(s12,-2)*
      s13*s23*s24*pow(s124,-1) + 4*pow(s12,-2)*s13*s23*pow(s24,2)*pow(
      s124,-2) + 4*pow(s12,-2)*s13*pow(s24,2)*s34*pow(s124,-2) + 4*pow(
      s12,-2)*s13*pow(s24,2)*pow(s124,-1) + 2*pow(s12,-2)*pow(s13,2)*
      pow(s24,2)*pow(s34,-2) + 2*pow(s12,-2)*pow(s13,2)*pow(s24,2)*pow(
      s124,-2) - 4*pow(s12,-2)*s14*s23*s24*pow(s123,-1) + 4*pow(s12,-2)
      *s14*pow(s23,2)*s24*pow(s123,-2) + 4*pow(s12,-2)*s14*pow(s23,2)*
      s34*pow(s123,-2) + 4*pow(s12,-2)*s14*pow(s23,2)*pow(s123,-1) + 2*
      pow(s12,-2)*pow(s14,2)*pow(s23,2)*pow(s34,-2) + 2*pow(s12,-2)*
      pow(s14,2)*pow(s23,2)*pow(s123,-2) - 8*pow(s12,-2)*s23*s24*s34*
      pow(s123,-1) - 8*pow(s12,-2)*s23*s24*s34*pow(s124,-1) - 4*pow(
      s12,-2)*s23*s24*pow(s34,2)*pow(s123,-1)*pow(s124,-1) - 8*pow(
      s12,-2)*s23*s24 + 4*pow(s12,-2)*s23*pow(s24,2)*s34*pow(s124,-2)
       - 4*pow(s12,-2)*s23*pow(s24,2)*pow(s123,-1) + 4*pow(s12,-2)*s23*
      pow(s24,2)*pow(s124,-1);
      A +=  + 4*pow(s12,-2)*pow(s23,2)*s24*s34*pow(s123,-2) + 4*pow(
      s12,-2)*pow(s23,2)*s24*pow(s123,-1) - 4*pow(s12,-2)*pow(s23,2)*
      s24*pow(s124,-1) + 2*pow(s12,-2)*pow(s23,2)*pow(s24,2)*pow(
      s123,-2) + 2*pow(s12,-2)*pow(s23,2)*pow(s24,2)*pow(s124,-2) + 4*
      pow(s12,-2)*pow(s23,2)*s34*pow(s123,-1) + 2*pow(s12,-2)*pow(
      s23,2)*pow(s34,2)*pow(s123,-2) + 2*pow(s12,-2)*pow(s23,2) + 4*
      pow(s12,-2)*pow(s24,2)*s34*pow(s124,-1) + 2*pow(s12,-2)*pow(
      s24,2)*pow(s34,2)*pow(s124,-2) + 2*pow(s12,-2)*pow(s24,2) + 12*
      pow(s12,-1)*s13*pow(s14,-1)*pow(s23,-1)*s24*s34 + 12*pow(s12,-1)*
      s13*pow(s14,-1)*pow(s23,-1)*pow(s24,2) + 4*pow(s12,-1)*s13*pow(
      s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s34,-1) + 4*pow(s12,-1)*s13*
      pow(s14,-1)*pow(s23,-1)*pow(s34,2) + 12*pow(s12,-1)*s13*pow(
      s14,-1)*s23*s24*pow(s34,-1) + 16*pow(s12,-1)*s13*pow(s14,-1)*s23
       + 4*pow(s12,-1)*s13*pow(s14,-1)*pow(s23,2)*pow(s34,-1) + 28*pow(
      s12,-1)*s13*pow(s14,-1)*s24 + 12*pow(s12,-1)*s13*pow(s14,-1)*pow(
      s24,2)*pow(s34,-1);
      A +=  + 16*pow(s12,-1)*s13*pow(s14,-1)*s34 + 12*pow(s12,-1)*s13*
      s14*pow(s23,-1)*s24*pow(s34,-1) + 4*pow(s12,-1)*s13*s14*pow(
      s23,-1)*s34*pow(s234,-1) + 20*pow(s12,-1)*s13*s14*pow(s23,-1) + 
      14*pow(s12,-1)*s13*s14*pow(s34,-1) + 8*pow(s12,-1)*s13*s14*pow(
      s234,-1) + 4*pow(s12,-1)*s13*pow(s14,2)*pow(s23,-1)*pow(s34,-1)
       + 4*pow(s12,-1)*s13*pow(s14,2)*pow(s23,-1)*pow(s234,-1) + 10*
      pow(s12,-1)*s13*pow(s14,2)*pow(s34,-1)*pow(s234,-1) + 28*pow(
      s12,-1)*s13*pow(s23,-1)*s24 + 12*pow(s12,-1)*s13*pow(s23,-1)*pow(
      s24,2)*pow(s34,-1) + 20*pow(s12,-1)*s13*pow(s23,-1)*s34 + 4*pow(
      s12,-1)*s13*pow(s23,-1)*pow(s34,2)*pow(s234,-1) + 4*pow(s12,-1)*
      s13*pow(s23,-1)*pow(s34,3)*pow(s124,-1)*pow(s234,-1) + 14*pow(
      s12,-1)*s13*s23*pow(s34,-1) - 8*pow(s12,-1)*s13*s23*pow(s124,-1)
       + 10*pow(s12,-1)*s13*pow(s23,2)*pow(s34,-1)*pow(s124,-1) + 16*
      pow(s12,-1)*s13*s24*pow(s34,-1) - 8*pow(s12,-1)*s13*s24*s34*pow(
      s124,-1)*pow(s234,-1) + 4*pow(s12,-1)*s13*s24*pow(s124,-1) + 4*
      pow(s12,-1)*s13*s24*pow(s234,-1);
      A +=  + 8*pow(s12,-1)*s13*pow(s24,2)*pow(s34,-1)*pow(s124,-1) + 8
      *pow(s12,-1)*s13*pow(s24,2)*pow(s34,-1)*pow(s234,-1) - 4*pow(
      s12,-1)*s13*pow(s24,2)*pow(s124,-1)*pow(s234,-1) - 8*pow(s12,-1)*
      s13*pow(s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) - 10*pow(
      s12,-1)*s13*s34*pow(s124,-1) + 8*pow(s12,-1)*s13*s34*pow(s234,-1)
       + 4*pow(s12,-1)*s13*pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 14*
      pow(s12,-1)*s13 + 12*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*pow(
      s23,-1)*s24 + 6*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*
      pow(s24,2)*pow(s34,-1) + 6*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*
      pow(s23,-1)*s34 + 6*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*s23*pow(
      s34,-1) + 12*pow(s12,-1)*pow(s13,2)*pow(s14,-1)*s24*pow(s34,-1)
       + 16*pow(s12,-1)*pow(s13,2)*pow(s14,-1) + 6*pow(s12,-1)*pow(
      s13,2)*s14*pow(s23,-1)*pow(s34,-1) + 6*pow(s12,-1)*pow(s13,2)*s14
      *pow(s23,-1)*pow(s234,-1) + 10*pow(s12,-1)*pow(s13,2)*s14*pow(
      s34,-1)*pow(s234,-1) + 12*pow(s12,-1)*pow(s13,2)*pow(s23,-1)*s24*
      pow(s34,-1);
      A +=  + 6*pow(s12,-1)*pow(s13,2)*pow(s23,-1)*s34*pow(s234,-1) + 6
      *pow(s12,-1)*pow(s13,2)*pow(s23,-1)*pow(s34,2)*pow(s124,-1)*pow(
      s234,-1) + 18*pow(s12,-1)*pow(s13,2)*pow(s23,-1) + 10*pow(s12,-1)
      *pow(s13,2)*s23*pow(s34,-1)*pow(s124,-1) - 8*pow(s12,-1)*pow(
      s13,2)*s24*pow(s34,-1)*pow(s124,-1) - 8*pow(s12,-1)*pow(s13,2)*
      s24*pow(s34,-1)*pow(s234,-1) + 12*pow(s12,-1)*pow(s13,2)*pow(
      s24,2)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 18*pow(s12,-1)*
      pow(s13,2)*pow(s34,-1) + 6*pow(s12,-1)*pow(s13,2)*s34*pow(
      s124,-1)*pow(s234,-1) - 4*pow(s12,-1)*pow(s13,2)*pow(s124,-1) + 4
      *pow(s12,-1)*pow(s13,2)*pow(s234,-1) + 4*pow(s12,-1)*pow(s13,3)*
      pow(s14,-1)*pow(s23,-1)*s24*pow(s34,-1) + 4*pow(s12,-1)*pow(
      s13,3)*pow(s14,-1)*pow(s23,-1) + 4*pow(s12,-1)*pow(s13,3)*pow(
      s14,-1)*pow(s34,-1) + 4*pow(s12,-1)*pow(s13,3)*pow(s23,-1)*pow(
      s34,-1) + 4*pow(s12,-1)*pow(s13,3)*pow(s23,-1)*s34*pow(s124,-1)*
      pow(s234,-1) + 4*pow(s12,-1)*pow(s13,3)*pow(s23,-1)*pow(s234,-1)
       - 8*pow(s12,-1)*pow(s13,3)*s24*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1);
      A +=  + 6*pow(s12,-1)*pow(s13,3)*pow(s34,-1)*pow(s124,-1) + 6*
      pow(s12,-1)*pow(s13,3)*pow(s34,-1)*pow(s234,-1) + 4*pow(s12,-1)*
      pow(s13,3)*pow(s124,-1)*pow(s234,-1) + 2*pow(s12,-1)*pow(s13,4)*
      pow(s14,-1)*pow(s23,-1)*pow(s34,-1) + 2*pow(s12,-1)*pow(s13,4)*
      pow(s23,-1)*pow(s124,-1)*pow(s234,-1) + 2*pow(s12,-1)*pow(s13,4)*
      pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 4*pow(s12,-1)*pow(s14,-1)
      *pow(s23,-1)*s24*pow(s34,2) + 6*pow(s12,-1)*pow(s14,-1)*pow(
      s23,-1)*pow(s24,2)*s34 + 4*pow(s12,-1)*pow(s14,-1)*pow(s23,-1)*
      pow(s24,3) + 2*pow(s12,-1)*pow(s14,-1)*pow(s23,-1)*pow(s24,4)*
      pow(s34,-1) + 2*pow(s12,-1)*pow(s14,-1)*pow(s23,-1)*pow(s34,3) + 
      4*pow(s12,-1)*pow(s14,-1)*s23*s24*s34*pow(s134,-1) + 20*pow(
      s12,-1)*pow(s14,-1)*s23*s24 + 6*pow(s12,-1)*pow(s14,-1)*s23*pow(
      s24,2)*pow(s34,-1) + 6*pow(s12,-1)*pow(s14,-1)*s23*pow(s24,2)*
      pow(s134,-1) + 6*pow(s12,-1)*pow(s14,-1)*s23*s34 - 4*pow(s12,-1)*
      pow(s14,-1)*s23*pow(s34,2)*pow(s123,-1) + 4*pow(s12,-1)*pow(
      s14,-1)*s23*pow(s34,2)*pow(s134,-1);
      A +=  + 4*pow(s12,-1)*pow(s14,-1)*pow(s23,2)*s24*pow(s34,-1) + 4*
      pow(s12,-1)*pow(s14,-1)*pow(s23,2)*s24*pow(s134,-1) + 2*pow(
      s12,-1)*pow(s14,-1)*pow(s23,2)*s34*pow(s123,-1) - 2*pow(s12,-1)*
      pow(s14,-1)*pow(s23,2)*s34*pow(s134,-1) + 12*pow(s12,-1)*pow(
      s14,-1)*pow(s23,2) + 2*pow(s12,-1)*pow(s14,-1)*pow(s23,3)*pow(
      s34,-1) - 2*pow(s12,-1)*pow(s14,-1)*pow(s23,3)*pow(s123,-1) + 2*
      pow(s12,-1)*pow(s14,-1)*pow(s23,3)*pow(s134,-1) + 20*pow(s12,-1)*
      pow(s14,-1)*s24*s34 + 4*pow(s12,-1)*pow(s14,-1)*s24*pow(s34,2)*
      pow(s134,-1) + 4*pow(s12,-1)*pow(s14,-1)*s24*pow(s34,3)*pow(
      s123,-1)*pow(s134,-1) + 6*pow(s12,-1)*pow(s14,-1)*pow(s24,2)*s34*
      pow(s134,-1) + 6*pow(s12,-1)*pow(s14,-1)*pow(s24,2)*pow(s34,2)*
      pow(s123,-1)*pow(s134,-1) + 18*pow(s12,-1)*pow(s14,-1)*pow(s24,2)
       + 4*pow(s12,-1)*pow(s14,-1)*pow(s24,3)*pow(s34,-1) + 4*pow(
      s12,-1)*pow(s14,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(s134,-1) + 4*
      pow(s12,-1)*pow(s14,-1)*pow(s24,3)*pow(s134,-1) + 2*pow(s12,-1)*
      pow(s14,-1)*pow(s24,4)*pow(s123,-1)*pow(s134,-1);
      A +=  + 12*pow(s12,-1)*pow(s14,-1)*pow(s34,2) + 2*pow(s12,-1)*
      pow(s14,-1)*pow(s34,4)*pow(s123,-1)*pow(s134,-1) + 18*pow(s12,-1)
      *s14*pow(s23,-1)*s24 + 6*pow(s12,-1)*s14*pow(s23,-1)*pow(s24,2)*
      pow(s34,-1) + 8*pow(s12,-1)*s14*pow(s23,-1)*s34 + 4*pow(s12,-1)*
      s14*pow(s23,-1)*pow(s34,2)*pow(s234,-1) + 8*pow(s12,-1)*s14*s23*
      s24*pow(s34,-1)*pow(s134,-1) + 8*pow(s12,-1)*s14*s23*s24*pow(
      s123,-2) + 4*pow(s12,-1)*s14*s23*pow(s34,-1) + 8*pow(s12,-1)*s14*
      s23*s34*pow(s123,-2) + 14*pow(s12,-1)*s14*s23*pow(s123,-1) - 2*
      pow(s12,-1)*s14*pow(s23,2)*pow(s34,-1)*pow(s123,-1) + 4*pow(
      s12,-1)*s14*pow(s23,2)*pow(s34,-1)*pow(s134,-1) + 8*pow(s12,-1)*
      s14*s24*pow(s34,-1) + 24*pow(s12,-1)*s14*s24*s34*pow(s123,-1)*
      pow(s134,-1) + 12*pow(s12,-1)*s14*s24*s34*pow(s123,-1)*pow(
      s234,-1) - 28*pow(s12,-1)*s14*s24*pow(s123,-1) + 8*pow(s12,-1)*
      s14*s24*pow(s134,-1) + 4*pow(s12,-1)*s14*s24*pow(s234,-1) - 10*
      pow(s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s123,-1) + 10*pow(
      s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s134,-1);
      A +=  + 8*pow(s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s234,-1) + 
      24*pow(s12,-1)*s14*pow(s24,2)*pow(s123,-1)*pow(s134,-1) + 12*pow(
      s12,-1)*s14*pow(s24,2)*pow(s123,-1)*pow(s234,-1) + 8*pow(s12,-1)*
      s14*pow(s24,3)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 8*pow(
      s12,-1)*s14*pow(s24,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 24
      *pow(s12,-1)*s14*s34*pow(s123,-1) + 6*pow(s12,-1)*s14*s34*pow(
      s134,-1) + 8*pow(s12,-1)*s14*s34*pow(s234,-1) + 10*pow(s12,-1)*
      s14*pow(s34,2)*pow(s123,-1)*pow(s134,-1) + 4*pow(s12,-1)*s14*pow(
      s34,2)*pow(s123,-1)*pow(s234,-1) - 4*pow(s12,-1)*s14 + 4*pow(
      s12,-1)*pow(s14,2)*pow(s23,-1)*s24*pow(s34,-1) - 2*pow(s12,-1)*
      pow(s14,2)*pow(s23,-1)*s34*pow(s234,-1) + 10*pow(s12,-1)*pow(
      s14,2)*pow(s23,-1) + 2*pow(s12,-1)*pow(s14,2)*s23*pow(s34,-1)*
      pow(s123,-1) + 4*pow(s12,-1)*pow(s14,2)*s23*pow(s123,-2) - 10*
      pow(s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s123,-1) + 6*pow(
      s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s134,-1) + 8*pow(s12,-1)*
      pow(s14,2)*s24*pow(s34,-1)*pow(s234,-1);
      A +=  + 20*pow(s12,-1)*pow(s14,2)*s24*pow(s123,-1)*pow(s134,-1)
       + 12*pow(s12,-1)*pow(s14,2)*s24*pow(s123,-1)*pow(s234,-1) + 12*
      pow(s12,-1)*pow(s14,2)*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*pow(
      s134,-1) + 12*pow(s12,-1)*pow(s14,2)*pow(s24,2)*pow(s34,-1)*pow(
      s123,-1)*pow(s234,-1) + 2*pow(s12,-1)*pow(s14,2)*pow(s34,-1) + 10
      *pow(s12,-1)*pow(s14,2)*s34*pow(s123,-1)*pow(s134,-1) + 6*pow(
      s12,-1)*pow(s14,2)*s34*pow(s123,-1)*pow(s234,-1) - 16*pow(s12,-1)
      *pow(s14,2)*pow(s123,-1) + 4*pow(s12,-1)*pow(s14,2)*pow(s134,-1)
       + 4*pow(s12,-1)*pow(s14,2)*pow(s234,-1) + 2*pow(s12,-1)*pow(
      s14,3)*pow(s23,-1)*pow(s34,-1) + 2*pow(s12,-1)*pow(s14,3)*pow(
      s23,-1)*pow(s234,-1) + 8*pow(s12,-1)*pow(s14,3)*s24*pow(s34,-1)*
      pow(s123,-1)*pow(s134,-1) + 8*pow(s12,-1)*pow(s14,3)*s24*pow(
      s34,-1)*pow(s123,-1)*pow(s234,-1) - 4*pow(s12,-1)*pow(s14,3)*pow(
      s34,-1)*pow(s123,-1) + 2*pow(s12,-1)*pow(s14,3)*pow(s34,-1)*pow(
      s134,-1) + 6*pow(s12,-1)*pow(s14,3)*pow(s34,-1)*pow(s234,-1) + 6*
      pow(s12,-1)*pow(s14,3)*pow(s123,-1)*pow(s134,-1);
      A +=  + 4*pow(s12,-1)*pow(s14,3)*pow(s123,-1)*pow(s234,-1) + 2*
      pow(s12,-1)*pow(s14,4)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*
      pow(s12,-1)*pow(s14,4)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) + 14
      *pow(s12,-1)*pow(s23,-1)*s24*s34 + 4*pow(s12,-1)*pow(s23,-1)*s24*
      pow(s34,2)*pow(s124,-1) + 2*pow(s12,-1)*pow(s23,-1)*pow(s24,2)*
      s34*pow(s124,-1) + 14*pow(s12,-1)*pow(s23,-1)*pow(s24,2) + 4*pow(
      s12,-1)*pow(s23,-1)*pow(s24,3)*pow(s34,-1) + 2*pow(s12,-1)*pow(
      s23,-1)*pow(s24,3)*pow(s124,-1) + 8*pow(s12,-1)*pow(s23,-1)*pow(
      s34,2) + 2*pow(s12,-1)*pow(s23,-1)*pow(s34,4)*pow(s124,-1)*pow(
      s234,-1) + 8*pow(s12,-1)*s23*s24*pow(s34,-1) + 8*pow(s12,-1)*s23*
      s24*s34*pow(s123,-2) - 8*pow(s12,-1)*s23*s24*s34*pow(s123,-1)*
      pow(s124,-1) + 12*pow(s12,-1)*s23*s24*s34*pow(s124,-1)*pow(
      s134,-1) - 12*pow(s12,-1)*s23*s24*pow(s124,-1) + 20*pow(s12,-1)*
      s23*s24*pow(s134,-1) - 2*pow(s12,-1)*s23*pow(s24,2)*pow(s34,-1)*
      pow(s123,-1) + 8*pow(s12,-1)*s23*pow(s24,2)*pow(s34,-1)*pow(
      s124,-1);
      A +=  + 2*pow(s12,-1)*s23*pow(s24,2)*pow(s34,-1)*pow(s134,-1) + 4
      *pow(s12,-1)*s23*pow(s24,2)*pow(s123,-2) - 12*pow(s12,-1)*s23*
      pow(s24,2)*pow(s124,-1)*pow(s134,-1) + 8*pow(s12,-1)*s23*pow(
      s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 12*pow(s12,-1)*s23
      *s34*pow(s123,-1) - 2*pow(s12,-1)*s23*s34*pow(s124,-1) + 4*pow(
      s12,-1)*s23*pow(s34,2)*pow(s123,-2) - 4*pow(s12,-1)*s23*pow(
      s34,2)*pow(s124,-1)*pow(s134,-1) + 14*pow(s12,-1)*s23 - 2*pow(
      s12,-1)*pow(s23,2)*s24*pow(s34,-1)*pow(s123,-1) + 8*pow(s12,-1)*
      pow(s23,2)*s24*pow(s34,-1)*pow(s124,-1) - 2*pow(s12,-1)*pow(
      s23,2)*s24*pow(s34,-1)*pow(s134,-1) - 12*pow(s12,-1)*pow(s23,2)*
      s24*pow(s124,-1)*pow(s134,-1) + 12*pow(s12,-1)*pow(s23,2)*pow(
      s24,2)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(s12,-1)*pow(
      s23,2)*pow(s34,-1) - 8*pow(s12,-1)*pow(s23,2)*s34*pow(s123,-1)*
      pow(s124,-1) + 6*pow(s12,-1)*pow(s23,2)*s34*pow(s124,-1)*pow(
      s134,-1) - 6*pow(s12,-1)*pow(s23,2)*pow(s123,-1) - 12*pow(s12,-1)
      *pow(s23,2)*pow(s124,-1);
      A +=  + 8*pow(s12,-1)*pow(s23,2)*pow(s134,-1) + 8*pow(s12,-1)*
      pow(s23,3)*s24*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 6*pow(
      s12,-1)*pow(s23,3)*pow(s34,-1)*pow(s124,-1) - 2*pow(s12,-1)*pow(
      s23,3)*pow(s34,-1)*pow(s134,-1) - 4*pow(s12,-1)*pow(s23,3)*pow(
      s124,-1)*pow(s134,-1) + 2*pow(s12,-1)*pow(s23,4)*pow(s34,-1)*pow(
      s124,-1)*pow(s134,-1) - 26*pow(s12,-1)*s24*s34*pow(s123,-1) + 4*
      pow(s12,-1)*s24*s34*pow(s124,-1) + 6*pow(s12,-1)*s24*s34*pow(
      s134,-1) - 4*pow(s12,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s124,-1)
       + 16*pow(s12,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s134,-1) + 4*
      pow(s12,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s234,-1) - 4*pow(
      s12,-1)*s24*pow(s34,2)*pow(s124,-1)*pow(s134,-1) + 22*pow(s12,-1)
      *s24 + 12*pow(s12,-1)*pow(s24,2)*pow(s34,-1) - 8*pow(s12,-1)*pow(
      s24,2)*s34*pow(s123,-1)*pow(s124,-1) + 18*pow(s12,-1)*pow(s24,2)*
      s34*pow(s123,-1)*pow(s134,-1) + 6*pow(s12,-1)*pow(s24,2)*s34*pow(
      s123,-1)*pow(s234,-1) + 6*pow(s12,-1)*pow(s24,2)*s34*pow(s124,-1)
      *pow(s134,-1);
      A +=  + 4*pow(s12,-1)*pow(s24,2)*s34*pow(s124,-1)*pow(s234,-1) - 
      22*pow(s12,-1)*pow(s24,2)*pow(s123,-1) - 8*pow(s12,-1)*pow(s24,2)
      *pow(s124,-1) + 16*pow(s12,-1)*pow(s24,2)*pow(s134,-1) - 4*pow(
      s12,-1)*pow(s24,3)*pow(s34,-1)*pow(s123,-1) + 4*pow(s12,-1)*pow(
      s24,3)*pow(s34,-1)*pow(s134,-1) + 12*pow(s12,-1)*pow(s24,3)*pow(
      s123,-1)*pow(s134,-1) + 4*pow(s12,-1)*pow(s24,3)*pow(s123,-1)*
      pow(s234,-1) - 4*pow(s12,-1)*pow(s24,3)*pow(s124,-1)*pow(s134,-1)
       + 2*pow(s12,-1)*pow(s24,3)*pow(s124,-1)*pow(s234,-1) + 2*pow(
      s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*pow(
      s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) + 2*pow(
      s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(
      s12,-1)*pow(s24,4)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) - 8*pow(
      s12,-1)*s34 - 20*pow(s12,-1)*pow(s34,2)*pow(s123,-1) - 16*pow(
      s12,-1)*pow(s34,2)*pow(s124,-1) + 4*pow(s12,-1)*pow(s34,2)*pow(
      s134,-1) - 8*pow(s12,-1)*pow(s34,3)*pow(s123,-1)*pow(s124,-1) + 6
      *pow(s12,-1)*pow(s34,3)*pow(s123,-1)*pow(s134,-1);
      A +=  + 2*pow(s12,-1)*pow(s34,3)*pow(s123,-1)*pow(s234,-1) + 2*
      pow(s12,-1)*pow(s34,3)*pow(s124,-1)*pow(s134,-1) + 2*pow(s12,-1)*
      pow(s34,3)*pow(s124,-1)*pow(s234,-1) - 4*s12*s13*pow(s14,-2)*pow(
      s23,-2)*s24*s34 + 12*s12*s13*pow(s14,-1)*pow(s23,-1)*s24*pow(
      s34,-1) + 14*s12*s13*pow(s14,-1)*pow(s23,-1) - 4*s12*s13*pow(
      s14,-1)*s24*pow(s34,-1)*pow(s234,-1) + 24*s12*s13*pow(s14,-1)*
      pow(s34,-1) + 4*s12*s13*pow(s14,-1)*pow(s234,-1) - 4*s12*s13*pow(
      s23,-2)*s34*pow(s234,-1) + 4*s12*s13*pow(s23,-2)*pow(s34,2)*pow(
      s234,-2) + 18*s12*s13*pow(s23,-1)*pow(s34,-1) + 8*s12*s13*pow(
      s23,-1)*s34*pow(s234,-2) - 12*s12*s13*pow(s23,-1)*pow(s234,-1) + 
      4*s12*s13*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 12*s12*s13*pow(
      s34,-1)*pow(s234,-1) + 8*s12*s13*pow(s234,-2) + 6*s12*pow(s13,2)*
      pow(s14,-1)*pow(s23,-1)*pow(s34,-1) + 10*s12*pow(s13,2)*pow(
      s14,-1)*pow(s23,-1)*pow(s234,-1) + 6*s12*pow(s13,2)*pow(s14,-1)*
      pow(s34,-1)*pow(s234,-1) + 4*s12*pow(s14,-2)*s23*pow(s34,2)*pow(
      s134,-2);
      A +=  - 4*s12*pow(s14,-2)*s24*s34*pow(s134,-1) + 4*s12*pow(
      s14,-2)*s24*pow(s34,2)*pow(s134,-2) + 4*s12*pow(s14,-2)*pow(
      s34,2)*pow(s134,-1) + 14*s12*pow(s14,-1)*pow(s23,-1)*s24 + 6*s12*
      pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,-1) + 10*s12*pow(
      s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s134,-1) - 6*s12*pow(s14,-1)*
      pow(s23,-1)*s34 + 8*s12*pow(s14,-1)*pow(s23,-1)*pow(s34,2)*pow(
      s134,-1) + 8*s12*pow(s14,-1)*pow(s23,-1)*pow(s34,2)*pow(s234,-1)
       - 8*s12*pow(s14,-1)*pow(s23,-1)*pow(s34,3)*pow(s134,-1)*pow(
      s234,-1) + 12*s12*pow(s14,-1)*s23*pow(s34,-1) + 8*s12*pow(s14,-1)
      *s23*s34*pow(s134,-2) - 14*s12*pow(s14,-1)*s23*pow(s134,-1) + 14*
      s12*pow(s14,-1)*s24*pow(s34,-1) + 8*s12*pow(s14,-1)*s24*s34*pow(
      s134,-2) + 4*s12*pow(s14,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - 
      8*s12*pow(s14,-1)*s24*pow(s134,-1) - 4*s12*pow(s14,-1)*s24*pow(
      s234,-1) + 4*s12*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s234,-1)
       - 4*s12*pow(s14,-1)*pow(s24,2)*pow(s134,-1)*pow(s234,-1) + 12*
      s12*pow(s14,-1)*s34*pow(s134,-1);
      A +=  + 4*s12*pow(s14,-1)*s34*pow(s234,-1) - 4*s12*pow(s14,-1)*
      pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 4*s12*s14*pow(s23,-2)*pow(
      s34,2)*pow(s234,-2) + 4*s12*s14*pow(s23,-1)*s24*pow(s34,-1)*pow(
      s134,-1) + 8*s12*s14*pow(s23,-1)*pow(s34,-1) - 12*s12*s14*pow(
      s23,-1)*s34*pow(s134,-1)*pow(s234,-1) + 8*s12*s14*pow(s23,-1)*s34
      *pow(s234,-2) + 12*s12*s14*pow(s23,-1)*pow(s134,-1) - 10*s12*s14*
      pow(s23,-1)*pow(s234,-1) + 8*s12*s14*s23*pow(s34,-1)*pow(s134,-2)
       - 8*s12*s14*s24*pow(s34,-2)*pow(s134,-1) - 8*s12*s14*s24*pow(
      s34,-2)*pow(s234,-1) + 8*s12*s14*s24*pow(s34,-1)*pow(s134,-2) - 8
      *s12*s14*s24*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 4*s12*s14*
      pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 12*s12*s14*pow(s34,-1)*pow(
      s134,-1) - 6*s12*s14*pow(s34,-1)*pow(s234,-1) - 12*s12*s14*pow(
      s134,-1)*pow(s234,-1) + 8*s12*s14*pow(s234,-2) + 4*s12*pow(s14,2)
      *pow(s23,-1)*pow(s34,-1)*pow(s134,-1) - 4*s12*pow(s14,2)*pow(
      s23,-1)*pow(s134,-1)*pow(s234,-1) + 4*s12*pow(s14,2)*s23*pow(
      s34,-2)*pow(s134,-2);
      A +=  + 4*s12*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-2) + 4*s12*
      pow(s14,2)*pow(s34,-2)*pow(s134,-1) - 8*s12*pow(s14,2)*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) + 4*s12*pow(s23,-2)*pow(s34,2)*
      pow(s234,-1) + 20*s12*pow(s23,-1)*s24*pow(s34,-1) + 8*s12*pow(
      s23,-1)*s24*pow(s134,-1) + 6*s12*pow(s23,-1)*pow(s24,2)*pow(
      s34,-1)*pow(s134,-1) + 12*s12*pow(s23,-1)*s34*pow(s134,-1) + 20*
      s12*pow(s23,-1)*s34*pow(s234,-1) - 12*s12*pow(s23,-1)*pow(s34,2)*
      pow(s134,-1)*pow(s234,-1) - 8*s12*pow(s23,-1) - 14*s12*s23*pow(
      s34,-1)*pow(s134,-1) + 12*s12*s23*pow(s134,-2) - 12*s12*s24*pow(
      s34,-1)*pow(s134,-1) + 4*s12*s24*pow(s34,-1)*pow(s234,-1) + 12*
      s12*s24*pow(s134,-2) - 12*s12*s24*pow(s134,-1)*pow(s234,-1) + 4*
      s12*pow(s24,2)*pow(s34,-2)*pow(s234,-1) - 8*s12*pow(s24,2)*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) - 8*s12*pow(s34,-1) - 12*s12*
      s34*pow(s134,-1)*pow(s234,-1) + 28*s12*pow(s134,-1) + 24*s12*pow(
      s234,-1) + 4*pow(s12,2)*s13*pow(s14,-1)*pow(s23,-1)*pow(s34,-1)
       + 10*pow(s12,2)*s13*pow(s14,-1)*pow(s23,-1)*pow(s234,-1);
      A +=  + 4*pow(s12,2)*s13*pow(s14,-1)*pow(s34,-1)*pow(s234,-1) + 2
      *pow(s12,2)*pow(s14,-2)*pow(s23,-2)*pow(s34,2) + 2*pow(s12,2)*
      pow(s14,-2)*pow(s34,2)*pow(s134,-2) + 4*pow(s12,2)*pow(s14,-1)*
      pow(s23,-1)*s24*pow(s34,-1) + 10*pow(s12,2)*pow(s14,-1)*pow(
      s23,-1)*s24*pow(s134,-1) - 8*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*
      s34*pow(s134,-1) - 8*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*s34*pow(
      s234,-1) + 12*pow(s12,2)*pow(s14,-1)*pow(s23,-1)*pow(s34,2)*pow(
      s134,-1)*pow(s234,-1) + 10*pow(s12,2)*pow(s14,-1)*pow(s23,-1) + 2
      *pow(s12,2)*pow(s14,-1)*s24*pow(s34,-1)*pow(s234,-1) - 6*pow(
      s12,2)*pow(s14,-1)*s24*pow(s134,-1)*pow(s234,-1) + 8*pow(s12,2)*
      pow(s14,-1)*pow(s34,-1) + 4*pow(s12,2)*pow(s14,-1)*s34*pow(
      s134,-2) + 6*pow(s12,2)*pow(s14,-1)*s34*pow(s134,-1)*pow(s234,-1)
       - 6*pow(s12,2)*pow(s14,-1)*pow(s134,-1) - 2*pow(s12,2)*pow(
      s14,-1)*pow(s234,-1) - 2*pow(s12,2)*s14*pow(s23,-1)*pow(s34,-1)*
      pow(s134,-1) + 6*pow(s12,2)*s14*pow(s23,-1)*pow(s134,-1)*pow(
      s234,-1);
      A +=  - 4*pow(s12,2)*s14*s24*pow(s34,-2)*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*s14*pow(s34,-1)*pow(s134,-2) + 2*pow(
      s12,2)*pow(s14,2)*pow(s34,-2)*pow(s134,-2) + 2*pow(s12,2)*pow(
      s23,-2)*pow(s34,2)*pow(s234,-2) + 4*pow(s12,2)*pow(s23,-1)*s24*
      pow(s34,-1)*pow(s134,-1) + 10*pow(s12,2)*pow(s23,-1)*pow(s34,-1)
       + 12*pow(s12,2)*pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) + 4*
      pow(s12,2)*pow(s23,-1)*s34*pow(s234,-2) - 4*pow(s12,2)*pow(
      s23,-1)*pow(s134,-1) - 12*pow(s12,2)*pow(s23,-1)*pow(s234,-1) - 4
      *pow(s12,2)*s24*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 2*pow(
      s12,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 14*pow(s12,2)*pow(
      s34,-1)*pow(s134,-1) - 14*pow(s12,2)*pow(s34,-1)*pow(s234,-1) + 6
      *pow(s12,2)*pow(s134,-2) + 12*pow(s12,2)*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*pow(s234,-2) + 2*pow(s12,3)*pow(s14,-1)*
      pow(s23,-1)*pow(s34,-1) - 8*pow(s12,3)*pow(s14,-1)*pow(s23,-1)*
      s34*pow(s134,-1)*pow(s234,-1) + 6*pow(s12,3)*pow(s14,-1)*pow(
      s23,-1)*pow(s134,-1);
      A +=  + 6*pow(s12,3)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) + 2*
      pow(s12,3)*pow(s14,-1)*pow(s34,-1)*pow(s234,-1) - 4*pow(s12,3)*
      pow(s14,-1)*pow(s134,-1)*pow(s234,-1) + 2*pow(s12,3)*pow(s23,-1)*
      pow(s34,-1)*pow(s134,-1) - 4*pow(s12,3)*pow(s23,-1)*pow(s134,-1)*
      pow(s234,-1) - 8*pow(s12,3)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1)
       + 2*pow(s12,4)*pow(s14,-1)*pow(s23,-1)*pow(s134,-1)*pow(s234,-1)
       + 4*s13*pow(s14,-2)*s23*pow(s24,2)*pow(s124,-2) - 4*s13*pow(
      s14,-2)*s24*s34*pow(s124,-1) + 4*s13*pow(s14,-2)*pow(s24,2)*s34*
      pow(s124,-2) + 4*s13*pow(s14,-2)*pow(s24,2)*pow(s124,-1) + 8*s13*
      pow(s14,-1)*pow(s23,-1)*s24*s34*pow(s124,-1) + 26*s13*pow(s14,-1)
      *pow(s23,-1)*s24 + 12*s13*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(
      s34,-1) + 2*s13*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s123,-1)
       + 8*s13*pow(s14,-1)*pow(s23,-1)*s34 - 2*s13*pow(s14,-1)*pow(
      s23,-1)*pow(s34,2)*pow(s123,-1) + 2*s13*pow(s14,-1)*pow(s23,-1)*
      pow(s34,2)*pow(s124,-1) + 8*s13*pow(s14,-1)*pow(s23,-1)*pow(
      s34,2)*pow(s234,-1);
      A +=  + 8*s13*pow(s14,-1)*pow(s23,-1)*pow(s34,3)*pow(s124,-1)*
      pow(s234,-1) - 4*s13*pow(s14,-1)*s23*s24*pow(s34,-1)*pow(s124,-1)
       + 24*s13*pow(s14,-1)*s23*pow(s34,-1) - 10*s13*pow(s14,-1)*s23*
      pow(s124,-1) + 4*s13*pow(s14,-1)*pow(s23,2)*pow(s34,-1)*pow(
      s124,-1) + 24*s13*pow(s14,-1)*s24*pow(s34,-1) - 4*s13*pow(s14,-1)
      *s24*s34*pow(s124,-1)*pow(s234,-1) - 4*s13*pow(s14,-1)*s24*pow(
      s234,-1) + 4*s13*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s124,-1)
       + 4*s13*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s234,-1) + 4*s13*
      pow(s14,-1)*pow(s24,2)*pow(s124,-1)*pow(s234,-1) - 4*s13*pow(
      s14,-1)*pow(s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) - 2*s13*
      pow(s14,-1)*s34*pow(s123,-1) - 8*s13*pow(s14,-1)*s34*pow(s124,-1)
       + 4*s13*pow(s14,-1)*s34*pow(s234,-1) + 8*s13*pow(s14,-1)*pow(
      s34,2)*pow(s124,-1)*pow(s234,-1) + 18*s13*pow(s14,-1) - 8*s13*s14
      *pow(s23,-2)*s34*pow(s123,-1) - 8*s13*s14*pow(s23,-2)*s34*pow(
      s234,-1) + 4*s13*s14*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + 14*s13
      *s14*pow(s23,-1)*pow(s34,-1);
      A +=  - 8*s13*s14*pow(s23,-1)*s34*pow(s123,-1)*pow(s234,-1) + 4*
      s13*s14*pow(s23,-1)*s34*pow(s124,-1)*pow(s234,-1) + 8*s13*s14*
      pow(s23,-1)*s34*pow(s234,-2) + 4*s13*s14*pow(s23,-1)*pow(s123,-1)
       - 10*s13*s14*pow(s23,-1)*pow(s234,-1) - 4*s13*s14*s24*pow(
      s34,-2)*pow(s234,-1) - 4*s13*s14*s24*pow(s34,-1)*pow(s123,-1)*
      pow(s234,-1) + 4*s13*s14*s24*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 4*s13*s14*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 2*s13*
      s14*pow(s34,-1)*pow(s123,-1) - 4*s13*s14*pow(s34,-1)*pow(s124,-1)
       - 4*s13*s14*pow(s34,-1)*pow(s234,-1) - 16*s13*s14*pow(s123,-1)*
      pow(s234,-1) + 4*s13*s14*pow(s124,-1)*pow(s234,-1) + 8*s13*s14*
      pow(s234,-2) - 4*s13*pow(s14,2)*pow(s23,-2)*s34*pow(s123,-1)*pow(
      s234,-1) + 4*s13*pow(s14,2)*pow(s23,-1)*pow(s34,-1)*pow(s123,-1)
       - 4*s13*pow(s14,2)*pow(s23,-1)*pow(s123,-1)*pow(s234,-1) - 6*s13
      *pow(s14,2)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 4*s13*pow(
      s23,-2)*s24*s34*pow(s123,-1) - 8*s13*pow(s23,-2)*s34 - 4*s13*pow(
      s23,-2)*pow(s34,2)*pow(s123,-1);
      A +=  + 4*s13*pow(s23,-2)*pow(s34,2)*pow(s234,-1) + 28*s13*pow(
      s23,-1)*s24*pow(s34,-1) + 8*s13*pow(s23,-1)*s24*pow(s123,-1) - 6*
      s13*pow(s23,-1)*s34*pow(s123,-1) + 12*s13*pow(s23,-1)*s34*pow(
      s124,-1) + 4*s13*pow(s23,-1)*s34*pow(s234,-1) - 8*s13*pow(s23,-1)
      *pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 20*s13*pow(s23,-1) + 4*
      s13*s23*pow(s34,-1)*pow(s124,-1) + 4*s13*s23*pow(s124,-2) + 4*s13
      *s24*pow(s34,-1)*pow(s123,-1) + 4*s13*s24*pow(s34,-1)*pow(
      s124,-1) - 8*s13*s24*pow(s123,-1)*pow(s234,-1) + 16*s13*s24*pow(
      s124,-1)*pow(s234,-1) + 4*s13*pow(s24,2)*pow(s34,-2)*pow(s234,-1)
       - 4*s13*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 4*s13
      *pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 18*s13*pow(
      s34,-1) - 4*s13*s34*pow(s123,-1)*pow(s234,-1) + 4*s13*s34*pow(
      s124,-2) - 8*s13*s34*pow(s124,-1)*pow(s234,-1) + 4*s13*pow(
      s124,-1) + 4*s13*pow(s234,-1) + 2*pow(s13,2)*pow(s14,-2)*pow(
      s23,-2)*pow(s24,2) + 2*pow(s13,2)*pow(s14,-2)*pow(s24,2)*pow(
      s124,-2);
      A +=  + 12*pow(s13,2)*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,-1) - 2
      *pow(s13,2)*pow(s14,-1)*pow(s23,-1)*s24*pow(s123,-1) + 4*pow(
      s13,2)*pow(s14,-1)*pow(s23,-1)*s24*pow(s124,-1) - 2*pow(s13,2)*
      pow(s14,-1)*pow(s23,-1)*s34*pow(s123,-1) - 2*pow(s13,2)*pow(
      s14,-1)*pow(s23,-1)*s34*pow(s124,-1) + 8*pow(s13,2)*pow(s14,-1)*
      pow(s23,-1)*s34*pow(s234,-1) + 12*pow(s13,2)*pow(s14,-1)*pow(
      s23,-1)*pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 10*pow(s13,2)*pow(
      s14,-1)*pow(s23,-1) + 6*pow(s13,2)*pow(s14,-1)*s23*pow(s34,-1)*
      pow(s124,-1) - 6*pow(s13,2)*pow(s14,-1)*s24*pow(s34,-1)*pow(
      s124,-1) - 6*pow(s13,2)*pow(s14,-1)*s24*pow(s34,-1)*pow(s234,-1)
       - 6*pow(s13,2)*pow(s14,-1)*s24*pow(s124,-1)*pow(s234,-1) + 6*
      pow(s13,2)*pow(s14,-1)*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 24*pow(s13,2)*pow(s14,-1)*pow(s34,-1) + 12*pow(s13,2)*
      pow(s14,-1)*s34*pow(s124,-1)*pow(s234,-1) - 4*pow(s13,2)*pow(
      s14,-1)*pow(s124,-1) + 6*pow(s13,2)*pow(s14,-1)*pow(s234,-1) + 4*
      pow(s13,2)*s14*pow(s23,-2)*s24*pow(s123,-2);
      A +=  + 4*pow(s13,2)*s14*pow(s23,-2)*s34*pow(s123,-2) + 4*pow(
      s13,2)*s14*pow(s23,-2)*pow(s123,-1) + 2*pow(s13,2)*s14*pow(
      s23,-1)*pow(s34,-1)*pow(s123,-1) - 8*pow(s13,2)*s14*pow(s23,-1)*
      pow(s123,-1)*pow(s234,-1) - 4*pow(s13,2)*s14*pow(s34,-1)*pow(
      s123,-1)*pow(s234,-1) + 2*pow(s13,2)*pow(s14,2)*pow(s23,-2)*pow(
      s123,-2) + 4*pow(s13,2)*pow(s23,-2)*s24*s34*pow(s123,-2) + 4*pow(
      s13,2)*pow(s23,-2)*s24*pow(s123,-1) + 2*pow(s13,2)*pow(s23,-2)*
      pow(s24,2)*pow(s123,-2) + 4*pow(s13,2)*pow(s23,-2)*s34*pow(
      s123,-1) - 4*pow(s13,2)*pow(s23,-2)*s34*pow(s234,-1) + 2*pow(
      s13,2)*pow(s23,-2)*pow(s34,2)*pow(s123,-2) + 2*pow(s13,2)*pow(
      s23,-2)*pow(s34,2)*pow(s234,-2) + 2*pow(s13,2)*pow(s23,-2) + 14*
      pow(s13,2)*pow(s23,-1)*pow(s34,-1) - 6*pow(s13,2)*pow(s23,-1)*s34
      *pow(s124,-1)*pow(s234,-1) + 4*pow(s13,2)*pow(s23,-1)*s34*pow(
      s234,-2) - 6*pow(s13,2)*pow(s23,-1)*pow(s123,-1) + 4*pow(s13,2)*
      pow(s23,-1)*pow(s124,-1) - 8*pow(s13,2)*pow(s23,-1)*pow(s234,-1)
       + 2*pow(s13,2)*s24*pow(s34,-1)*pow(s123,-1)*pow(s234,-1);
      A +=  + 6*pow(s13,2)*s24*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 
      2*pow(s13,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 2*pow(s13,2)*
      pow(s34,-1)*pow(s124,-1) - 2*pow(s13,2)*pow(s34,-1)*pow(s234,-1)
       + 2*pow(s13,2)*pow(s123,-1)*pow(s234,-1) + 2*pow(s13,2)*pow(
      s124,-2) - 2*pow(s13,2)*pow(s124,-1)*pow(s234,-1) + 4*pow(s13,2)*
      pow(s234,-2) + 4*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*pow(s34,-1)
       + 8*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*s34*pow(s124,-1)*pow(
      s234,-1) - 2*pow(s13,3)*pow(s14,-1)*pow(s23,-1)*pow(s124,-1) + 6*
      pow(s13,3)*pow(s14,-1)*pow(s23,-1)*pow(s234,-1) - 4*pow(s13,3)*
      pow(s14,-1)*s24*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 4*pow(
      s13,3)*pow(s14,-1)*pow(s34,-1)*pow(s124,-1) + 4*pow(s13,3)*pow(
      s14,-1)*pow(s34,-1)*pow(s234,-1) + 8*pow(s13,3)*pow(s14,-1)*pow(
      s124,-1)*pow(s234,-1) + 2*pow(s13,3)*pow(s23,-1)*pow(s34,-1)*pow(
      s123,-1) - 2*pow(s13,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) + 2
      *pow(s13,4)*pow(s14,-1)*pow(s23,-1)*pow(s124,-1)*pow(s234,-1) + 2
      *pow(s13,4)*pow(s14,-1)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1);
      A +=  - 8*pow(s14,-2)*s23*s24*s34*pow(s124,-1) - 8*pow(s14,-2)*
      s23*s24*s34*pow(s134,-1) + 4*pow(s14,-2)*s23*s24*pow(s34,2)*pow(
      s134,-2) + 4*pow(s14,-2)*s23*pow(s24,2)*s34*pow(s124,-2) + 4*pow(
      s14,-2)*s23*pow(s24,2)*pow(s124,-1) + 4*pow(s14,-2)*s23*pow(
      s34,2)*pow(s134,-1) - 4*pow(s14,-2)*pow(s23,2)*s24*s34*pow(
      s124,-1)*pow(s134,-1) + 2*pow(s14,-2)*pow(s23,2)*pow(s24,2)*pow(
      s124,-2) + 2*pow(s14,-2)*pow(s23,2)*pow(s34,2)*pow(s134,-2) - 8*
      pow(s14,-2)*s24*s34 - 4*pow(s14,-2)*s24*pow(s34,2)*pow(s124,-1)
       + 4*pow(s14,-2)*s24*pow(s34,2)*pow(s134,-1) + 4*pow(s14,-2)*pow(
      s24,2)*s34*pow(s124,-1) - 4*pow(s14,-2)*pow(s24,2)*s34*pow(
      s134,-1) + 2*pow(s14,-2)*pow(s24,2)*pow(s34,2)*pow(s124,-2) + 2*
      pow(s14,-2)*pow(s24,2)*pow(s34,2)*pow(s134,-2) + 2*pow(s14,-2)*
      pow(s24,2) + 2*pow(s14,-2)*pow(s34,2) + 8*pow(s14,-1)*pow(s23,-1)
      *s24*s34 - 10*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,2)*pow(s123,-1)
       + 10*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,2)*pow(s124,-1) + 8*
      pow(s14,-1)*pow(s23,-1)*s24*pow(s34,2)*pow(s134,-1);
      A +=  + 8*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,3)*pow(s123,-1)*
      pow(s124,-1) + 8*pow(s14,-1)*pow(s23,-1)*s24*pow(s34,3)*pow(
      s123,-1)*pow(s134,-1) - 10*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*s34
      *pow(s123,-1) + 6*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*s34*pow(
      s124,-1) + 8*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*s34*pow(s134,-1)
       + 12*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,2)*pow(s123,-1)*
      pow(s124,-1) + 12*pow(s14,-1)*pow(s23,-1)*pow(s24,2)*pow(s34,2)*
      pow(s123,-1)*pow(s134,-1) + 10*pow(s14,-1)*pow(s23,-1)*pow(s24,2)
       + 4*pow(s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s34,-1) + 8*pow(
      s14,-1)*pow(s23,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(s124,-1) + 8*
      pow(s14,-1)*pow(s23,-1)*pow(s24,3)*s34*pow(s123,-1)*pow(s134,-1)
       - 4*pow(s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s123,-1) + 2*pow(
      s14,-1)*pow(s23,-1)*pow(s24,3)*pow(s124,-1) + 6*pow(s14,-1)*pow(
      s23,-1)*pow(s24,3)*pow(s134,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(
      s24,4)*pow(s123,-1)*pow(s124,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(
      s24,4)*pow(s123,-1)*pow(s134,-1);
      A +=  + 4*pow(s14,-1)*pow(s23,-1)*pow(s34,2) - 4*pow(s14,-1)*pow(
      s23,-1)*pow(s34,3)*pow(s123,-1) + 4*pow(s14,-1)*pow(s23,-1)*pow(
      s34,3)*pow(s124,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(
      s123,-1)*pow(s124,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(
      s123,-1)*pow(s134,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(
      s124,-1)*pow(s234,-1) + 2*pow(s14,-1)*pow(s23,-1)*pow(s34,4)*pow(
      s134,-1)*pow(s234,-1) + 14*pow(s14,-1)*s23*s24*pow(s34,-1) + 12*
      pow(s14,-1)*s23*s24*s34*pow(s123,-1)*pow(s124,-1) + 4*pow(s14,-1)
      *s23*s24*s34*pow(s123,-1)*pow(s134,-1) - 8*pow(s14,-1)*s23*s24*
      s34*pow(s124,-1)*pow(s134,-1) + 8*pow(s14,-1)*s23*s24*s34*pow(
      s134,-2) - 4*pow(s14,-1)*s23*s24*pow(s123,-1) + 6*pow(s14,-1)*s23
      *s24*pow(s124,-1) - 12*pow(s14,-1)*s23*s24*pow(s134,-1) + 4*pow(
      s14,-1)*s23*pow(s24,2)*pow(s34,-1)*pow(s124,-1) + 6*pow(s14,-1)*
      s23*pow(s24,2)*pow(s123,-1)*pow(s124,-1) - 8*pow(s14,-1)*s23*pow(
      s24,2)*pow(s124,-1)*pow(s134,-1) - 4*pow(s14,-1)*s23*s34*pow(
      s123,-1);
      A +=  - 2*pow(s14,-1)*s23*s34*pow(s124,-1) + 14*pow(s14,-1)*s23*
      s34*pow(s134,-1) + 6*pow(s14,-1)*s23*pow(s34,2)*pow(s123,-1)*pow(
      s124,-1) + 2*pow(s14,-1)*s23*pow(s34,2)*pow(s123,-1)*pow(s134,-1)
       - 8*pow(s14,-1)*s23*pow(s34,2)*pow(s124,-1)*pow(s134,-1) - 8*
      pow(s14,-1)*s23 + 2*pow(s14,-1)*pow(s23,2)*s24*pow(s34,-1)*pow(
      s124,-1) - 4*pow(s14,-1)*pow(s23,2)*s24*pow(s123,-1)*pow(s124,-1)
       - 4*pow(s14,-1)*pow(s23,2)*s24*pow(s124,-1)*pow(s134,-1) + 8*
      pow(s14,-1)*pow(s23,2)*pow(s34,-1) - 4*pow(s14,-1)*pow(s23,2)*s34
      *pow(s123,-1)*pow(s124,-1) - 2*pow(s14,-1)*pow(s23,2)*s34*pow(
      s123,-1)*pow(s134,-1) + 4*pow(s14,-1)*pow(s23,2)*s34*pow(s134,-2)
       - 16*pow(s14,-1)*pow(s23,2)*pow(s124,-1) - 14*pow(s14,-1)*pow(
      s23,2)*pow(s134,-1) + 2*pow(s14,-1)*pow(s23,3)*pow(s34,-1)*pow(
      s124,-1) + 2*pow(s14,-1)*pow(s23,3)*pow(s123,-1)*pow(s124,-1) - 8
      *pow(s14,-1)*pow(s23,3)*pow(s124,-1)*pow(s134,-1) + 16*pow(
      s14,-1)*s24*s34*pow(s123,-1) - 14*pow(s14,-1)*s24*s34*pow(
      s124,-1);
      A +=  + 4*pow(s14,-1)*s24*s34*pow(s134,-1) - 12*pow(s14,-1)*s24*
      pow(s34,2)*pow(s123,-1)*pow(s124,-1) - 8*pow(s14,-1)*s24*pow(
      s34,2)*pow(s123,-1)*pow(s134,-1) - 4*pow(s14,-1)*s24*pow(s34,2)*
      pow(s134,-1)*pow(s234,-1) + 20*pow(s14,-1)*s24 + 14*pow(s14,-1)*
      pow(s24,2)*pow(s34,-1) - 12*pow(s14,-1)*pow(s24,2)*s34*pow(
      s123,-1)*pow(s124,-1) - 6*pow(s14,-1)*pow(s24,2)*s34*pow(s123,-1)
      *pow(s134,-1) + 4*pow(s14,-1)*pow(s24,2)*s34*pow(s124,-1)*pow(
      s234,-1) + 4*pow(s14,-1)*pow(s24,2)*s34*pow(s134,-2) - 2*pow(
      s14,-1)*pow(s24,2)*s34*pow(s134,-1)*pow(s234,-1) + 6*pow(s14,-1)*
      pow(s24,2)*pow(s123,-1) - 8*pow(s14,-1)*pow(s24,2)*pow(s124,-1)
       - 6*pow(s14,-1)*pow(s24,2)*pow(s134,-1) - 4*pow(s14,-1)*pow(
      s24,3)*pow(s123,-1)*pow(s124,-1) + 2*pow(s14,-1)*pow(s24,3)*pow(
      s124,-1)*pow(s234,-1) - 2*pow(s14,-1)*pow(s24,3)*pow(s134,-1)*
      pow(s234,-1) + 2*pow(s14,-1)*pow(s24,4)*pow(s34,-1)*pow(s124,-1)*
      pow(s234,-1) + 14*pow(s14,-1)*s34 + 6*pow(s14,-1)*pow(s34,2)*pow(
      s123,-1);
      A +=  - 12*pow(s14,-1)*pow(s34,2)*pow(s124,-1) - 4*pow(s14,-1)*
      pow(s34,2)*pow(s134,-1) - 4*pow(s14,-1)*pow(s34,3)*pow(s123,-1)*
      pow(s124,-1) - 4*pow(s14,-1)*pow(s34,3)*pow(s123,-1)*pow(s134,-1)
       + 2*pow(s14,-1)*pow(s34,3)*pow(s124,-1)*pow(s234,-1) + 4*s14*
      pow(s23,-2)*pow(s34,2)*pow(s234,-1) + 20*s14*pow(s23,-1)*s24*pow(
      s34,-1) + 12*s14*pow(s23,-1)*s24*s34*pow(s123,-1)*pow(s124,-1) + 
      24*s14*pow(s23,-1)*s24*s34*pow(s123,-1)*pow(s134,-1) - 26*s14*
      pow(s23,-1)*s24*pow(s123,-1) + 2*s14*pow(s23,-1)*s24*pow(s124,-1)
       + 12*s14*pow(s23,-1)*s24*pow(s134,-1) + 6*s14*pow(s23,-1)*pow(
      s24,2)*pow(s34,-1)*pow(s134,-1) + 6*s14*pow(s23,-1)*pow(s24,2)*
      pow(s123,-1)*pow(s124,-1) + 18*s14*pow(s23,-1)*pow(s24,2)*pow(
      s123,-1)*pow(s134,-1) + 4*s14*pow(s23,-1)*pow(s24,3)*pow(s34,-1)*
      pow(s123,-1)*pow(s134,-1) - 16*s14*pow(s23,-1)*s34*pow(s123,-1)
       + 2*s14*pow(s23,-1)*s34*pow(s124,-1) + 12*s14*pow(s23,-1)*s34*
      pow(s234,-1) + 6*s14*pow(s23,-1)*pow(s34,2)*pow(s123,-1)*pow(
      s124,-1);
      A +=  + 10*s14*pow(s23,-1)*pow(s34,2)*pow(s123,-1)*pow(s134,-1)
       - 8*s14*pow(s23,-1)*pow(s34,2)*pow(s123,-1)*pow(s234,-1) + 2*s14
      *pow(s23,-1)*pow(s34,2)*pow(s124,-1)*pow(s234,-1) + 6*s14*pow(
      s23,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) - 8*s14*pow(s23,-1)
       - 4*s14*s23*s24*pow(s34,-2)*pow(s134,-1) - 4*s14*s23*s24*pow(
      s34,-1)*pow(s123,-1)*pow(s134,-1) - 4*s14*s23*s24*pow(s34,-1)*
      pow(s124,-1)*pow(s134,-1) + 8*s14*s23*s24*pow(s34,-1)*pow(
      s134,-2) - 4*s14*s23*pow(s34,-1)*pow(s123,-1) - 4*s14*s23*pow(
      s34,-1)*pow(s124,-1) + 14*s14*s23*pow(s34,-1)*pow(s134,-1) + 6*
      s14*s23*pow(s123,-1)*pow(s124,-1) + 6*s14*s23*pow(s123,-1)*pow(
      s134,-1) - 2*s14*pow(s23,2)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1)
       - 6*s14*pow(s23,2)*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 4*s14
      *pow(s23,2)*pow(s34,-1)*pow(s134,-2) - 8*s14*s24*pow(s34,-2) + 4*
      s14*s24*pow(s34,-1)*pow(s123,-1) + 8*s14*s24*pow(s34,-1)*pow(
      s134,-1) - 6*s14*s24*pow(s34,-1)*pow(s234,-1) + 8*s14*s24*pow(
      s123,-2);
      A +=  - 12*s14*s24*pow(s123,-1)*pow(s124,-1) - 24*s14*s24*pow(
      s123,-1)*pow(s134,-1) + 4*s14*s24*pow(s123,-1)*pow(s234,-1) + 4*
      s14*s24*pow(s124,-1)*pow(s134,-1) - 2*s14*s24*pow(s124,-1)*pow(
      s234,-1) - 4*s14*pow(s24,2)*pow(s34,-2)*pow(s134,-1) + 4*s14*pow(
      s24,2)*pow(s34,-2)*pow(s234,-1) - 6*s14*pow(s24,2)*pow(s34,-1)*
      pow(s123,-1)*pow(s134,-1) - 4*s14*pow(s24,2)*pow(s34,-1)*pow(
      s123,-1)*pow(s234,-1) - 4*s14*pow(s24,2)*pow(s34,-1)*pow(s124,-1)
      *pow(s134,-1) + 2*s14*pow(s24,2)*pow(s34,-1)*pow(s124,-1)*pow(
      s234,-1) + 4*s14*pow(s24,2)*pow(s34,-1)*pow(s134,-2) + 14*s14*
      pow(s34,-1) + 8*s14*s34*pow(s123,-2) - 24*s14*s34*pow(s123,-1)*
      pow(s134,-1) - 4*s14*s34*pow(s123,-1)*pow(s234,-1) - 6*s14*s34*
      pow(s124,-1)*pow(s134,-1) + 2*s14*s34*pow(s124,-1)*pow(s234,-1)
       + 28*s14*pow(s123,-1) - 12*s14*pow(s134,-1) + 22*s14*pow(
      s234,-1) + 2*pow(s14,2)*pow(s23,-2)*pow(s34,2)*pow(s234,-2) + 4*
      pow(s14,2)*pow(s23,-1)*s24*pow(s34,-1)*pow(s134,-1) + 4*pow(
      s14,2)*pow(s23,-1)*s24*pow(s123,-1)*pow(s124,-1);
      A +=  + 16*pow(s14,2)*pow(s23,-1)*s24*pow(s123,-1)*pow(s134,-1)
       + 6*pow(s14,2)*pow(s23,-1)*pow(s24,2)*pow(s34,-1)*pow(s123,-1)*
      pow(s134,-1) + 8*pow(s14,2)*pow(s23,-1)*pow(s34,-1) + 4*pow(
      s14,2)*pow(s23,-1)*s34*pow(s123,-1)*pow(s124,-1) + 10*pow(s14,2)*
      pow(s23,-1)*s34*pow(s123,-1)*pow(s134,-1) - 2*pow(s14,2)*pow(
      s23,-1)*s34*pow(s124,-1)*pow(s234,-1) + 4*pow(s14,2)*pow(s23,-1)*
      s34*pow(s134,-1)*pow(s234,-1) + 4*pow(s14,2)*pow(s23,-1)*s34*pow(
      s234,-2) - 20*pow(s14,2)*pow(s23,-1)*pow(s123,-1) + 2*pow(s14,2)*
      pow(s23,-1)*pow(s124,-1) - 16*pow(s14,2)*pow(s23,-1)*pow(s234,-1)
       + 4*pow(s14,2)*s23*s24*pow(s34,-2)*pow(s134,-2) + 4*pow(s14,2)*
      s23*pow(s34,-2)*pow(s134,-1) + 2*pow(s14,2)*s23*pow(s34,-1)*pow(
      s123,-1)*pow(s134,-1) + 4*pow(s14,2)*s23*pow(s34,-1)*pow(s124,-1)
      *pow(s134,-1) + 2*pow(s14,2)*pow(s23,2)*pow(s34,-2)*pow(s134,-2)
       + 4*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-1) - 4*pow(s14,2)*s24*
      pow(s34,-2)*pow(s234,-1) - 4*pow(s14,2)*s24*pow(s34,-1)*pow(
      s123,-1)*pow(s134,-1);
      A +=  - 6*pow(s14,2)*s24*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 
      2*pow(s14,2)*s24*pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(
      s14,2)*s24*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 2*pow(s14,2)*
      pow(s24,2)*pow(s34,-2)*pow(s134,-2) + 2*pow(s14,2)*pow(s24,2)*
      pow(s34,-2)*pow(s234,-2) + 2*pow(s14,2)*pow(s34,-2) + 6*pow(
      s14,2)*pow(s34,-1)*pow(s123,-1) - 4*pow(s14,2)*pow(s34,-1)*pow(
      s134,-1) - 6*pow(s14,2)*pow(s34,-1)*pow(s234,-1) + 4*pow(s14,2)*
      pow(s123,-2) - 4*pow(s14,2)*pow(s123,-1)*pow(s124,-1) - 16*pow(
      s14,2)*pow(s123,-1)*pow(s134,-1) + 6*pow(s14,2)*pow(s123,-1)*pow(
      s234,-1) - 4*pow(s14,2)*pow(s124,-1)*pow(s134,-1) - 2*pow(s14,2)*
      pow(s124,-1)*pow(s234,-1) + 4*pow(s14,2)*pow(s234,-2) + 4*pow(
      s14,3)*pow(s23,-1)*s24*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*
      pow(s14,3)*pow(s23,-1)*pow(s123,-1)*pow(s124,-1) + 6*pow(s14,3)*
      pow(s23,-1)*pow(s123,-1)*pow(s134,-1) - 8*pow(s14,3)*pow(s23,-1)*
      pow(s123,-1)*pow(s234,-1) + 2*pow(s14,3)*pow(s23,-1)*pow(s134,-1)
      *pow(s234,-1);
      A +=  - 4*pow(s14,3)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1) - 4*
      pow(s14,3)*pow(s34,-1)*pow(s123,-1)*pow(s234,-1) - 2*pow(s14,3)*
      pow(s34,-1)*pow(s124,-1)*pow(s134,-1) + 2*pow(s14,4)*pow(s23,-1)*
      pow(s34,-1)*pow(s123,-1)*pow(s134,-1) + 2*pow(s23,-2)*pow(s34,2)
       - 24*pow(s23,-1)*s24*s34*pow(s123,-1) + 4*pow(s23,-1)*s24*s34*
      pow(s124,-1) + 12*pow(s23,-1)*s24*s34*pow(s134,-1) + 12*pow(
      s23,-1)*s24*pow(s34,2)*pow(s123,-1)*pow(s124,-1) + 20*pow(s23,-1)
      *s24*pow(s34,2)*pow(s123,-1)*pow(s134,-1) + 10*pow(s23,-1)*s24 + 
      18*pow(s23,-1)*pow(s24,2)*pow(s34,-1) + 12*pow(s23,-1)*pow(s24,2)
      *s34*pow(s123,-1)*pow(s124,-1) + 24*pow(s23,-1)*pow(s24,2)*s34*
      pow(s123,-1)*pow(s134,-1) - 18*pow(s23,-1)*pow(s24,2)*pow(
      s123,-1) + 4*pow(s23,-1)*pow(s24,2)*pow(s124,-1) + 12*pow(s23,-1)
      *pow(s24,2)*pow(s134,-1) + 4*pow(s23,-1)*pow(s24,3)*pow(s34,-1)*
      pow(s134,-1) + 4*pow(s23,-1)*pow(s24,3)*pow(s123,-1)*pow(s124,-1)
       + 12*pow(s23,-1)*pow(s24,3)*pow(s123,-1)*pow(s134,-1) + 2*pow(
      s23,-1)*pow(s24,4)*pow(s34,-1)*pow(s123,-1)*pow(s134,-1);
      A +=  + 14*pow(s23,-1)*s34 - 16*pow(s23,-1)*pow(s34,2)*pow(
      s123,-1) + 10*pow(s23,-1)*pow(s34,2)*pow(s124,-1) - 8*pow(s23,-1)
      *pow(s34,2)*pow(s234,-1) + 4*pow(s23,-1)*pow(s34,3)*pow(s123,-1)*
      pow(s124,-1) + 6*pow(s23,-1)*pow(s34,3)*pow(s123,-1)*pow(s134,-1)
       - 4*pow(s23,-1)*pow(s34,3)*pow(s124,-1)*pow(s234,-1) + 4*pow(
      s23,-1)*pow(s34,3)*pow(s134,-1)*pow(s234,-1) + 4*s23*s24*pow(
      s34,-1)*pow(s124,-1) - 8*s23*s24*pow(s34,-1)*pow(s134,-1) + 12*
      s23*s24*pow(s123,-1)*pow(s124,-1) - 24*s23*s24*pow(s124,-1)*pow(
      s134,-1) + 12*s23*s24*pow(s134,-2) + 4*s23*pow(s24,2)*pow(s34,-1)
      *pow(s124,-1)*pow(s134,-1) + 6*s23*s34*pow(s123,-1)*pow(s134,-1)
       + 4*s23*s34*pow(s124,-2) - 12*s23*pow(s123,-1) + 10*s23*pow(
      s124,-1) + 32*s23*pow(s134,-1) + 6*pow(s23,2)*s24*pow(s34,-1)*
      pow(s124,-1)*pow(s134,-1) + 6*pow(s23,2)*pow(s34,-1)*pow(s124,-1)
       - 6*pow(s23,2)*pow(s34,-1)*pow(s134,-1) - 4*pow(s23,2)*pow(
      s123,-1)*pow(s124,-1) - 4*pow(s23,2)*pow(s123,-1)*pow(s134,-1) + 
      2*pow(s23,2)*pow(s124,-2);
      A +=  + 6*pow(s23,2)*pow(s134,-2) + 4*pow(s23,3)*pow(s34,-1)*pow(
      s124,-1)*pow(s134,-1) + 18*s24*pow(s34,-1) + 8*s24*s34*pow(
      s123,-2) - 24*s24*s34*pow(s123,-1)*pow(s124,-1) - 28*s24*s34*pow(
      s123,-1)*pow(s134,-1) + 6*s24*s34*pow(s124,-1)*pow(s134,-1) + 4*
      s24*s34*pow(s124,-1)*pow(s234,-1) + 28*s24*pow(s123,-1) - 12*s24*
      pow(s124,-1) + 8*s24*pow(s134,-1) + 4*s24*pow(s234,-1) + 2*pow(
      s24,2)*pow(s34,-2) - 6*pow(s24,2)*pow(s34,-1)*pow(s134,-1) - 4*
      pow(s24,2)*pow(s34,-1)*pow(s234,-1) + 4*pow(s24,2)*pow(s123,-2)
       - 12*pow(s24,2)*pow(s123,-1)*pow(s124,-1) - 8*pow(s24,2)*pow(
      s123,-1)*pow(s134,-1) - 4*pow(s24,2)*pow(s124,-1)*pow(s134,-1) - 
      4*pow(s24,2)*pow(s124,-1)*pow(s234,-1) + 6*pow(s24,2)*pow(
      s134,-2) + 4*pow(s24,3)*pow(s34,-1)*pow(s124,-1)*pow(s234,-1) + 
      26*s34*pow(s123,-1) + 10*s34*pow(s124,-1) - 12*s34*pow(s134,-1)
       - 4*s34*pow(s234,-1) + 4*pow(s34,2)*pow(s123,-2) - 16*pow(s34,2)
      *pow(s123,-1)*pow(s134,-1) + 2*pow(s34,2)*pow(s124,-2) - 4*pow(
      s34,2)*pow(s124,-1)*pow(s134,-1);
      A +=  - 4*pow(s34,2)*pow(s124,-1)*pow(s234,-1);

    //Normalize
    A /= pow2(s1234);

    // Return
    return A;

  }

  //***********************************************************************
  // G40 (g_a,q_b,qbar_c,g_d)
  // GGG notation (a = 1, b = 3, c = 4, d = 2)
  // From GGG form code, private communication
  double Matching::G40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double A =
       - 9 + 4*pow(s12,-2)*s13*s14*s23*s24*pow(s34,-2) - 2*pow(s12,-2)*
      pow(s13,2)*pow(s24,2)*pow(s34,-2) - 2*pow(s12,-2)*pow(s14,2)*pow(
      s23,2)*pow(s34,-2) + pow(s12,-1)*pow(s13,-1)*s14*s23 - 2*pow(
      s12,-1)*pow(s13,-1)*s23*s24*s34*pow(s134,-1) + 2*pow(s12,-1)*pow(
      s13,-1)*s23*s24 + pow(s12,-1)*pow(s13,-1)*s23*pow(s24,2)*pow(
      s34,-1) + pow(s12,-1)*pow(s13,-1)*s23*pow(s24,2)*pow(s134,-1) - 
      pow(s12,-1)*pow(s13,-1)*s23*s34 + pow(s12,-1)*pow(s13,-1)*s23*
      pow(s34,2)*pow(s134,-1) + pow(s12,-1)*pow(s13,-1)*pow(s23,3)*pow(
      s34,-1) + pow(s12,-1)*pow(s13,-1)*pow(s23,3)*pow(s134,-1) - 2*
      pow(s12,-1)*s13*s14*pow(s24,-1)*s34*pow(s234,-1) + 2*pow(s12,-1)*
      s13*s14*pow(s24,-1) + 2*pow(s12,-1)*s13*s14*pow(s34,-1) - 4*pow(
      s12,-1)*s13*s14*pow(s234,-1) + pow(s12,-1)*s13*pow(s14,2)*pow(
      s34,-1)*pow(s234,-1) + 6*pow(s12,-1)*s13*s23*pow(s34,-1) + 4*pow(
      s12,-1)*s13*s24*pow(s34,-1) + 2*pow(s12,-1)*s13*s24*pow(s234,-1)
       + 2*pow(s12,-1)*s13*pow(s24,2)*pow(s34,-1)*pow(s234,-1);
      A +=  + pow(s12,-1)*s13*s34*pow(s234,-1) - pow(s12,-1)*s13 + pow(
      s12,-1)*pow(s13,2)*s14*pow(s24,-1)*pow(s34,-1) + pow(s12,-1)*pow(
      s13,2)*s14*pow(s24,-1)*pow(s234,-1) + pow(s12,-1)*pow(s13,2)*s14*
      pow(s34,-1)*pow(s234,-1) - 2*pow(s12,-1)*pow(s13,2)*s24*pow(
      s34,-1)*pow(s234,-1) + 4*pow(s12,-1)*pow(s13,2)*pow(s34,-1) - 2*
      pow(s12,-1)*pow(s13,2)*pow(s234,-1) + pow(s12,-1)*pow(s13,3)*pow(
      s34,-1)*pow(s234,-1) + pow(s12,-1)*s14*s23*pow(s24,-1) - 2*pow(
      s12,-1)*s14*pow(s23,2)*pow(s34,-1)*pow(s134,-1) - pow(s12,-1)*s14
      *pow(s24,-1)*s34 + pow(s12,-1)*s14*pow(s24,-1)*pow(s34,2)*pow(
      s234,-1) + 2*pow(s12,-1)*s14*s24*pow(s34,-1) + 2*pow(s12,-1)*s14*
      s24*pow(s134,-1) + 4*pow(s12,-1)*s14*s24*pow(s234,-1) + 2*pow(
      s12,-1)*s14*pow(s24,2)*pow(s34,-1)*pow(s134,-1) + 2*pow(s12,-1)*
      s14*pow(s24,2)*pow(s34,-1)*pow(s234,-1) + 3*pow(s12,-1)*s14*s34*
      pow(s234,-1) - 2*pow(s12,-1)*s14 + 2*pow(s12,-1)*pow(s14,2)*s23*
      pow(s34,-1)*pow(s134,-1) + 2*pow(s12,-1)*pow(s14,2)*s24*pow(
      s34,-1)*pow(s134,-1);
      A +=  + 2*pow(s12,-1)*pow(s14,2)*s24*pow(s34,-1)*pow(s234,-1) + 2
      *pow(s12,-1)*pow(s14,2)*pow(s34,-1) + 2*pow(s12,-1)*pow(s14,2)*
      pow(s234,-1) + pow(s12,-1)*pow(s14,3)*pow(s24,-1)*pow(s34,-1) + 
      pow(s12,-1)*pow(s14,3)*pow(s24,-1)*pow(s234,-1) + pow(s12,-1)*
      pow(s14,3)*pow(s34,-1)*pow(s234,-1) + 2*pow(s12,-1)*s23*s24*pow(
      s34,-1) - 4*pow(s12,-1)*s23*s24*pow(s134,-1) + pow(s12,-1)*s23*
      pow(s24,2)*pow(s34,-1)*pow(s134,-1) + pow(s12,-1)*s23*s34*pow(
      s134,-1) + pow(s12,-1)*pow(s23,2)*s24*pow(s34,-1)*pow(s134,-1) + 
      4*pow(s12,-1)*pow(s23,2)*pow(s34,-1) + pow(s12,-1)*pow(s23,3)*
      pow(s34,-1)*pow(s134,-1) + pow(s12,-1)*s24*s34*pow(s134,-1) - 
      pow(s12,-1)*s24 + 2*pow(s12,-1)*pow(s24,2)*pow(s34,-1) + pow(
      s12,-1)*pow(s24,3)*pow(s34,-1)*pow(s134,-1) - 2*s12*pow(s13,-1)*
      s23*pow(s34,-1) - 2*s12*pow(s13,-1)*s23*s34*pow(s134,-2) + 8*s12*
      pow(s13,-1)*s23*pow(s134,-1) + s12*pow(s13,-1)*pow(s24,-1)*s34 + 
      2*s12*pow(s13,-1)*s24*pow(s34,-1) - 2*s12*pow(s13,-1)*s24*s34*
      pow(s134,-2);
      A +=  + 6*s12*pow(s13,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - 4*
      s12*pow(s13,-1)*s24*pow(s134,-1) - 4*s12*pow(s13,-1)*s24*pow(
      s234,-1) - 2*s12*pow(s13,-1)*pow(s24,2)*pow(s34,-1)*pow(s234,-1)
       + 4*s12*pow(s13,-1)*pow(s24,2)*pow(s134,-1)*pow(s234,-1) - 4*s12
      *pow(s13,-1)*s34*pow(s134,-1) - 2*s12*pow(s13,-1)*s34*pow(
      s234,-1) + 3*s12*pow(s13,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1)
       + 2*s12*pow(s13,-1) - 2*s12*s13*pow(s24,-1)*s34*pow(s234,-2) - 8
      *s12*s13*s24*pow(s34,-1)*pow(s234,-2) - 4*s12*s13*pow(s24,2)*pow(
      s34,-2)*pow(s234,-2) + 6*s12*s13*pow(s34,-1)*pow(s234,-1) - 6*s12
      *s13*pow(s234,-2) + 2*s12*s14*pow(s24,-1)*s34*pow(s134,-1)*pow(
      s234,-1) - 2*s12*s14*pow(s24,-1)*s34*pow(s234,-2) + 4*s12*s14*
      pow(s24,-1)*pow(s234,-1) + 8*s12*s14*s24*pow(s34,-2)*pow(s134,-1)
       + 8*s12*s14*s24*pow(s34,-2)*pow(s234,-1) + 8*s12*s14*s24*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) - 8*s12*s14*s24*pow(s34,-1)*
      pow(s234,-2) - 4*s12*s14*pow(s24,2)*pow(s34,-2)*pow(s234,-2) - 4*
      s12*s14*pow(s34,-1)*pow(s134,-1);
      A +=  + 8*s12*s14*pow(s34,-1)*pow(s234,-1) + 12*s12*s14*pow(
      s134,-1)*pow(s234,-1) - 6*s12*s14*pow(s234,-2) - 4*s12*pow(s14,2)
      *s23*pow(s34,-2)*pow(s134,-2) - 2*s12*pow(s14,2)*pow(s24,-1)*pow(
      s34,-1)*pow(s134,-1) + 4*s12*pow(s14,2)*pow(s24,-1)*pow(s134,-1)*
      pow(s234,-1) - 4*s12*pow(s14,2)*s24*pow(s34,-2)*pow(s134,-2) - 4*
      s12*pow(s14,2)*pow(s34,-2)*pow(s134,-1) + 8*s12*pow(s14,2)*pow(
      s34,-1)*pow(s134,-1)*pow(s234,-1) + 8*s12*s23*pow(s34,-1)*pow(
      s134,-1) - 2*s12*s23*pow(s134,-2) - 2*s12*pow(s24,-1)*s34*pow(
      s234,-1) + s12*pow(s24,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1)
       - 2*s12*s24*pow(s34,-1)*pow(s134,-1) - 12*s12*s24*pow(s34,-1)*
      pow(s234,-1) - 2*s12*s24*pow(s134,-2) + 12*s12*s24*pow(s134,-1)*
      pow(s234,-1) - 4*s12*pow(s24,2)*pow(s34,-2)*pow(s234,-1) + 8*s12*
      pow(s24,2)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) + 12*s12*pow(
      s34,-1) + 6*s12*s34*pow(s134,-1)*pow(s234,-1) - 6*s12*pow(
      s134,-1) - 10*s12*pow(s234,-1) + pow(s12,2)*pow(s13,-1)*s24*pow(
      s34,-1)*pow(s234,-1);
      A +=  - 3*pow(s12,2)*pow(s13,-1)*s24*pow(s134,-1)*pow(s234,-1) - 
      pow(s12,2)*pow(s13,-1)*pow(s34,-1) - pow(s12,2)*pow(s13,-1)*s34*
      pow(s134,-2) - 3*pow(s12,2)*pow(s13,-1)*s34*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*pow(s13,-1)*pow(s134,-1) + pow(s12,2)*
      pow(s13,-1)*pow(s234,-1) - pow(s12,2)*s14*pow(s24,-1)*pow(s34,-1)
      *pow(s134,-1) + 3*pow(s12,2)*s14*pow(s24,-1)*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*s14*s24*pow(s34,-2)*pow(s134,-1)*pow(
      s234,-1) + 4*pow(s12,2)*s14*pow(s34,-1)*pow(s134,-1)*pow(s234,-1)
       - 2*pow(s12,2)*pow(s14,2)*pow(s34,-2)*pow(s134,-2) - pow(s12,2)*
      pow(s24,-1)*s34*pow(s234,-2) + pow(s12,2)*pow(s24,-1)*pow(
      s234,-1) - 4*pow(s12,2)*s24*pow(s34,-1)*pow(s234,-2) - 2*pow(
      s12,2)*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 5*pow(s12,2)*pow(
      s34,-1)*pow(s134,-1) + 5*pow(s12,2)*pow(s34,-1)*pow(s234,-1) - 
      pow(s12,2)*pow(s134,-2) - 3*pow(s12,2)*pow(s234,-2) + pow(s12,3)*
      pow(s13,-1)*pow(s134,-1)*pow(s234,-1) + pow(s12,3)*pow(s24,-1)*
      pow(s134,-1)*pow(s234,-1);
      A +=  + 2*pow(s12,3)*pow(s34,-1)*pow(s134,-1)*pow(s234,-1) - pow(
      s13,-1)*s14*s23*pow(s24,-1) + 2*pow(s13,-1)*s23*s24*pow(s34,-1)
       - 2*pow(s13,-1)*s23*s24*s34*pow(s134,-2) - 2*pow(s13,-1)*s23*s34
      *pow(s134,-1) + 2*pow(s13,-1)*s23 - 2*pow(s13,-1)*pow(s23,2)*pow(
      s34,-1) - pow(s13,-1)*pow(s23,2)*s34*pow(s134,-2) + 6*pow(s13,-1)
      *pow(s23,2)*pow(s134,-1) + 3*pow(s13,-1)*s24*s34*pow(s234,-1) - 3
      *pow(s13,-1)*s24*pow(s34,2)*pow(s134,-1)*pow(s234,-1) - 2*pow(
      s13,-1)*s24 - 2*pow(s13,-1)*pow(s24,2)*pow(s34,-1) - pow(s13,-1)*
      pow(s24,2)*s34*pow(s134,-2) - 4*pow(s13,-1)*pow(s24,2)*s34*pow(
      s134,-1)*pow(s234,-1) + 2*pow(s13,-1)*pow(s24,2)*pow(s134,-1) + 4
      *pow(s13,-1)*pow(s24,2)*pow(s234,-1) + 2*pow(s13,-1)*pow(s24,3)*
      pow(s34,-1)*pow(s234,-1) - 2*pow(s13,-1)*pow(s24,3)*pow(s134,-1)*
      pow(s234,-1) - 2*pow(s13,-1)*s34 + 2*pow(s13,-1)*pow(s34,2)*pow(
      s134,-1) + pow(s13,-1)*pow(s34,2)*pow(s234,-1) - pow(s13,-1)*pow(
      s34,3)*pow(s134,-1)*pow(s234,-1) - 2*s13*s14*pow(s24,-1)*s34*pow(
      s234,-2);
      A +=  + 2*s13*s14*pow(s24,-1)*pow(s234,-1) + 4*s13*s14*s24*pow(
      s34,-2)*pow(s234,-1) - 8*s13*s14*s24*pow(s34,-1)*pow(s234,-2) - 4
      *s13*s14*pow(s24,2)*pow(s34,-2)*pow(s234,-2) + 4*s13*s14*pow(
      s34,-1)*pow(s234,-1) - 6*s13*s14*pow(s234,-2) - 2*s13*pow(s24,-1)
      *s34*pow(s234,-1) - 14*s13*s24*pow(s34,-1)*pow(s234,-1) - 4*s13*
      pow(s24,2)*pow(s34,-2)*pow(s234,-1) + 12*s13*pow(s34,-1) - 12*s13
      *pow(s234,-1) - pow(s13,2)*pow(s24,-1)*s34*pow(s234,-2) - 4*pow(
      s13,2)*s24*pow(s34,-1)*pow(s234,-2) - 2*pow(s13,2)*pow(s24,2)*
      pow(s34,-2)*pow(s234,-2) + 4*pow(s13,2)*pow(s34,-1)*pow(s234,-1)
       - 3*pow(s13,2)*pow(s234,-2) + 4*s14*s23*s24*pow(s34,-2)*pow(
      s134,-1) - 8*s14*s23*pow(s34,-1)*pow(s134,-1) - s14*pow(s24,-1)*
      s34*pow(s134,-1) - 2*s14*pow(s24,-1)*s34*pow(s234,-1) + s14*pow(
      s24,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 2*s14*pow(s24,-1)
       + 8*s14*s24*pow(s34,-2) + 6*s14*s24*pow(s34,-1)*pow(s134,-1) + 4
      *s14*pow(s24,2)*pow(s34,-2)*pow(s134,-1) - 4*s14*pow(s24,2)*pow(
      s34,-2)*pow(s234,-1);
      A +=  + 2*s14*pow(s34,-1) + 4*s14*pow(s134,-1) - 4*pow(s14,2)*s23
      *s24*pow(s34,-2)*pow(s134,-2) - 4*pow(s14,2)*s23*pow(s34,-2)*pow(
      s134,-1) - 2*pow(s14,2)*pow(s23,2)*pow(s34,-2)*pow(s134,-2) + 2*
      pow(s14,2)*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1) - pow(s14,2)
      *pow(s24,-1)*s34*pow(s234,-2) - 2*pow(s14,2)*pow(s24,-1)*pow(
      s134,-1) + 4*pow(s14,2)*pow(s24,-1)*pow(s234,-1) - 4*pow(s14,2)*
      s24*pow(s34,-2)*pow(s134,-1) + 4*pow(s14,2)*s24*pow(s34,-2)*pow(
      s234,-1) - 4*pow(s14,2)*s24*pow(s34,-1)*pow(s234,-2) - 2*pow(
      s14,2)*pow(s24,2)*pow(s34,-2)*pow(s134,-2) - 2*pow(s14,2)*pow(
      s24,2)*pow(s34,-2)*pow(s234,-2) - 2*pow(s14,2)*pow(s34,-2) + 6*
      pow(s14,2)*pow(s34,-1)*pow(s134,-1) + 10*pow(s14,2)*pow(s34,-1)*
      pow(s234,-1) - 3*pow(s14,2)*pow(s234,-2) - 2*pow(s14,3)*pow(
      s24,-1)*pow(s34,-1)*pow(s134,-1) + 2*pow(s14,3)*pow(s24,-1)*pow(
      s134,-1)*pow(s234,-1) - 2*s23*s24*pow(s134,-2) + 12*s23*pow(
      s34,-1) - 4*s23*pow(s134,-1) + 6*pow(s23,2)*pow(s34,-1)*pow(
      s134,-1);
      A +=  - pow(s23,2)*pow(s134,-2) - pow(s24,-1)*s34 + pow(s24,-1)*
      pow(s34,2)*pow(s234,-1) - 4*s24*pow(s34,-1) - 2*s24*pow(s134,-1)
       + 8*s24*pow(s234,-1) - 2*pow(s24,2)*pow(s34,-2) + 4*pow(s24,2)*
      pow(s34,-1)*pow(s134,-1) + 6*pow(s24,2)*pow(s34,-1)*pow(s234,-1)
       - pow(s24,2)*pow(s134,-2) + 3*s34*pow(s134,-1) + 5*s34*pow(
      s234,-1);

    //Normalize
    A /= pow2(s1234);

    // Return
    return A;

  }

  //***********************************************************************
  // ~G40 (g_a,q_b,qbar_c,g_d)
  // GGG notation (a = 1, b = 3, c = 4, d = 2)
  // From GGG form code, private communication
  double Matching::G40NLC(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate argument # to GGG notation
    double s12=m2(pn[a],pn[d]);
    double s13=m2(pn[a],pn[b]);
    double s14=m2(pn[a],pn[c]);
    double s23=m2(pn[d],pn[b]);
    double s24=m2(pn[d],pn[c]);
    double s34=m2(pn[b],pn[c]);
    double s234=m2(pn[d],pn[b],pn[c]);
    double s134=m2(pn[a],pn[b],pn[c]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);

    double ANLC =
       - 2 + 2*s12*pow(s13,-1)*pow(s14,-1)*s23 + 2*s12*pow(s13,-1)*pow(
      s14,-1)*s24 - s12*pow(s13,-1)*pow(s23,-1)*pow(s34,3)*pow(s134,-1)
      *pow(s234,-1) - 2*s12*pow(s13,-1)*s23*s34*pow(s134,-2) + s12*pow(
      s13,-1)*pow(s24,-1)*s34 - 2*s12*pow(s13,-1)*s24*s34*pow(s134,-2)
       + 4*s12*pow(s13,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) - 2*s12*
      pow(s13,-1)*s24*pow(s134,-1) - 2*s12*pow(s13,-1)*s24*pow(s234,-1)
       - 4*s12*pow(s13,-1)*s34*pow(s134,-1) - 2*s12*pow(s13,-1)*s34*
      pow(s234,-1) + 2*s12*pow(s13,-1)*pow(s34,2)*pow(s134,-1)*pow(
      s234,-1) + 2*s12*pow(s13,-1) + 2*s12*s13*pow(s23,-1)*pow(s24,-1)
       - 2*s12*s13*pow(s23,-1)*s34*pow(s234,-2) - 2*s12*s13*pow(s24,-1)
      *s34*pow(s234,-2) - 2*s12*s13*pow(s24,-1)*pow(s234,-1) - 4*s12*
      s13*pow(s234,-2) + s12*pow(s14,-1)*pow(s23,-1)*s34 - 2*s12*pow(
      s14,-1)*s23*s34*pow(s134,-2) - 2*s12*pow(s14,-1)*s23*pow(s134,-1)
       - s12*pow(s14,-1)*pow(s24,-1)*pow(s34,3)*pow(s134,-1)*pow(
      s234,-1);
      ANLC +=  - 2*s12*pow(s14,-1)*s24*s34*pow(s134,-2) - 4*s12*pow(
      s14,-1)*s24*s34*pow(s134,-1)*pow(s234,-1) + 2*s12*pow(s14,-1)*s24
      *pow(s234,-1) - 2*s12*pow(s14,-1)*pow(s34,2)*pow(s134,-1)*pow(
      s234,-1) + 2*s12*s14*pow(s23,-1)*pow(s24,-1) + 4*s12*s14*pow(
      s23,-1)*s34*pow(s134,-1)*pow(s234,-1) - 2*s12*s14*pow(s23,-1)*s34
      *pow(s234,-2) - 2*s12*s14*pow(s23,-1)*pow(s134,-1) - 2*s12*s14*
      pow(s23,-1)*pow(s234,-1) - 4*s12*s14*pow(s24,-1)*s34*pow(s134,-1)
      *pow(s234,-1) - 2*s12*s14*pow(s24,-1)*s34*pow(s234,-2) + 2*s12*
      s14*pow(s24,-1)*pow(s134,-1) - 4*s12*s14*pow(s234,-2) - 2*s12*
      pow(s23,-1)*s34*pow(s134,-1) - 4*s12*pow(s23,-1)*s34*pow(s234,-1)
       + 2*s12*pow(s23,-1)*pow(s34,2)*pow(s134,-1)*pow(s234,-1) + 2*s12
      *pow(s23,-1) - 4*s12*s23*pow(s134,-2) - 2*s12*pow(s24,-1)*pow(
      s34,2)*pow(s134,-1)*pow(s234,-1) - 4*s12*s24*pow(s134,-2) - 8*s12
      *s34*pow(s134,-1)*pow(s234,-1) - 4*s12*pow(s134,-1) - 4*s12*pow(
      s234,-1) + 2*pow(s12,2)*pow(s13,-1)*pow(s14,-1) - pow(s12,2)*pow(
      s13,-1)*s34*pow(s134,-2);
      ANLC +=  - 3*pow(s12,2)*pow(s13,-1)*s34*pow(s134,-1)*pow(s234,-1)
       - pow(s12,2)*pow(s13,-1)*pow(s134,-1) + pow(s12,2)*pow(s13,-1)*
      pow(s234,-1) - pow(s12,2)*pow(s14,-1)*s34*pow(s134,-2) - 3*pow(
      s12,2)*pow(s14,-1)*s34*pow(s134,-1)*pow(s234,-1) - pow(s12,2)*
      pow(s14,-1)*pow(s134,-1) + pow(s12,2)*pow(s14,-1)*pow(s234,-1) + 
      2*pow(s12,2)*pow(s23,-1)*pow(s24,-1) - 3*pow(s12,2)*pow(s23,-1)*
      s34*pow(s134,-1)*pow(s234,-1) - pow(s12,2)*pow(s23,-1)*s34*pow(
      s234,-2) + pow(s12,2)*pow(s23,-1)*pow(s134,-1) - pow(s12,2)*pow(
      s23,-1)*pow(s234,-1) - 3*pow(s12,2)*pow(s24,-1)*s34*pow(s134,-1)*
      pow(s234,-1) - pow(s12,2)*pow(s24,-1)*s34*pow(s234,-2) + pow(
      s12,2)*pow(s24,-1)*pow(s134,-1) - pow(s12,2)*pow(s24,-1)*pow(
      s234,-1) - 2*pow(s12,2)*pow(s134,-2) - 4*pow(s12,2)*pow(s134,-1)*
      pow(s234,-1) - 2*pow(s12,2)*pow(s234,-2) - pow(s12,3)*pow(s13,-1)
      *pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1) - pow(s12,3)*pow(
      s14,-1)*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1) + 2*pow(s13,-1)
      *pow(s14,-1)*pow(s23,2);
      ANLC +=  + 2*pow(s13,-1)*pow(s14,-1)*pow(s24,2) - pow(s13,-1)*s14
      *s23*pow(s24,-1) - 2*pow(s13,-1)*s23*s24*s34*pow(s134,-2) - 2*
      pow(s13,-1)*s23*s34*pow(s134,-1) + 2*pow(s13,-1)*s23 - pow(
      s13,-1)*pow(s23,2)*s34*pow(s134,-2) + pow(s13,-1)*pow(s23,2)*pow(
      s134,-1) + 2*pow(s13,-1)*s24*s34*pow(s134,-1) + 2*pow(s13,-1)*s24
      *s34*pow(s234,-1) - 2*pow(s13,-1)*s24*pow(s34,2)*pow(s134,-1)*
      pow(s234,-1) - 2*pow(s13,-1)*s24 - pow(s13,-1)*pow(s24,2)*s34*
      pow(s134,-2) - 2*pow(s13,-1)*pow(s24,2)*s34*pow(s134,-1)*pow(
      s234,-1) - pow(s13,-1)*pow(s24,2)*pow(s134,-1) + 2*pow(s13,-1)*
      pow(s24,2)*pow(s234,-1) + pow(s13,-1)*pow(s34,2)*pow(s234,-1) - 
      pow(s13,-1)*pow(s34,3)*pow(s134,-1)*pow(s234,-1) - s13*pow(
      s14,-1)*pow(s23,-1)*s24 - 2*s13*s14*pow(s23,-1)*s34*pow(s234,-2)
       - 2*s13*s14*pow(s24,-1)*s34*pow(s234,-2) - 4*s13*s14*pow(
      s234,-2) - 2*s13*pow(s23,-1)*s34*pow(s234,-1) + 2*s13*pow(s23,-1)
       - 2*s13*pow(s234,-1) + 2*pow(s13,2)*pow(s23,-1)*pow(s24,-1) - 
      pow(s13,2)*pow(s23,-1)*s34*pow(s234,-2);
      ANLC +=  + pow(s13,2)*pow(s23,-1)*pow(s234,-1) - pow(s13,2)*pow(
      s24,-1)*s34*pow(s234,-2) - pow(s13,2)*pow(s24,-1)*pow(s234,-1) - 
      2*pow(s13,2)*pow(s234,-2) - 2*pow(s14,-1)*s23*s24*s34*pow(
      s134,-2) - pow(s14,-1)*pow(s23,2)*s34*pow(s134,-2) - pow(s14,-1)*
      pow(s23,2)*pow(s134,-1) + 2*pow(s14,-1)*s24*s34*pow(s234,-1) - 2*
      pow(s14,-1)*s24*pow(s34,2)*pow(s134,-1)*pow(s234,-1) - pow(
      s14,-1)*pow(s24,2)*s34*pow(s134,-2) - 2*pow(s14,-1)*pow(s24,2)*
      s34*pow(s134,-1)*pow(s234,-1) + pow(s14,-1)*pow(s24,2)*pow(
      s134,-1) + 2*pow(s14,-1)*pow(s24,2)*pow(s234,-1) + pow(s14,-1)*
      pow(s34,2)*pow(s234,-1) - pow(s14,-1)*pow(s34,3)*pow(s134,-1)*
      pow(s234,-1) + 2*s14*pow(s23,-1)*s34*pow(s134,-1) + 2*s14*pow(
      s23,-1)*s34*pow(s234,-1) - 2*s14*pow(s23,-1)*pow(s34,2)*pow(
      s134,-1)*pow(s234,-1) - 2*s14*pow(s23,-1) + 2*s14*pow(s24,-1)*s34
      *pow(s134,-1) - 2*s14*pow(s24,-1)*pow(s34,2)*pow(s134,-1)*pow(
      s234,-1) - 2*s14*pow(s234,-1) + 2*pow(s14,2)*pow(s23,-1)*pow(
      s24,-1);
      ANLC +=  - 2*pow(s14,2)*pow(s23,-1)*s34*pow(s134,-1)*pow(s234,-1)
       - pow(s14,2)*pow(s23,-1)*s34*pow(s234,-2) + 2*pow(s14,2)*pow(
      s23,-1)*pow(s134,-1) - pow(s14,2)*pow(s23,-1)*pow(s234,-1) - 2*
      pow(s14,2)*pow(s24,-1)*s34*pow(s134,-1)*pow(s234,-1) - pow(s14,2)
      *pow(s24,-1)*s34*pow(s234,-2) + 2*pow(s14,2)*pow(s24,-1)*pow(
      s134,-1) + pow(s14,2)*pow(s24,-1)*pow(s234,-1) - 2*pow(s14,2)*
      pow(s234,-2) + pow(s23,-1)*pow(s34,2)*pow(s134,-1) - pow(s23,-1)*
      pow(s34,3)*pow(s134,-1)*pow(s234,-1) - 4*s23*s24*pow(s134,-2) - 2
      *s23*pow(s134,-1) - 2*pow(s23,2)*pow(s134,-2) + pow(s24,-1)*pow(
      s34,2)*pow(s134,-1) - pow(s24,-1)*pow(s34,3)*pow(s134,-1)*pow(
      s234,-1) - 2*s24*pow(s134,-1) - 2*pow(s24,2)*pow(s134,-2) - 2*s34
      *pow(s134,-1) - 2*s34*pow(s234,-1);

    //Normalize
    ANLC /= pow2(s1234);

    // Return
    return ANLC;

  }

  //***********************************************************************
  // H40 (q_a,qbar_b,q'_c,qbar'_d)
  // From GGG, JHEP 0509:056,2005
  double Matching::H40(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate Vincia ordering to GGG notation
    double s12=m2(pn[a],pn[b]);
    double s13=m2(pn[a],pn[c]);
    double s14=m2(pn[a],pn[d]);
    double s23=m2(pn[b],pn[c]);
    double s24=m2(pn[b],pn[d]);
    double s34=m2(pn[c],pn[d]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A = 
        2.0*pow2(s13*s24 - s14*s23)/pow2(s12)/pow2(s34)
      + ( - 2.0*s13*s24 - 2.0*s14*s23
	  + pow2(s13) + pow2(s14) + pow2(s23) + pow2(s24) )/s12/s34
      + 2; 
    
    A /= pow2(s1234);
    
    return A;
    
  }

  //***********************************************************************
  // H40 (q_a,qbar_b,q'_c,qbar'_d)
  // From GGG Form code, private communication
  double Matching::H40frm(vector<Vec4>& pn, int a, int b, int c, int d) {

    // Compute invariants
    // Translate Vincia ordering to GGG notation
    double s12=m2(pn[a],pn[b]);
    double s13=m2(pn[a],pn[c]);
    double s14=m2(pn[a],pn[d]);
    double s23=m2(pn[b],pn[c]);
    double s24=m2(pn[b],pn[d]);
    double s34=m2(pn[c],pn[d]);
    double s1234=m2(pn[a],pn[b],pn[c],pn[d]);
    
    double A = 
      2 - 4*pow(s12,-2)*s13*s14*s23*s24*pow(s34,-2) + 2*pow(s12,-2)*
      pow(s13,2)*pow(s24,2)*pow(s34,-2) + 2*pow(s12,-2)*pow(s14,2)*pow(
      s23,2)*pow(s34,-2) - 2*pow(s12,-1)*s13*s24*pow(s34,-1) + pow(
      s12,-1)*pow(s13,2)*pow(s34,-1) - 2*pow(s12,-1)*s14*s23*pow(
      s34,-1) + pow(s12,-1)*pow(s14,2)*pow(s34,-1) + pow(s12,-1)*pow(
      s23,2)*pow(s34,-1) + pow(s12,-1)*pow(s24,2)*pow(s34,-1);

    A /= pow2(s1234);
    return A;
  }

  //***********************************************************************
  // Z->3 at one Loop

  //***********************************************************************
  // Leading-color 3-parton 1-loop finite terms
  double Matching::A31fin(double sar,double srb,double sab,double s) {

    // Rewrite to GGG notation (1,3,2) instead of (a,r,b)
    double s13  = sar;
    double y13  = sar/s;
    double s23  = srb;
    double y23  = srb/s;
    double s12  = sab;

    // A31fin (A30 piece, includes CF, so only multiply by CA here)
    double A= -A30(sar,srb,sab,s) * (A31auxR(y13,y23)
					+ 5.0/3.0*log(y13*y23)); 

    // Remaining terms
    A += ( 1.0 + (s12+s23)/(2*s13) + (s12+s13)/(2*s23) 
	       - s13/(2*(s12+s13)) - s23/(2*(s12+s23))
	       + log(y13)*(2-s13*s23/(2*pow2(s12+s13))+2*(s13-s23)/(s12+s23))
	       + log(y23)*(2-s13*s23/(2*pow2(s12+s13))+2*(s23-s13)/(s12+s23))
	       )/s;
    // Return answer
    return A;

  }
  // Leading-color 3-parton 1-loop finite terms
  double Matching::A31finNLC(double sar,double srb,double sab,double s) {

    // Rewrite to GGG notation (1,3,2) instead of (a,r,b)
    double s13  = sar;
    double y13  = sar/s;
    double s23  = srb;
    double y23  = srb/s;
    double s12  = sab;
    double y12  = sab/s;

    // A31fin (A30 piece, includes CF, so just divide by -CA here)
    double A= A30(sar,srb,sab,s) * (A31auxR(y12,y13) + A31auxR(y12,y23)
				    + 1.5*log(y12));

    // Remaining terms 
    A += ( - (s12+s23)/(2*s13) - (s12+s13)/(2*s23) 
	   + s12/(2*(s12+s13)) + s12/(2*(s12+s23)) + 2*s12/(s13+s23) 
	   + 2*log(y12)*(2*s12/(s13+s23)+pow2(s12)/pow2(s13+s23))
	   + log(y13)/2*(s12*s23/pow2(s12+s23)+4*(s12+s13)/(s12+s23))
	   + log(y23)/2*(s12*s13/pow2(s12+s13)+4*(s12+s23)/(s12+s13))
	   + A31auxR(y12,y13)*(2*s12+s13)/s23 
	   + A31auxR(y12,y23)*(2*s12+s23)/s13
	   )/s;
    // Return answer
    return A;

  }
  // The R function, eq.(5.11) in the GGG paper.
  double Matching::A31auxR(double y, double z) {
    double R=log(y)*log(z)-log(y)*log(1-y)-log(z)*log(1-z)
      + pow2(M_PI)/6 - Li2(y)-Li2(z);
    return R;
  }

  // *********************************************************************
  // Kosower's sector functions

  double Matching::Asec30gks(double sar, double srb, double s) {

    double A = 2.0*s/sar/srb - 2.0/sar - 2.0/srb + srb/sar/s + sar/srb/s
      - pow2(srb/s)/sar - pow2(sar/s)/srb;

    return A;

  }

  double Matching::Dsec30gks(double sar,double srb,double s) { 
    return Dsec30ska(sar,srb,s);}
  double Matching::Esec30gks(double sar,double srb,double s) {
    return Esec30ska(sar,srb,s);}
  double Matching::Fsec30gks(double sar,double srb,double s) {
    return Fsec30ska(sar,srb,s);}
  double Matching::Gsec30gks(double sar,double srb,double s) {
    return Gsec30ska(sar,srb,s);}

  // *********************************************************************
  // Skands' experimental sector functions

  double Matching::Asec30ska(double sar,double srb,double s) {
    return A30(sar,srb,s-sar-srb,s);
  }
  double Matching::Dsec30ska(double sar,double srb,double s) {
    return D30(sar,srb,s-sar-srb,s)  
      - 2.0/srb + sar/srb/s - pow2(sar/s)/srb + 2.0*s/srb/(s-sar);
  }
  double Matching::Esec30ska(double sar,double srb,double s) {
    return 1.0/srb - 2.0*sar/srb/s + 2.0 * pow2(sar/s)/srb;
  }
  double Matching::Fsec30ska(double sar,double srb,double s) {
    return 2.0*F30(sar,srb,s-sar-srb,s)  
      - 2.0*s/sar/srb + 2.0*s/srb/(s-sar) + 2.0*s/sar/(s-srb);
  }
  double Matching::Gsec30ska(double sar,double srb,double s) {
    return 0.0/sar + 0.0/srb + 0.0*sar/srb/s;
  }

}
 
